#define GEOM_SOURCE
#include "geomlib/polynomial.h"
#include "geomlib/bernstein.h"
#include "geomlib/bezier.h"


/* geom_bezier_line_intersection_times
 * Function returns list of parameters, where the curve crosses a line (in parametric form). */

#if 0
int geom_bezier_line_intersection_times(struct geom_bezier *b, struct geom_point *a, struct geom_point *v, int flags, double result[])
{
	int i;
	double coef[4];

	BASSERT(geom_bezier_valid(b) && geom_point_valid(a) && geom_point_valid(v) &&
		!(flags & ~(GEOM_SOLVE_LEFT_ONLY | GEOM_SOLVE_MULTIPLICITY)) && result);

	/* nonrational curve */
	if (b->flags & GEOM_BEZIER_NONRATIONAL) {
		for (i = 0; i <= b->degree; i++) {
			if (!isfinite(coef[i] = v->x * (b->pt[i].y - a->y) + v->y * (b->pt[i].x - a->x)))
				return -1;
		}
	}

	/* rational curve */
	else {
		for (i = 0; i <= b->degree; i++) {
			if (!isfinite(coef[i] = (v->x * (b->pt[i].y - a->y) + v->y * (b->pt[i].x - a->x)) * b->pt[i].w))
				return -1;
		}
	}

	/* solve polynomial */
	return geom_bernstein_solve(b->degree, coef, GEOM_SOLVE_UNIT_INTERVAL | flags, result);
}
#endif


/* geom_bezier_resultant_matrix
 *
 * Function computes from two polynomials f and g of degree n
 * the coefficient matrix used when converting curve to its implicit form.
 *
 * This matrix represents expression:
 *   (f(s)*g(t) - f(t)*g(s)) / (s-t) = sum (m[i][j] * s^i * t^j)
 */

static int geom_bezier_resultant_matrix(int n, double f[4], double g[4], double m[3][3])
{
	int i, j, k;
	double c;

	/* clean matrix */
	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
			m[i][j] = 0;

	/* compute matrix entries */
	for (i = 0; i <= n; i++) {
		for (j = 0; j <= i; j++) {
			c = f[i] * g[j];
			for (k = 0; k < i - j; k++)
				m[i - 1 - k][k + j] += c;
		}
		for (j = i + 1; j <= n; j++) {
			c = -f[i] * g[j];
			for (k = 0; k < j - i; k++)
				m[j - 1 - k][k + i] += c;
		}
	}

	/* test for numeric overflow */
	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
			if (!isfinite(m[i][j]))
				return -1;

	return 0;
}


/* geom_bezier_to_power_matrix */

/*static int geom_bezier_to_power_matrix(struct geom_bezier *b, double p[3][4])
{
	int i;
	double bernstein[3][4];

	BASSERT(geom_bezier_valid(b) && p);

	for (i = 0; i <= b->degree; i++) {
		bernstein[0][i] = b->pt[i].x * b->pt[i].w;
		bernstein[1][i] = b->pt[i].y * b->pt[i].w;
		bernstein[2][i] = b->pt[i].w;
	}

	for (i = 0; i < 3; i++)
		if (geom_bernstein_to_power(b->degree, bernstein[i], p[i]) < 0)
			return -1;

	return 0;
}*/


/* geom_bezier_to_implicit_symbolic_matrix
 *
 * Function express symbolic matrix
 *
 *   P[i][j] = (X * p[0][i][j][0] + Y * p[1][i][j] + p[2][i][j]),
 *
 * which determinant (polynome in X and Y) represents the implicit form of bezier b.
 */

static int geom_bezier_to_implicit_symbolic_matrix(struct geom_bezier *b, double p[3][3][3])
{
	int i;
	double coef[3][4];

	ITEM_ASSERT(bezier, b);
	PASSERT(p);

	for (i = 0; i <= BEZIER_DEGREE(b); i++) {
		coef[2][i] = geom_binomial(BEZIER_DEGREE(b), i) * b->pt[i].w;
		coef[0][i] = coef[2][i] * b->pt[i].x;
		coef[1][i] = coef[2][i] * b->pt[i].y;
		if (!isfinite(coef[0][i]) || !isfinite(coef[1][i]))
			return -1;
	}

	if (geom_bezier_resultant_matrix(BEZIER_DEGREE(b), coef[2], coef[0], p[1]) < 0)
		return -1;
	if (geom_bezier_resultant_matrix(BEZIER_DEGREE(b), coef[0], coef[1], p[2]) < 0)
		return -1;
	if (geom_bezier_resultant_matrix(BEZIER_DEGREE(b), coef[1], coef[2], p[0]) < 0)
		return -1;

	return 0;
}


/* geom_bezier_to_implicit_matrix */

/*int geom_bezier_to_implicit_matrix(struct geom_bezier *b, double m[4][4])
{
	int i, j, k, pk, n, perm[4], perm_tmp[4];
	double p[3][3][3], q[5][5];

	BASSERT(geom_bezier_valid(b) && m);

	if (geom_bezier_to_implicit_symbolic_matrix(b, p) < 0)
		return -1;

	n = b->degree;
	for (i = 0; i <= n; i++)
		for (j = 0; j <= n; j++)
			m[i][j] = 0;

	for (i = 1; i <= n + 1; i++)
		q[0][i] = q[i][0] = 0;

	geom_permutation_generate_init(n, perm, perm_tmp);
	while (geom_permutation_generate(n, perm, perm_tmp)) {
		for (i = 1; i <= n + 1; i++)
			for (j = 1; j <= n + 2 - i; j++)
				q[i][j] = 0;
		q[1][1] = geom_permutation_sign(n, perm);
		for (k = 0; k < n; k++) {
			pk = perm[k] - 1;
			for (i = k + 1; i >= 0; i--)
				for (j = k + 1 - i; j >= 0; j--)
					q[i + 1][j + 1] = p[0][k][pk] * q[i][j + 1] +
					    p[1][k][pk] * q[i + 1][j] + p[2][k][pk] * q[i + 1][j + 1];
		}
		for (i = 0; i <= n; i++)
			for (j = 0; j <= n - i; j++)
				m[i][j] += q[i + 1][j + 1];
	}

	for (i = 0; i <= n; i++)
		for (j = 0; j <= n - i; j++)
			if (!isfinite(m[i][j]))
				return -1;

	return 0;
}*/


/* geom_bezier_intersection_times
   TODO:
   - zatim spis jen takovy pokus... neni moc numericky
     stabilni ani prilis rychle
   - nepodporuje vicenasobne body a podobne zakernosti
   - bylo by vhodne specialne osetrit spolecne krajni body
*/

int geom_bezier_intersection_times(struct geom_bezier *b1, struct geom_bezier *b2, uns flags, double result1[], double result2[])
{
	int d, n, m, nm, i, j, k, perm[4], perm_tmp[4];
	double p[3][3][3], q[3][3][4], det[10], a[4][10], /*b[3][3], */t, s;
	struct geom_point pt, pt2;
	struct geom_bezier b3, b4;
	int err = GEOM_ERR_NUMERIC;
	b3.bbox = (struct geom_rectangle){{0,0}, {1,1}};

	ITEM_ASSERT(bezier, b1);
	ITEM_ASSERT(bezier, b2);

	GEOM_INSTANCE_INIT(&b3, bezier);
	GEOM_INSTANCE_INIT(&b4, bezier);

	switch (BEZIER_DEGREE(b2)) {
		case GEOM_BEZIER_QUADRATIC:
			INSTANCE_FLAGS(&b3) = INSTANCE_FLAGS(b2);
			BEZIER_DEGREE(&b3) = GEOM_BEZIER_LINEAR;
			b3.pt[0] = b2->pt[0];
			b3.pt[1] = b2->pt[2];
			geom_bezier_elevate(&b3, GEOM_BEZIER_QUADRATIC, &b4);
			if (0.001 >
					SQR(b2->pt[1].x - b4.pt[1].x) +
					SQR(b2->pt[1].y - b4.pt[1].x) +
					SQR(b2->pt[1].w - b4.pt[1].w)) {
				DBG("decreasing second curve degree to %d\n", BEZIER_DEGREE(&b4));
				b2 = &b4;
			}
			break;

		case GEOM_BEZIER_CUBIC:
			INSTANCE_FLAGS(&b3) = INSTANCE_FLAGS(b2);
			b3.pt[0] = b2->pt[0];
			for (BEZIER_DEGREE(&b3) = GEOM_BEZIER_LINEAR; BEZIER_DEGREE(&b3) <= GEOM_BEZIER_QUADRATIC; BEZIER_DEGREE(&b3)++) {
				if (BEZIER_DEGREE(&b3) == GEOM_BEZIER_LINEAR)
					b3.pt[1] = b2->pt[3];
				else {
					/* TODO */
					b3.pt[1].x = b2->pt[0].x + 1.5 * (b2->pt[1].x - b2->pt[0].x);
					b3.pt[1].y = b2->pt[0].y + 1.5 * (b2->pt[1].y - b2->pt[0].y);
					b3.pt[2] = b2->pt[3];
				}
				geom_bezier_elevate(&b3, GEOM_BEZIER_CUBIC, &b4);
				if (0.001 >
						SQR(b2->pt[1].x - b4.pt[1].x) +
						SQR(b2->pt[1].y - b4.pt[1].x) +
						SQR(b2->pt[1].w - b4.pt[1].w) +
						SQR(b2->pt[2].x - b4.pt[2].x) +
						SQR(b2->pt[2].y - b4.pt[2].x) +
						SQR(b2->pt[2].w - b4.pt[2].w)) {
					DBG("decreasing second curve degree to %d\n", BEZIER_DEGREE(&b4));
					b2 = &b4;
					break;
				}
			}
			break;
	};

	/* convert b2 to implicit form P */
	if (geom_bezier_to_implicit_symbolic_matrix(b2, p) < 0)
		goto EXIT;

	DBG("second curve symbolic matrix:\n");
	for (i = 0; i < BEZIER_DEGREE(b2); i++)
		for (j = 0; j < BEZIER_DEGREE(b2); j++)
		{
			DBG("  [%d,%d]", i, j);
			for (k = 0; k < 3; k++)
				DBG(" %f", p[k][i][j]);
			DBG("\n");
		}

	/* substitute b1 to this symbolic matrix and multiply by w,
	 * matrix Q now contains bernstein polynomials in t of degree n (coefficients stored in q[i][j][0..n]) */
	DBG("matrix after first curve substitution:\n");
	n = BEZIER_DEGREE(b1);
	m = BEZIER_DEGREE(b2);
	nm = n * m;
	for (i = 0; i < m; i++)
		for (j = 0; j < m; j++)
		{
			DBG("  [%d,%d]", i, j);
			for (k = 0; k <= n; k++)
			{
				if (!isfinite(q[i][j][k] = b1->pt[k].w * (b1->pt[k].x * p[0][i][j] + b1->pt[k].y * p[1][i][j] + p[2][i][j])))
					goto EXIT;
				DBG(" %f", q[i][j][k]);
			}
			DBG("\n");
		}

	/* compute determinant of matrix Q (degree n*m bernstein polynomial), store to det[0..nm] */
	for (i = 0; i <= nm; i++)
		det[i] = 0;

	geom_permutation_generate_init(m, perm, perm_tmp);
	while (geom_permutation_generate(m, perm, perm_tmp)) {
		d = 0;
		a[0][0] = geom_permutation_sign(m, perm);
		for (i = 0; i < m; i++) {
			d = geom_bernstein_mul(d, a[i], n, q[i][perm[i] - 1], a[i + 1]);
			if (d < 0)
				goto EXIT;
		}
		for (i = 0; i <= nm; i++)
			det[i] += a[m][i];
	}

	for (i = 0; i <= nm; i++)
		if (!isfinite(det[i]))
			goto EXIT;

	DBG("bernstein polynomial to solve:");
	for (i = 0; i <= nm; i++)
		DBG(" %f", det[i]);
	DBG("\n");

	/* det[0..nm] bernstein polynomial roots are the intersection times (on curve b1) */
	if ((n = geom_bernstein_solve(nm, det, GEOM_SOLVE_UNIT_INTERVAL | flags, result1)) < 0)
		goto EXIT;

	DBG("found %d roots:", n);
	for (i = 0; i < n; i++)
		DBG(" %f", result1[i]);
	DBG("\n");

	/* */
	if (1 /*result2*/) {
		j = 0;
		for (i = 0; i < n; i++) {
			result1[j] = result1[i];
			if (geom_bezier_point_at_time(b1, result1[j], &pt) < 0)
				goto EXIT;

			/*for (j = 0; j < m; j++)
				for (k = 0; k < m; k++)
					b[j][k] = pt.x * p[0][j][k] + pt.y * p[1][j][k] + p[2][j][k];*/
			/* TODO */

			t = geom_bezier_time_of(b2, &pt);
			if (isnan(t))
				goto EXIT;
			geom_bezier_point_at_time(b2, t, &pt2);
			s = hypot(pt.x - pt2.x, pt.y - pt2.y);
			if (isfinite(s) && fabs(s) < 0.001) {
				if (result2)
					result2[j] = t;
				j++;
			}
		}
		n = j;
	}

	DBG("found %d intersections\n", n);

	err = n;
EXIT:	
	GEOM_INSTANCE_DESTROY2(&b3, bezier);
	GEOM_INSTANCE_DESTROY2(&b4, bezier);
	return err;
}


/* geom_beziers_intersections */

int geom_beziers_intersections(struct geom_bezier *bezier1, struct geom_bezier *bezier2, uns flags, struct garr *result)
{
	ITEM_ASSERT(bezier, bezier1);
	ITEM_ASSERT(bezier, bezier2);
	PASSERT(result);

	/* TODO */
	{
		double times1[9];
		double times2[9];
		int i, n;
		struct geom_intersection *inter;
		if ((n = geom_bezier_intersection_times(bezier1, bezier2, 0, times1, times2)) < 0)
			return GEOM_ERR_NUMERIC;
		for (i = 0; i < n; i++) {
			inter = (struct geom_intersection *)garr_get(result, i);
			inter->time1 = times1[i];
			inter->time2 = times2[i];
			if (!isfinite(inter->time1) || !isfinite(inter->time2))
				return GEOM_ERR_NUMERIC;
			if (flags & GEOM_INTERSECTION_POINT)
				geom_bezier_point_at_time(bezier1, inter->time1, &inter->point);
		}
		return n;
	}
}


/* geom_bezier_solve_y */

#if 0
int geom_bezier_solve_y(struct geom_bezier *bezier, real min_time, real max_time, real y, uns flags, struct geom_curve_point *result)
{
	int n, d;
	double time;

	FUNC;
	ITEM_ASSERT(bezier, bezier);

	if (min_time > max_time)
		ERET(0);
	d = BEZIER_DEGREE(bezier);
	if (IS_EQUAL(y, bezier->pt[0].y)) {
		if (min_time > 0)
			ERET(0);
		result->time = 0;
		result->point.x = bezier->pt[0].x;
		goto FOUND;
	}
	if (IS_EQUAL(y, bezier->pt[d].y)) {
		if (max_time < 1)
			ERET(0);
		result->time = 1;
		result->point.x = bezier->pt[d].x;
		goto FOUND;
	}
	if (bezier->pt[0].y <= bezier->pt[d].y) {
		if (y < bezier->pt[0].y || y > bezier->pt[d].y)
			ERET(0);
	}
	else {
		if (y > bezier->pt[0].y || y < bezier->pt[d].y)
			ERET(0);
	}

	ETEST_RET(n = geom_bezier_y_intersection_times(bezier, y, GEOM_SOLVE_LEFT_ONLY, &time));
	if (n > 0) {
		RTEST_NUM(result->time = time);
		if (result->time < min_time || result->time > max_time)
			ERET(0);
		if (flags & GEOM_CURVE_POINT_POINT)
			ETEST_RET(geom_bezier_point_at_time(bezier, result->time, &result->point));
		goto FOUND;
	}
	ERET(0);
FOUND:
	result->point.y = y;

	ERET(1);
}
#endif


