#ifndef _VRR_GEOMLIB_CURVE_H_
#define _VRR_GEOMLIB_CURVE_H_

/** @file
 *
 * GEOMLIB - Elementary curves ancestor.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "geomlib/group.h"


/*----------------------------------------------*
 *                                              *
 *                BEZIER CLASS                  *
 *                                              *
 *----------------------------------------------*/


#define geom_curve_VMT    geom_item_VMT
#define geom_curve_INIT   geom_item_INIT  GEOM_CURVE_DEFAULTS
GEOM_CLASS_HEAD(curve, item);


/*----------------------------------------------*
 *                                              *
 *       DEFAULTS FOR CURVES AND PATHS          *
 *                                              *
 *----------------------------------------------*/


#define GEOM_CURVE_DEFAULTS \
	\
	.destroy = (void (*)(void *)) &geom_curve_destroy, \
	.copy = (void (*)(void *, void *)) &geom_curve_copy, \
	.after_change = (void (*)(void *)) &geom_curve_after_change, \
	\
	.tlength = &geom_item_tlength__one, \
	.alength = &geom_item_alength__from_atime, \
	.blength = &geom_item_blength__expand, \
	\
	.time_to_atime = &geom_item_time_to_atime__expand, \
	.times_to_atimes = &geom_item_times_to_atimes__cycle, \
	.atime_to_time = &geom_item_atime_to_time__expand, \
	.atimes_to_times = &geom_item_atimes_to_times__cycle, \
	\
	.time_to_ratime = &geom_item_time_to_ratime__from_atime, \
	.times_to_ratimes = &geom_item_times_to_ratimes__cycle, \
	.ratime_to_time = &geom_item_ratime_to_time__from_atime, \
	.ratimes_to_times = &geom_item_ratimes_to_times__cycle, \
	\
	.time_dif_btime = 1, \
	.time_to_btime = &geom_item_time_to_btime__from_atime, \
	.times_to_btimes = &geom_item_times_to_btimes__cycle, \
	.btime_to_time = &geom_item_btime_to_time__from_atime, \
	.btimes_to_times = &geom_item_btimes_to_times__cycle, \
	\
	.to_bezier = &geom_item_to_bezier__expand, \
	.conversion_append = &geom_item_conversion_append__expand, \
	\
	.point_start = &geom_item_point_start__from_time, \
	.point_end = &geom_item_point_end__from_time, \
	.point_at_time = &geom_item_point_at_time__expand, \
	\
	.derivation_start = &geom_item_derivation_start__from_time, \
	.derivation_end = &geom_item_derivation_end__from_time, \
	.derivation_at_time = &geom_item_derivation_at_time__expand, \
	\
	.bbox = &geom_item_bbox__expand, \
	.nearest_to_point = &geom_item_nearest_to_point__expand, \
	.distance_times = &geom_item_distance_times__expand, \

#define GEOM_CURVE_TIME_EQUALS_BTIME \
	.time_dif_btime = 0, \
	.time_to_btime = &geom_item_time_to_btime__copy, \
	.times_to_btimes = &geom_item_times_to_btimes__copy, \
	.btime_to_time = &geom_item_btime_to_time__copy, \
	.btimes_to_times = &geom_item_btimes_to_times__copy, \


static inline void geom_curve_create(struct geom_curve *curve) { geom_item_create(&curve->item); }

static inline void geom_curve_destroy(struct geom_curve *curve)
{
	geom_item_expansion_release(curve);
	geom_item_destroy(&curve->item);
}


/*----------------------------------------------*
 *                                              *
 *        GENERAL METHODS IMPLEMENTATION        *
 *                                              *
 *----------------------------------------------*/


/* Parametrizations */

static inline real geom_item_tlength__one(void UNUSED *item) { return 1; }
real geom_item_alength__from_atime(void *item);
static inline int geom_item_blength__one(void UNUSED *item) { return 1; }
int geom_item_blength__expand(void *item);

real geom_item_time_to_atime__expand(void *item, real time);
int geom_item_times_to_atimes__cycle(void *item, uns count, real *times, real *atimes);
real geom_item_atime_to_time__expand(void *item, real atime);
int geom_item_atimes_to_times__cycle(void *item, uns count, real *atimes, real *times);

real geom_item_time_to_ratime__from_atime(void *item, real time);
int geom_item_times_to_ratimes__cycle(void *item, uns count, real *times, real *ratimes);
real geom_item_ratime_to_time__from_atime(void *item, real ratime);
int geom_item_ratimes_to_times__cycle(void *item, uns count, real *ratimes, real *times);

static inline real geom_item_time_to_btime__copy(void UNUSED *item, struct geom_expansion UNUSED *expansion, real time) { return time; }
real geom_item_time_to_btime__from_atime(void *item, struct geom_expansion *expansion, real time);
int geom_item_times_to_btimes__copy(void *item, struct geom_expansion *expansion, uns count, real *times, real *btimes);
int geom_item_times_to_btimes__cycle(void *item, struct geom_expansion *expansion, uns count, real *times, real *btimes);
int geom_item_times_to_btimes__from_atimes(void *item, struct geom_expansion *expansion, uns count, real *times, real *btimes);
static inline real geom_item_btime_to_time__copy(void UNUSED *item, struct geom_expansion UNUSED *expansion, real btime) { return btime; }
real geom_item_btime_to_time__from_atime(void *item, struct geom_expansion *expansion, real btime);
int geom_item_btimes_to_times__copy(void *item, struct geom_expansion *expansion, uns count, real *btimes, real *times);
int geom_item_btimes_to_times__cycle(void *item, struct geom_expansion *expansion, uns count, real *btimes, real *times);
int geom_item_btimes_to_times__from_atimes(void *item, struct geom_expansion *expansion, uns count, real *btimes, real *times);


/* Conversions */

int geom_item_to_bezier__expand(void *item, struct geom_bezier *bezier);
int geom_item_conversion_append__expand(void *item, uns flags, struct geom_fpath *result);


/* Geometric properties */

int geom_item_point_start__from_time(void *item, struct geom_point *result);
int geom_item_point_end__from_time(void *item, struct geom_point *result);
int geom_item_point_at_time__expand(void *item, real time, struct geom_point *result);

int geom_item_derivation_start__from_time(void *item, struct geom_vector *result);
int geom_item_derivation_end__from_time(void *item, struct geom_vector *result);
int geom_item_derivation_at_time__expand(void *item, real time, struct geom_vector *result);

int geom_item_bbox__expand(void *item, struct geom_rectangle *bbox);
int geom_item_nearest_to_point__expand(void *item, struct geom_point *point, uns flags, struct geom_nearest *nearest);
int geom_item_direction_times__expand(void *item, struct geom_vector *dir, struct garr *result); // FIXME: geom_bezier_direction_times broken
int geom_item_distance_times__expand(void *item, struct geom_point *point, real dist, struct garr *result);


/*----------------------------------------------*
 *                                              *
 *          ELEMETARY CURVES OVERRIDES          *
 *                                              *
 *----------------------------------------------*/


#define GEOM_PATH_ALENGTH_VALID  0x1

static inline void geom_curve_copy(struct geom_curve *curve, struct geom_curve *src)
{
	geom_item_copy__default(curve, src);
	geom_item_after_change(curve);
}

static inline void geom_path_after_change(struct geom_path *path)
{
	geom_item_expansion_release(path);
	GEOM_INSTANCE_FLAGS(path) &= ~GEOM_PATH_ALENGTH_VALID;
}

static inline void geom_curve_after_change(struct geom_curve *curve)
{
	GEOM_ITEM_FLAGS(curve) &= ~GEOM_ITEM_EXPANSION_VALID;
	//geom_item_expansion_free(curve);*/
	/*if ((path = geom_item_owner(curve)) && (GEOM_ITEM_FLAGS(path) & GEOM_ITEM_PATH))
		geom_path_after_change(path);*/
}


#endif
