#ifndef _VRR_GEOM_RTREE_H_
#define _VRR_GEOM_RTREE_H_

/** @file
 *
 * 	GEOMLIB - Implementation of Rtree planar search data structure.
 *
 *	Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 *	This program is distributed under GNU General Public License, see COPYING.
 */

#include "geomlib/types.h"

#define GEOM_RTREE_QUERY_SHOW             0x00000004
#define GEOM_RTREE_QUERY_HIDE             0x00000008
#define GEOM_RTREE_QUERY_UPDATE           0x00000010
#define GEOM_RTREE_QUERY_UPDATE_BBOX      0x00000020

#define GEOM_RTREE_MAX_HEIGHT 32

/* WALK_QUERY_IN_RTREE */

static inline struct geom_rtree_query *query_first_in_rtree(struct geom_rtree *rtree)
{
	cnode *n = clist_head(&rtree->lquery);
	return n ? SKIP_BACK(struct geom_rtree_query, nquery, n) : NULL;
}

static inline struct geom_rtree_query *query_next_in_rtree(struct geom_rtree *rtree, struct geom_rtree_query *query)
{
	cnode *n = clist_next(&rtree->lquery, &query->nquery);
	return n ? SKIP_BACK(struct geom_rtree_query, nquery, n) : NULL;
}

#define WALK_QUERY_IN_RTREE(_rtree, _query) \
	for (struct geom_rtree_query *_query = query_first_in_rtree(_rtree); \
			_query; _query = query_next_in_rtree(_rtree, _query))


/* WALK_NODE_IN_RTREE_FILTER */

#define WALK_NODE_IN_RTREE_FILTER_BEGIN(_name, _rtree, _node, _filter) \
	{ if ((_rtree)->root) { \
		struct geom_rtree_node *_node, **_name##_stack[GEOM_RTREE_MAX_HEIGHT + 1]; \
		int _name##_top, _name##_cnt[GEOM_RTREE_MAX_HEIGHT + 1]; \
		_name##_cnt[0] = _name##_cnt[_name##_top = 1] = 1; \
		_name##_stack[1] = &(_rtree)->root; \
		do { \
			_node = *(_name##_stack[_name##_top]++); \
			_name##_cnt[_name##_top]--; \
			if (_filter) { \
				if (_node->height) { \
					_name##_stack[++_name##_top] = _node->child; \
					_name##_cnt[_name##_top] = _node->count; \
				}; {

#define WALK_NODE_IN_RTREE_FILTER_END(_name) }} \
			while (!_name##_cnt[_name##_top]) _name##_top--; } \
		while (_name##_top); } \
	_name##_exit:; if (0) goto _name##_exit; }

#define WALK_NODE_IN_RTREE_FILTER_BREAK(_name) do { goto _name##_exit; } while(0)


/* WALK_NODE_IN_RTREE */

#define WALK_NODE_IN_RTREE_BEGIN(_name, _rtree, _node) \
	WALK_NODE_IN_RTREE_FILTER_BEGIN(_name, _rtree, _node, 1)
#define WALK_NODE_IN_RTREE_END(_name) WALK_NODE_IN_RTREE_FILTER_END(_name)
#define WALK_NODE_IN_RTREE_BREAK(_name) WALK_NODE_IN_RTREE_FILTER_BREAK(_name)


/* WALK_OBJ_IN_RTREE_FILTER */

#define WALK_OBJ_IN_RTREE_FILTER_BEGIN(_name, _rtree, _node, _obj, _node_filter, _obj_filter) \
	WALK_NODE_IN_RTREE_FILTER_BEGIN(_name, _rtree, _node, _node_filter) { \
		if (!_node->height) { \
			for (int _name##_i = 0; _name##_i < _node->count; _name##_i++) { \
				struct geom_rtree_obj *_obj = (struct geom_rtree_obj *)_node->child[_name##_i]; \
				if (_obj_filter) {

#define WALK_OBJ_IN_RTREE_FILTER_END(_name) }}}} WALK_NODE_IN_RTREE_FILTER_END(_name)
#define WALK_OBJ_IN_RTREE_FILTER_BREAK(_name) WALK_NODE_IN_RTREE_FILTER_BREAK(_name)


/* WALK_OBJ_IN_RTREE */

#define WALK_OBJ_IN_RTREE_BEGIN(_name, _rtree, _obj) \
	WALK_OBJ_IN_RTREE_FILTER_BEGIN(_name, _rtree, _name##_node, _obj, 1, 1)
#define WALK_OBJ_IN_RTREE_END(_name) WALK_OBJ_IN_RTREE_FILTER_END(_name)
#define WALK_OBJ_IN_RTREE_BREAK(_name) WALK_OBJ_IN_RTREE_FILTER_BREAK(_name)


/* GEOM_RTREE_POINT_QUERY */

#define GEOM_RTREE_POINT_QUERY_BEGIN(_name, _rtree, _point, _obj) \
	WALK_OBJ_IN_RTREE_FILTER_BEGIN(_name, _rtree, _name##_node, _obj, \
		geom_rtree_box_point(&_name##_node->bbox, (_point)), \
		geom_rtree_box_point(&_obj->bbox, (_point)))
#define GEOM_RTREE_POINT_QUERY_END(_name) WALK_OBJ_IN_RTREE_FILTER_END(_name)
#define GEOM_RTREE_POINT_QUERY_BREAK(_name) WALK_OBJ_IN_RTREE_FILTER_BREAK(_name)


/* GEOM_RTREE_RECT_INTERSECT_QUERY */

#define GEOM_RTREE_RECT_INTERSECT_QUERY_BEGIN(_name, _rtree, _rect, _obj) \
	WALK_OBJ_IN_RTREE_FILTER_BEGIN(_name, _rtree, _name##_node, _obj, \
		geom_rtree_box_intersect(&_name##_node->bbox, (_rect)), \
		geom_rtree_box_intersect(&_obj->bbox, (_rect)))
#define GEOM_RTREE_RECT_INTERSECT_QUERY_END(_name) WALK_OBJ_IN_RTREE_FILTER_END(_name)
#define GEOM_RTREE_RECT_INTERSECT_QUERY_BREAK(_name) WALK_OBJ_IN_RTREE_FILTER_BREAK(_name)


/* GEOM_RTREE_RECT_ENCLOSE_QUERY */

#define GEOM_RTREE_RECT_ENCLOSE_QUERY_BEGIN(_name, _rtree, _rect, _obj) \
	WALK_OBJ_IN_RTREE_FILTER_BEGIN(_name, _rtree, _name##_node, _obj, \
		geom_rtree_box_intersect((_rect), &_name##_node->bbox), \
		geom_rtree_box_enclose((_rect), &_obj->bbox))
#define GEOM_RTREE_RECT_ENCLOSE_QUERY_END(_name) WALK_OBJ_IN_RTREE_FILTER_END(_name)
#define GEOM_RTREE_RECT_ENCLOSE_QUERY_BREAK(_name) WALK_OBJ_IN_RTREE_FILTER_BREAK(_name)


/**
	\brief Returns whether two closed rectangles (including edges) have nonempty intersection.

	\param r1 pointer to first rectangle
	\param r2 pointer to second rectangle

	\return nonzero, when rectangles intersect
*/

static inline int geom_rtree_box_intersect(struct geom_rectangle *r1, const struct geom_rectangle *r2)
{
	return (r1->ll.x <= r2->ur.x && r2->ll.x <= r1->ur.x &&
		    r1->ll.y <= r2->ur.y && r2->ll.y <= r1->ur.y);
}


/**
	\brief Returns whether first rectangle contains the second.

	\param r1 pointer to first rectangle
	\param r2 pointer to second rectangle

	\return nonzero, when r1 contains r2
*/

static inline int geom_rtree_box_enclose(struct geom_rectangle *r1, const struct geom_rectangle *r2)
{
	return (r2->ll.x >= r1->ll.x && r2->ur.x <= r1->ur.x &&
			r2->ll.y >= r1->ll.y && r2->ur.y <= r1->ur.y);
}


/**
	\brief Returns whether closed rectangle (including edges) contains given point.

	\param r pointer to rectangle
	\param p pointer to point

	\return nonzero, when r contains p
*/

static inline int geom_rtree_box_point(struct geom_rectangle *r, struct geom_point *p)
{
	return (p->x >= r->ll.x && p->x <= r->ur.x && p->y >= r->ll.y && p->y <= r->ur.y);
}


/**
	\brief Initialize RTree data.

	\param rtree pointer to RTree structure
*/
void geom_rtree_init(struct geom_rtree *rtree);


/**
	\brief Cleans up internal RTree data.

	\param rtree pointer to RTree structure
*/
void geom_rtree_destroy(struct geom_rtree *rtree);


/**
	\brief Remove all objects from giver RTree.

	\param rtree pointer to RTree structure
*/
void geom_rtree_clear(struct geom_rtree *rtree);


/**
	\brief Add new object to Rtree.

	\param rtree pointer to RTree structure
	\param obj pointer to adding object
	\param bbox object's bounding box
*/
void geom_rtree_insert(struct geom_rtree *rtree, struct geom_rtree_obj *obj, struct geom_rectangle *bbox);


/**
	\brief Remove object from RTree.

	\param pointer to rtree structure
	\param obj pointer to removing object
*/
void geom_rtree_delete(struct geom_rtree *rtree, struct geom_rtree_obj *obj);


/**
	\brief Update object in its RTree.

	\param rtree pointer to rtree structure
	\param obj pointer to updating object
	\param flags update flags
	\param bbox object new bounding box (or NULL when not changing)
*/
void geom_rtree_update(struct geom_rtree *rtree, struct geom_rtree_obj *obj, uns flags, struct geom_rectangle *bbox);


/**
	\brief Assertion of whole RTree structure.

	\param rtree pointer to rtree main structure
*/
void geom_rtree_assert(struct geom_rtree *rtree);


/**
	\brief Dump whole RTree structure.

	\param rtree pointer to rtree main structure.
*/
void geom_rtree_dump(struct geom_rtree *rtree);


/**
	\brief Initialize dynamic query structure (rectangular intersection query).

	\param query pointer to initializing query
	\param rtree pointer to rtree to associate
	\param box pointer to query area
	\param show object show callback
	\param hide object hide callback
	\param update object update callback
*/
void geom_rtree_query_init(struct geom_rtree_query *query, struct geom_rtree *rtree, struct geom_rectangle *box, void *show, void *hide, void *update);


/**
	\brief Destroys dynamic query structure.

	\param query pointer to destroying query.
*/
void geom_rtree_query_destroy(struct geom_rtree_query *query);


#endif /* _VRR_GEOM_RTREE_H_ */
