#ifndef _VRR_GEOMLIB_TRANSFORM2_H_
#define _VRR_GEOMLIB_TRANSFORM2_H_

/** @file
 *
 * 	GEOMLIB - Transformations alternatives.
 *
 *	Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 *	This program is distributed under GNU General Public License, see COPYING.
 */



int geom_transform2_is_valid(const struct geom_transform2 *t);
int geom_transform2_set (struct geom_transform2 *t, real a, real b, real c, real d, real e, real f) WARN_UNUSED_RESULT;

int geom_transform2_set_identity (struct geom_transform2 *t);
int geom_transform2_set_move (struct geom_transform2 *t, const real x, const real y);
int geom_transform2_set_rotation (struct geom_transform2 *t, const real angle);
int geom_transform2_set_scale (struct geom_transform2 *t, const real x, const real y) WARN_UNUSED_RESULT;

/**
	\brief Initialize transformation matrix T so (T(0, 0) = (p1x, p1y), T(1, 0) = (p2x, p2y), T(0, 1) = (p3x, p3y)).

	\retval 0 success
	\retval GEOM_ERR_NUMERIC numeric overflow or singular matrix
*/
int geom_transform2_set_basis (struct geom_transform2 *t, real p1x, real p1y, real p2x, real p2y, real p3x, real p3y) WARN_UNUSED_RESULT;
int geom_transform2_invert (const struct geom_transform2 *t, struct geom_transform2 *result);
int geom_transform2_merge (const struct geom_transform2 *t1, const struct geom_transform2 *t2, struct geom_transform2 *result) WARN_UNUSED_RESULT;

// (T <= NT(x, y) * T)

int geom_transform2_translate (struct geom_transform2 *t, const real x, const real y) WARN_UNUSED_RESULT;
int geom_transform2_rotate (struct geom_transform2 *t, const real angle) WARN_UNUSED_RESULT;
int geom_transform2_scale (struct geom_transform2 *t, const real x, const real y) WARN_UNUSED_RESULT;

int geom_transform2_set_rotate_center (struct geom_transform2 *t, const real x, const real y, const real angle) WARN_UNUSED_RESULT;
int geom_transform2_set_scale_center (struct geom_transform2 *t, const real x, const real y, const real sx, const real sy) WARN_UNUSED_RESULT;

int geom_transform2_set_skew (struct geom_transform2 *t,
			      real x1, real y1, real x2, real y2,
			      real xf, real yf, real xt, real yt) WARN_UNUSED_RESULT;

/** Computes the skew affine transformation in which (zx,zy) and (ax,ay) are fixpoints
  and normalized (fx,fy) becomes normalized (tx,ty).

  \retval 0 success
  \retval GEOM_ERR_NUMERIC numeric overflow or singular matrix
*/
int geom_transform2_set_skew_normal (struct geom_transform2 *t, const real zx, const real zy,
	const real ax, const real ay, const real fx, const real fy, const real tx, const real ty) WARN_UNUSED_RESULT;

#endif

