/** @file
 * GUI - Contains the hard-coded initial menu command structure
 * 
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG
#include "lib/lib.h"

#include <gdk/gdkkeysyms.h>
#include "kernel/kernel.h"

#include "gui/cmdmgr.h"
#include "gui/main.h"
#include "gui/properties.h"
#include "build/path.h"

#define CMD static struct cmd

/* File category */
CMD file_exit_cmd = { "Exit", "Exit the program", CT_FUNC, CL_EVERYWHERE | CS_VIS_EN, LIB_PATH"exit.png", NULL, GDK_Q, GDK_CONTROL_MASK,
			{ .func = { VTM_ANYTHING, 0, RT_ANY, NULL, &on_file_exit } }, NULL, NULL };
CMD file_rerun_tex = { "Rerun TeX on all pages", "Rerun TeX on all pages", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_DOCUMENT | VTM_CT_DOC, -1, RT_META, NULL, &on_document_rerun_tex } }, NULL, &file_exit_cmd };
CMD file_sep = { .type = CT_SEPARATOR, .flags = CL_EVERYWHERE, .next = &file_rerun_tex };
CMD import_ipe5_cmd = { "Import Ipe 5", "Import an Ipe 5 file", CT_FUNC, CL_ALL_MENUS | CS_VIS_EN, 0, NULL, 0, 0,
			{ .func = { VTM_ANYTHING, 0, RT_ANY, NULL, &on_import_ipe5 }}, NULL,  &file_sep };
CMD import_svg_cmd = { "Import SVG", "Import a SVG file", CT_FUNC, CL_ALL_MENUS | CS_VIS_EN, 0, NULL, GDK_I, GDK_CONTROL_MASK,
			{ .func = { VTM_ANYTHING, 0, RT_ANY, NULL, &on_svg_file_import }}, NULL,  &import_ipe5_cmd };
CMD file_nn_sep = { .type = CT_SEPARATOR, .flags = CL_EVERYWHERE, .next = &import_svg_cmd };
CMD file_close_cmd = { "Close", "Close the file", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_DOCUMENT | VTM_CT_DOC, 1, RT_META, NULL, &on_file_close } }, NULL, &file_nn_sep };
CMD file_save_all_cmd = { "Save all", "Save all files" , CT_FUNC, CL_ALL_MENUS | CS_VIS_EN, 0, 0, GDK_S, GDK_CONTROL_MASK | GDK_SHIFT_MASK,
			{ .func = { VTM_ANYTHING, 0, RT_META, NULL, &on_file_save_all } }, NULL, &file_close_cmd };
CMD file_save_as_cmd = { "Save as...", "Save the file as..." , CT_FUNC, CL_UB_MENU | CL_VIEW_MENU | CS_VIS_EN, 0, 0, 0, 0,
			{ .func = { VTM_DOCUMENT | VTM_CT_DOC, 1, RT_META, NULL, &on_file_save_as } }, NULL, &file_save_all_cmd };
CMD file_save_cmd = { "Save", "Save the file" , CT_FUNC, CL_UB_MENU | CL_VIEW_MENU | CS_VIS_EN, 0, 0, GDK_S, GDK_CONTROL_MASK,
			{ .func = { VTM_DOCUMENT | VTM_CT_DOC, 1, RT_META, NULL, &on_file_save } }, NULL, &file_save_as_cmd };
CMD file_open_cmd = { "Open", "Open a file", CT_FUNC, CL_EVERYWHERE | CS_VIS_EN, LIB_PATH"open.png", NULL, GDK_O, GDK_CONTROL_MASK,
			{ .func = { VTM_ANYTHING, 0, RT_ANY, NULL, &on_file_open } }, NULL, &file_save_cmd };
CMD file_new_cmd = { "New", "New file", CT_FUNC, CL_EVERYWHERE | CS_VIS_EN, LIB_PATH"new.png", NULL, GDK_N, GDK_CONTROL_MASK,
			{ .func = { VTM_ANYTHING, 0, RT_ANY, NULL, &on_file_new } }, NULL, &file_open_cmd };


/* Import/Export category */
CMD page_export_svg = {"Export page to SVG", "Export page to SVG", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_TLO | VTM_CT_TLO, 1, RT_META, NULL, &on_export_svg }}, NULL, NULL };
CMD page_export_eps = { "Export page to EPS", "Export page to EPS", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_TLO | VTM_CT_TLO, 1, RT_META, NULL, &on_export_eps }}, NULL,  &page_export_svg };
CMD file_export_ps = { "Export document to PS", "Export document to PS", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_DOCUMENT | VTM_CT_DOC, 1, RT_META, NULL, &on_export_ps }}, NULL,  &page_export_eps };
CMD file_export_pdf = { "Export document to PDF", "Export document to PDF", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_DOCUMENT | VTM_CT_DOC, 1, RT_META, NULL, &on_export_pdf }}, NULL,  &file_export_ps };

/* Page category */
CMD page_move_back_cmd = { "Move to back", "Move the page up", CT_FUNC, CL_UB_MENU | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_TLO, 1, RT_META, NULL, &on_page_move_back }}, NULL,  NULL };
CMD page_move_front_cmd = { "Move to front", "Move the page up", CT_FUNC, CL_UB_MENU | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_TLO, 1, RT_META, NULL, &on_page_move_front }}, NULL,  &page_move_back_cmd };
CMD page_move_down_cmd = { "Move down", "Move the page up", CT_FUNC, CL_UB_MENU | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_TLO, 1, RT_META, NULL, &on_page_move_down }}, NULL,  &page_move_front_cmd };
CMD page_move_up_cmd = { "Move up", "Move the page up", CT_FUNC, CL_UB_MENU | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_TLO, 1, RT_META, NULL, &on_page_move_up }}, NULL,  &page_move_down_cmd };
CMD obj_paste_cmd = { "Paste", "Paste", CT_FUNC, CL_UB_MENU | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_DOCUMENT, 1, RT_META, NULL, &on_obj_paste }}, NULL,  &page_move_up_cmd };
CMD obj_cut_cmd = { "Cut", "Cut", CT_FUNC, CL_UB_MENU | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_TLO, -1, RT_META, NULL, &on_obj_cut }}, NULL,  &obj_paste_cmd };
CMD obj_copy_cmd = { "Copy", "Copy", CT_FUNC, CL_UB_MENU | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_TLO, -1, RT_META, NULL, &on_obj_copy }}, NULL,  &obj_cut_cmd };
CMD page_n_sp = { .type = CT_SEPARATOR, .flags = CL_UB_MENU, .next = &obj_copy_cmd };
CMD page_rerun_tex =  { "Rerun TeX", "Rerun TeX", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_TLO | VTM_CT_TLO, -1, RT_META, NULL, &on_page_rerun_tex }}, NULL,  &page_n_sp };
CMD page_new_cmd =  { "New", "New page", CT_FUNC, CL_UB_MENU | CL_VIEW_MENU | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_DOCUMENT | VTM_CT_DOC, -1, RT_META, NULL, &on_file_new_page }}, NULL,  &page_rerun_tex };
CMD new_view_cmd = { "Open in new view", "Open in new view", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_TLO | VTM_CT_TLO, 1, RT_META, NULL, &on_page_new_view }}, NULL,  &page_new_cmd };

/* Edit category */

static void paste_cmd_enabled( struct cmd * cmd UNUSED, GtkWidget * widget UNUSED )
{ COMBINE_FLAGS( (cmd->flags & CS_ENABLED) && context.tlo != tlo_clipboard &&
	group_get_first_selected_go(tlo_get_tlg(tlo_clipboard)), CS_ENABLED ); }

static void cutcopy_enabled( struct cmd * cmd UNUSED, GtkWidget * widget UNUSED )
{ COMBINE_FLAGS( (cmd->flags & CS_ENABLED) && context.tlo != tlo_clipboard, CS_ENABLED ); }

static void redo_enabled( struct cmd * cmd, GtkWidget * widget UNUSED )
{ COMBINE_FLAGS( tlo_universum->undo_present != trans_get_last_gui(tlo_universum), CS_ENABLED ); }

static void undo_enabled( struct cmd * cmd, GtkWidget * widget UNUSED )
{ COMBINE_FLAGS( tlo_universum->undo_present != &tlo_universum->undo_start, CS_ENABLED ); }

static void local_redo_enabled( struct cmd * cmd, GtkWidget * widget )
{
	void * v = g_object_get_data( G_OBJECT(widget), "vrrgroup" );
	if ( v )
	{
		struct obj_tlo * tlo = go_get_tlo(void2go(v));
		COMBINE_FLAGS( tlo->undo_present != trans_get_last_gui(tlo), CS_ENABLED );
	}
	else COMBINE_FLAGS( context.tlo && context.tlo->undo_present != trans_get_last_gui(context.tlo), CS_ENABLED );
}
static void local_undo_enabled( struct cmd * cmd, GtkWidget * widget )
{
	void * v = g_object_get_data( G_OBJECT(widget), "vrrgroup" );
	if ( v )
	{
		struct obj_tlo * tlo = go_get_tlo(void2go(v));
		COMBINE_FLAGS( tlo->undo_present != &tlo->undo_start, CS_ENABLED );
	}
	else COMBINE_FLAGS( context.tlo && context.tlo->undo_present != &context.tlo->undo_start, CS_ENABLED );
}
static void history_jump_enabled( struct cmd * cmd, GtkWidget * widget )
{
	void * v = g_object_get_data( G_OBJECT(widget), "vrrgroup" );
	struct obj_tlo * tlo = NULL;
	if ( v ) tlo = go_get_tlo(void2go(v));
	COMBINE_FLAGS( ( tlo || (tlo = tlo_universum) ) && trans_get_first_gui(tlo), CS_ENABLED ); //I really mean this! :)
}
#define GET_CONTEXT_GO(_go) \
	struct go * _go = NULL; \
	if ( context.o && context.o->kind == T_GO && go_get_group(o2go(context.o)) == context.group ) _go = o2go(context.o); \
	else if (context.group) go = group_get_first_selected_go(context.group)

static void can_move_up( struct cmd * cmd, GtkWidget * widget UNUSED )
{
	GET_CONTEXT_GO(go);
	FILTER_FLAGS( go && group_get_prev_go(context.group, go), CS_ENABLED );
}
static void can_move_down( struct cmd * cmd, GtkWidget * widget UNUSED )
{
	GET_CONTEXT_GO(go);
	FILTER_FLAGS( go && group_get_next_go(context.group, go), CS_ENABLED );
}
static void can_move_front( struct cmd * cmd, GtkWidget * widget UNUSED )
{
	GET_CONTEXT_GO(go);
	FILTER_FLAGS( go && go != group_get_first_go(context.group), CS_ENABLED );
}
static void can_move_back( struct cmd * cmd, GtkWidget * widget UNUSED )
{
	GET_CONTEXT_GO(go);
	FILTER_FLAGS( go && go != group_get_last_go(context.group), CS_ENABLED );
}
/* Basic transforms */
CMD edit_rot270 = { "Rotate by 270 degrees", "Rotate by 270 degrees", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_GO, -1, RT_GROUP, NULL, &on_rotate270 }}, NULL,  NULL };
CMD edit_rot180 = { "Rotate by 180 degrees", "Rotate by 180 degrees", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_GO, -1, RT_GROUP, NULL, &on_rotate180 }}, NULL,  &edit_rot270 };
CMD edit_rot90 = { "Rotate by 90 degrees", "Rotate by 90 degrees", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0, 
			{ .func = { VTM_GO, -1, RT_GROUP, NULL, &on_rotate90 }}, NULL,  &edit_rot180 };
CMD edit_hflip = { "Flip horizontally", "Flip horizontally", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_GO, -1, RT_GROUP, NULL, &on_flip_horizontal }}, NULL,  &edit_rot90 };
CMD edit_vflip = { "Flip vertically", "Flip vertically", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_GO, -1, RT_GROUP, NULL, &on_flip_vertical }}, NULL,  &edit_hflip };

CMD edit_create_proxy = { "Create Scheme proxy", "Create Scheme proxy", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_GO, 1, RT_GROUP, NULL, &on_create_proxy }}, NULL, NULL };
CMD edit_mm_sep = { .type = CT_SEPARATOR, .flags = CL_LOCAL_MENUS, .next = &edit_create_proxy };
//CMD edit_enclose_sel = { "Select dependent", "Add dependent to selection", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0,
//			{ .func = { VTM_CT_GRP, 1, RT_META, NULL, &on_go_enclose_sel }}, NULL,  &edit_mm_sep };
CMD edit_select_none =  { "Clear selection", "Clear selection", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL,
			GDK_A, GDK_SHIFT_MASK | GDK_CONTROL_MASK,
			{ .func = { VTM_CT_GRP, 1, RT_META, NULL, &on_group_select_none }}, NULL,  &edit_mm_sep };
CMD edit_select_all =  { "Select all", "Select all", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, GDK_A, GDK_CONTROL_MASK,
			{ .func = { VTM_CT_GRP, 1, RT_META, NULL, &on_group_select_all }}, NULL,  &edit_select_none };
CMD edit_m_sep = { .type = CT_SEPARATOR, .flags = CL_EVERYWHERE, .next = &edit_select_all };
CMD edit_move_back = { "Move to back", "Move to back", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, GDK_D, GDK_SHIFT_MASK,
			{ .func = { VTM_GO, 1, RT_GROUP, &can_move_back, &on_move_back }}, NULL,  &edit_m_sep };
CMD edit_move_front = { "Move to front", "Move to front", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, GDK_U, GDK_SHIFT_MASK,
			{ .func = { VTM_GO, 1, RT_GROUP, &can_move_front, &on_move_front }}, NULL,  &edit_move_back };
CMD edit_move_down = { "Move down", "Move down", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, GDK_D, 0,
			{ .func = { VTM_GO, 1, RT_GROUP, &can_move_down, &on_move_down }}, NULL,  &edit_move_front };
CMD edit_move_up = { "Move up", "Move up", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, GDK_U, 0,
			{ .func = { VTM_GO, 1, RT_GROUP, &can_move_up, &on_move_up }}, NULL,  &edit_move_down };
CMD edit_ungroup = { "Ungroup", "Ungroup the group", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, GDK_G, GDK_SHIFT_MASK | GDK_CONTROL_MASK,
			{ .func = { VTM_GROUP, 1, RT_GROUP, NULL, &on_group_ungroup }}, NULL,  NULL };
CMD edit_group = { "Group selected", "Group the selected objects", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, GDK_G, GDK_CONTROL_MASK,
			{ .func = { VTM_GO, -1, RT_GROUP, NULL, &on_group_selected }}, NULL,  &edit_ungroup };
CMD group_new_view = { "Open in new view", "Open in new view", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_GROUP, 1, RT_GROUP, NULL, &on_group_new_view }}, NULL,  &edit_group };
CMD edit_lr_sep = { .type = CT_SEPARATOR, .flags = CL_EVERYWHERE, .next = &edit_move_up };
CMD edit_redo_cmd =  { "Global redo", "Redo one action", CT_FUNC, CL_MAIN_MENU | CL_UB_MENU | CL_WUH_TBAR | CS_VIS_EN,
			LIB_PATH"redo.png", NULL, 0, 0,
			{ .func = { VTM_ANYTHING, 0, RT_META, &redo_enabled, &on_obj_redo }}, NULL,  &edit_lr_sep };
CMD edit_hist_jump =  { "Global jump", "Jump to selected action", CT_FUNC, CL_WUH_TBAR | CS_VIS_EN, LIB_PATH"jump.png", NULL, 0, 0,
			{ .func = { VTM_UNDO_HISTORY, 1, RT_META, &history_jump_enabled, &on_hist_jump }}, NULL,  &edit_redo_cmd };
CMD edit_undo_cmd =  { "Global undo", "Undo one action", CT_FUNC, CL_MAIN_MENU | CL_UB_MENU | CL_WUH_TBAR | CS_VIS_EN,
			LIB_PATH"undo.png", NULL, 0, 0,
			{ .func = { VTM_ANYTHING, 0, RT_META, &undo_enabled, &on_obj_undo }}, NULL,  &edit_hist_jump };
CMD edit_local_redo_cmd =  { "Redo", "Redo one action", CT_FUNC, CL_LOCAL_MENUS | CL_LWUH_TBAR | CS_VIS_EN,
			LIB_PATH"redo.png", NULL, GDK_R, GDK_CONTROL_MASK,
			{ .func = { VTM_CT_TLO, 1, RT_META, &local_redo_enabled, &on_tlo_redo }}, NULL,  &edit_undo_cmd };
CMD edit_local_hist_jump =  { "Jump", "Jump to selected action", CT_FUNC, CL_LWUH_TBAR | CS_VIS_EN, LIB_PATH"jump.png", NULL,
			GDK_J, GDK_CONTROL_MASK,
			{ .func = { VTM_UNDO_HISTORY, 1, RT_META, &history_jump_enabled, &on_hist_jump }}, NULL,  &edit_local_redo_cmd };
CMD edit_local_undo_cmd =  { "Undo", "Undo one action", CT_FUNC, CL_LOCAL_MENUS | CL_LWUH_TBAR | CS_VIS_EN,
			LIB_PATH"undo.png", NULL, GDK_Z, GDK_CONTROL_MASK,
			{ .func = { VTM_CT_TLO, 1, RT_META, &local_undo_enabled, &on_tlo_undo }}, NULL,  &edit_local_hist_jump };
CMD edit_t_sep = { .type = CT_SEPARATOR, .flags = CL_EVERYWHERE, .next = &edit_local_undo_cmd };
CMD edit_text_cmd =  { "Text", "Edit text", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, GDK_T, GDK_CONTROL_MASK,
			{ .func = { VTM_TEXT | VTM_TEX_TEXT, 1, RT_GROUP, NULL, &on_edit_text }}, NULL,  &edit_t_sep };
CMD edit_prop_cmd =  { "Properties", "Properties", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN,
			LIB_PATH"obj_properties.png", NULL, GDK_P, GDK_CONTROL_MASK,
			{ .func = { VTM_ANYTHING, -1, RT_ANY, NULL, &on_change_prop }}, NULL, &edit_text_cmd };
CMD edit_mb_sep = { .type = CT_SEPARATOR, .flags = CL_EVERYWHERE, .next = &edit_prop_cmd };
CMD transform_ctg = { "Transform", "", CT_CATEGORY, CL_VIEW_MENU, NULL, NULL, 0, 0, { .ctg = { 0, &edit_vflip, NULL }}, NULL, &edit_mb_sep };
CMD edit_delete_cmd = { "Delete", "Delete", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, GDK_Delete, 0,
			{ .func = { VTM_ANYTHING, -1, RT_ANY, NULL, &on_delete }}, NULL,  &transform_ctg };
CMD edit_paste_cmd =  { "Paste", "Paste clipboard contents", CT_FUNC, CL_LOCAL_MENUS | CC_STOCK | CS_VIS_EN, "gtk-paste", NULL,
			GDK_V, GDK_CONTROL_MASK,
			{ .func = { VTM_CT_TLO, 1, RT_META, &paste_cmd_enabled, &on_paste }}, NULL,  &edit_delete_cmd };
CMD edit_cut_cmd =  { "Cut", "Cut selection", CT_FUNC, CL_LOCAL_MENUS | CC_STOCK | CS_VIS_EN, "gtk-cut", NULL, GDK_X, GDK_CONTROL_MASK,
			{ .func = { VTM_GO, -1, RT_GROUP, &cutcopy_enabled, &on_cut }}, NULL,  &edit_paste_cmd };
CMD edit_copy_cmd =  { "Copy", "Copy selection", CT_FUNC, CL_LOCAL_MENUS | CC_STOCK | CS_VIS_EN, "gtk-copy", NULL, GDK_C, GDK_CONTROL_MASK,
			{ .func = { VTM_GO, -1, RT_GROUP, &cutcopy_enabled, &on_copy }}, NULL,  &edit_cut_cmd };

/* Snap category */
static void snap_depend_active( struct cmd * cmd, GtkWidget * widget UNUSED )
{ COMBINE_FLAGS( factory.snap_set & SNAP_DEPENDENCY, CS_CHECKED ); }

static void snap_inters_active( struct cmd * cmd, GtkWidget * widget UNUSED )
{ COMBINE_FLAGS( factory.snap_set & SNAP_INTERS, CS_CHECKED ); }

static void snap_lines_active( struct cmd * cmd, GtkWidget * widget UNUSED )
{ COMBINE_FLAGS( factory.snap_set & SNAP_LINES, CS_CHECKED ); }

static void snap_grid_active( struct cmd * cmd, GtkWidget * widget UNUSED )
{ COMBINE_FLAGS( factory.snap_set & SNAP_GRID, CS_CHECKED ); }

static void snap_hangers_active( struct cmd * cmd, GtkWidget * widget UNUSED )
{ COMBINE_FLAGS( factory.snap_set & SNAP_HANGERS, CS_CHECKED ); }

CMD snap_depend = { "Snap dependencies", "Snap dependencies on/off", CT_FUNC, CL_LOCAL_MENUS | CL_VIEW_TBAR | CC_CHECKABLE |
			CC_SAFE | CS_VIS_EN, LIB_PATH"snap_dependencies.png", NULL, GDK_Y, GDK_SHIFT_MASK,
			{ .func = { VTM_ANYTHING, 0, RT_META, &snap_depend_active, &on_snap_depend }}, NULL,  NULL };
CMD snap_inters_cmd = { "Snap to intersections", "Snap to intersections on/off", CT_FUNC, CL_LOCAL_MENUS | CL_VIEW_TBAR | CC_CHECKABLE |
			CC_SAFE | CS_VIS_EN, LIB_PATH"snap_intersections.png", NULL, GDK_I, GDK_SHIFT_MASK,
			{ .func = { VTM_ANYTHING, 0, RT_META, &snap_inters_active, &on_snap_inters }}, NULL,  &snap_depend };
CMD snap_lines_cmd = { "Snap to lines", "Snap to lines on/off", CT_FUNC, CL_LOCAL_MENUS | CL_VIEW_TBAR | CC_CHECKABLE | CC_SAFE | CS_VIS_EN,
			LIB_PATH"snap_lines.png", NULL, GDK_L, GDK_SHIFT_MASK,
			{ .func = { VTM_ANYTHING, 0, RT_META, &snap_lines_active, &on_snap_lines }}, NULL,  &snap_inters_cmd };
CMD snap_grid_cmd = { "Snap to grid", "Snap to grid on/off", CT_FUNC, CL_LOCAL_MENUS |  CL_VIEW_TBAR | CC_CHECKABLE | CC_SAFE | CS_VIS_EN,
			LIB_PATH"snap_grid.png", NULL, GDK_G, GDK_SHIFT_MASK,
			{ .func = { VTM_ANYTHING, 0, RT_META, &snap_grid_active, &on_snap_grid }}, NULL,  &snap_lines_cmd };
CMD snap_hangers_cmd = { "Snap to hangers", "Snap to hangers on/off", CT_FUNC, CL_LOCAL_MENUS |  CL_VIEW_TBAR | CC_CHECKABLE | CC_SAFE | CS_VIS_EN,
			LIB_PATH"snap_hangers.png", NULL, GDK_H, GDK_SHIFT_MASK,
			{ .func = { VTM_ANYTHING, 0, RT_META, &snap_hangers_active, &on_snap_hangers }}, NULL,  &snap_grid_cmd };

static void path_mode_active( struct cmd * cmd, GtkWidget * widget UNUSED )
{ COMBINE_FLAGS( factory.flags & OFF_PATH_ON, CS_CHECKED ); }

static void path_end_enabled( struct cmd * cmd, GtkWidget * widget UNUSED )
{ COMBINE_FLAGS( factory.flags & OFF_PATH_ACTIVE, CS_ENABLED ); }

/* Path category */
#ifdef GUI_DEBUG
CMD edit_pathize = { "Pathize selected", "Pathize selected", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_GO, -1, RT_GROUP, NULL, &on_group_pathize_selected }}, NULL,  NULL };
CMD path_sep = { .type = CT_SEPARATOR, .flags = CL_EVERYWHERE, .next = &edit_pathize };
#else
CMD path_sep = { .type = CT_SEPARATOR, .flags = 0, .next = NULL };
#endif
CMD path_end_close = { "Close + end path", "Close + end path", CT_FUNC, CL_LOCAL_MENUS | CL_VIEW_TBAR | CC_SAFE | CS_VIS_EN,
			LIB_PATH"path_end_close.png", NULL, GDK_colon, 0,
			{ .func = { VTM_ANYTHING, 0, RT_META, &path_end_enabled, &on_path_end_close }}, NULL,  &path_sep };
CMD path_end = { "End path", "End path", CT_FUNC, CL_LOCAL_MENUS | CL_VIEW_TBAR | CC_SAFE | CS_VIS_EN,
			LIB_PATH"path_end.png", NULL, GDK_semicolon, 0,
			{ .func = { VTM_ANYTHING, 0, RT_META, &path_end_enabled, &on_path_end }}, NULL,  &path_end_close };
CMD path_switch = { "Path mode", "Path mode on/off", CT_FUNC, CL_LOCAL_MENUS | CL_VIEW_TBAR |
			CC_CHECKABLE | CS_VIS_EN, LIB_PATH"path_on.png", NULL, GDK_period, 0,
			{ .func = { VTM_ANYTHING, 0, RT_META, &path_mode_active, &on_path_switch }}, NULL,  &path_end };
/* New category */

static void can_step_back( struct cmd * cmd UNUSED, GtkWidget * widget UNUSED )
{ COMBINE_FLAGS( context.tlo == factory.tlo && factory.state != factory.start_state, CS_ENABLED ); }

#ifdef GUI_DEBUG
CMD new_random_bezier_cubic = { "New random cubic bezier", "New random cubic bezier", CT_FUNC,
				CL_LOCAL_MENUS | CS_VIS_EN, LIB_PATH"bezier_cubic.png", NULL, 0, 0,
				{ .func = { VTM_CT_GRP, 1, RT_GROUP, NULL, &on_random_bezier_cubic }}, NULL, NULL };
CMD new_random_bezier_quad = { "New random quadratic bezier", "New random quadratic bezier", CT_FUNC,
				CL_LOCAL_MENUS | CS_VIS_EN, LIB_PATH"bezier_quad.png", NULL, 0, 0,
				{ .func = { VTM_CT_GRP, 1, RT_GROUP, NULL, &on_random_bezier_quad }}, NULL, &new_random_bezier_cubic };
CMD new_random_segment = { "New random segment", "New random segment", CT_FUNC,
			CL_LOCAL_MENUS | CS_VIS_EN, LIB_PATH"segment.png", NULL, 0, 0,
			{ .func = { VTM_CT_GRP, 1, RT_GROUP, NULL, &on_random_segment }}, NULL, &new_random_bezier_quad };
CMD new_random_point = { "New random point", "New random point", CT_FUNC,
			CL_LOCAL_MENUS | CS_VIS_EN, LIB_PATH"point.png", NULL, 0, 0,
			{ .func = { VTM_CT_GRP, 1, RT_GROUP, NULL, &on_random_point }}, NULL, &new_random_segment };
CMD nr_sep = { .type = CT_SEPARATOR, .flags = CL_EVERYWHERE, .next = &new_random_point };
CMD op_break_cmd = { "Break creation", "Break creation", CT_FUNC, CL_VIEW_MENU | CC_SAFE | CS_VIS_EN, NULL, NULL, GDK_Escape, 0,
			{ .func = { VTM_ANYTHING, 0, RT_ANY, NULL, &on_break_op }}, NULL, &nr_sep };
#else
CMD op_break_cmd = { "Break creation", "Break creation", CT_FUNC, CL_VIEW_MENU | CC_SAFE | CS_VIS_EN, NULL, NULL, GDK_Escape, 0,
			{ .func = { VTM_ANYTHING, 0, RT_ANY, NULL, &on_break_op }}, NULL, NULL };
#endif
CMD step_back_cmd = { "Step back", "Step back", CT_FUNC, CL_VIEW_MENU | CC_SAFE | CS_VIS_EN, NULL, NULL, GDK_BackSpace, 0,
			{ .func = { VTM_CT_TLO | VTM_TLO, 1, RT_META, &can_step_back, &on_step_back }}, NULL, &op_break_cmd };
CMD step_back_sep = { .type = CT_SEPARATOR, .flags = CL_EVERYWHERE, .next = &step_back_cmd };

CMD new_text = { "New text", "New text", CT_FACTORY_OP,
		CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"text.png", NULL, GDK_T, 0,
		{ .op = { &ofs_text_1 }}, NULL, NULL };
CMD new_tex_text = { "New TeX text", "New TeX text", CT_FACTORY_OP,
		CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"tex.png", NULL, GDK_X, 0,
		{ .op = { &ofs_tex_text_1 }}, NULL, &new_text };
CMD new_ellarc_center_radii = { "New ellarc center, radii", "New elliptic arc by center, radii", CT_FACTORY_OP,
				CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"ellarc_center_radii.png", NULL, GDK_Y, 0,
				{ .op = { &ofs_ellarc_center_radii_1 }}, NULL, NULL };
CMD new_ellipse_center_radii = { "New ellipse center, radii", "New ellipse by center, radii", CT_FACTORY_OP,
				CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"ellipse_center_radii.png", NULL, GDK_J, 0,
				{ .op = { &ofs_ellipse_center_radii_1 }}, NULL, &new_ellarc_center_radii };
CMD new_ellipse_xy1ecc = { "New ellipse center, point, ecc", "New ellipse by center, point, eccentricity", CT_FACTORY_OP,
			CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"ellipse_center_pt.png", NULL, GDK_N, 0,
			{ .op = { &ofs_ellipse_xy1ecc_1 }}, NULL, &new_ellipse_center_radii };
CMD new_ellipse_foci_pt = { "New ellipse foci, point", "New ellipse by foci, point", CT_FACTORY_OP,
			CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"ellipse_foci_pt.png", NULL, GDK_V, 0,
			{ .op = { &ofs_ellipse_foci_pt_1 }}, NULL, &new_ellipse_xy1ecc };
CMD new_ellipse_3pecc = { "New ellipse 3 points, rot, ecc", "New ellipse by 3 points, rotation, eccentricity", CT_FACTORY_OP,
			CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"ellipse_3points.png", NULL, GDK_E, 0,
			{ .op = { &ofs_ellipse_3p_ecc_1 }}, NULL, &new_ellipse_foci_pt };
CMD new_ellipse_3p = { "New ellipse 3 points", "New smallest ellipse by 3 points", CT_FACTORY_OP,
			CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"smallest_ellipse_3points.png", NULL, GDK_3, 0, 
			{ .op = { &ofs_ellipse_3p_smallest_1 }}, NULL, &new_ellipse_3pecc };
CMD new_circ_center_rad = { "New circle center, radius", "New circle by center, radius", CT_FACTORY_OP,
			CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"circle_center_radius.png", NULL, GDK_0, 0,
			{ .op = { &ofs_circ_center_radius_1 }}, NULL, NULL };
CMD new_circ_center_pt = { "New circle center, pt", "New circle by center, point", CT_FACTORY_OP,
			CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"circle_center_pt.png", NULL, GDK_C, 0,
			{ .op = { &ofs_circ_center_pt_1 }}, NULL, &new_circ_center_rad };
CMD new_circ_3p = { "New circle 3 points", "New circle by 3 points", CT_FACTORY_OP,
			CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"circle_3points.png", NULL, GDK_B, 0,
			{ .op = { &ofs_circ_3p_1 }}, NULL, &new_circ_center_pt };
CMD new_arc_center_rad = { "New arc center, radius", "New arc by center, radius", CT_FACTORY_OP,
			CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"arc_center_radius.png", NULL, GDK_Z, 0,
			{ .op = { &ofs_arc_center_radius_1 }}, NULL, &new_circ_3p };
CMD new_arc_3p = { "New arc 3 points", "New arc by 3 points", CT_FACTORY_OP,
			CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"arc_3points.png", NULL, GDK_A, 0,
			{ .op = { &ofs_arc_3p_1 }}, NULL, &new_arc_center_rad };
CMD new_bez_cub = { "New cubic bezier", "New cubic bezier", CT_FACTORY_OP,
			CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"bezier_cubic.png", NULL, GDK_K, 0,
			{ .op = { &ofs_bez_cub_1 }}, NULL, NULL };
CMD new_bez_quad = { "New quadratic bezier", "New quadratic bezier", CT_FACTORY_OP,
			CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"bezier_quad.png", NULL, GDK_Q, 0,
			{ .op = { &ofs_bez_quad_1 }}, NULL, &new_bez_cub };
CMD new_segment = { "New segment", "New segment", CT_FACTORY_OP,
			CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"segment.png", NULL, GDK_L, 0,
			{ .op = { &ofs_segment_1 }}, NULL, &new_bez_quad };

CMD new_ngon = { "New n-gon", "New n-gon", CT_FACTORY_OP,
		CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"ngon.png", NULL, GDK_G, 0,
		{ .op = { &ofs_ngon_1 }}, NULL, NULL };
CMD newpt_inters = { "New intersection point", "New intersection point", CT_FACTORY_OP,
		CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"intersection.png", NULL, GDK_I, 0,
		{ .op = { &ofs_point_inters }}, NULL, &new_ngon };
CMD newarrow_cmd = { "New arrow", "New arrow", CT_FACTORY_OP,
		CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"arrow.png", NULL, GDK_W, 0,
		{ .op = { &ofs_arrow }}, NULL, &newpt_inters };
CMD newdecpt_cmd = { "New decoration point", "New decoration point", CT_FACTORY_OP,
		CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"decoration.png", NULL, GDK_O, 0,
		{ .op = { &ofs_decorator_pt }}, NULL, &newarrow_cmd };
CMD newpt_cmd = { "New point", "New point", CT_FACTORY_OP,
		CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"point.png", NULL, GDK_P, 0,
		{ .op = { &ofs_point }}, NULL, &newdecpt_cmd };
/* Mode category */
CMD new_point_ctg;
CMD edit_mode_sep = { .type = CT_SEPARATOR, .flags = CL_LOCAL_MENUS, .next = &new_point_ctg };
CMD mode_rehang = { "Rehang", "Rehang anchors", CT_FACTORY_OP,
		CL_VIEW_TBAR | CL_VIEW_MENU | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"rehang_mode.png", NULL, GDK_H, 0,
		{ .op = { &ofs_rehang_1 }}, NULL, &edit_mode_sep };

CMD mode_transform = { "Transform", "Santiago's transform tool", CT_FACTORY_OP,
		CL_VIEW_TBAR | CL_VIEW_MENU | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"santiago_tm.png", NULL, GDK_M, 0,
		{ .op = { &ofs_transform_1 }}, NULL, &mode_rehang };

CMD mode_select = { "Select", "Select/Transform", CT_FACTORY_OP,
		CL_VIEW_TBAR | CL_VIEW_MENU | CC_CHECKABLE | CS_VIS_EN, LIB_PATH"select_mode.png", NULL, GDK_S, 0,
		{ .op = { &ofs_select }}, NULL, &mode_transform };

/* View category */
static void property_window_visible (struct cmd *cmd, GtkWidget *widget UNUSED)
{ COMBINE_FLAGS( context_property_window, CS_CHECKED); }

static void console_window_visible( struct cmd * cmd, GtkWidget * widget UNUSED )
{ COMBINE_FLAGS( console_window && console_window->visible, CS_CHECKED ); }

static void univ_browser_visible( struct cmd * cmd, GtkWidget * widget UNUSED )
{ COMBINE_FLAGS( wnd_univ_browser.window && wnd_univ_browser.visible, CS_CHECKED ); }

static void plugin_window_visible( struct cmd * cmd, GtkWidget * widget UNUSED )
{ COMBINE_FLAGS( wnd_plugins.visible, CS_CHECKED ); }

static void settings_window_visible( struct cmd * cmd, GtkWidget * widget UNUSED )
{ COMBINE_FLAGS( wnd_settings.window, CS_CHECKED ); }

static void style_window_visible( struct cmd * cmd, GtkWidget * widget UNUSED )
{ COMBINE_FLAGS( wnd_style.window, CS_CHECKED ); }

static void unit_window_visible( struct cmd * cmd, GtkWidget * widget UNUSED )
{ COMBINE_FLAGS( wnd_units.visible, CS_CHECKED ); }

CMD view_zoom_reset = { "Reset zoom and rotation", "Reset zoom and rotation", CT_FUNC, CL_VIEW_MENU |  CS_VIS_EN,
			NULL, NULL, GDK_R, 0, { .func = { VTM_VIEW, 1, RT_META, NULL, &on_view_zoom_reset }}, NULL,  NULL };
CMD view_out_kp = { "Zoom out", "Zoom out", CT_FUNC, CL_VIEW_MENU | CS_ENABLED, NULL, NULL, GDK_KP_Subtract, 0,
			{ .func = { VTM_VIEW, 1, RT_META, NULL, &on_view_zoom_out }}, NULL,  &view_zoom_reset };
CMD view_in_kp = { "Zoom in", "Zoom in", CT_FUNC, CL_VIEW_MENU |  CS_ENABLED, NULL, NULL, GDK_KP_Add, 0,
			{ .func = { VTM_VIEW, 1, RT_META, NULL, &on_view_zoom_in }}, NULL,  &view_out_kp };
CMD view_out = { "Zoom out", "Zoom out", CT_FUNC, CL_VIEW_MENU | CS_VIS_EN, NULL, NULL, GDK_minus, 0,
			{ .func = { VTM_VIEW, 1, RT_META, NULL, &on_view_zoom_out }}, NULL,  &view_in_kp };
CMD view_in = { "Zoom in", "Zoom in", CT_FUNC, CL_VIEW_MENU | CS_VIS_EN, NULL, NULL, GDK_equal, 0,
			{ .func = { VTM_VIEW, 1, RT_META, NULL, &on_view_zoom_in }}, NULL,  &view_out };
CMD view_fit_all = { "Fit all in view", "Fit all in view", CT_FUNC, CL_VIEW_MENU | CS_VIS_EN, NULL, NULL, GDK_KP_Multiply, 0,
			{ .func = { VTM_VIEW, 1, RT_META, NULL, &on_view_fit_all }}, NULL,  &view_in };
CMD view_fit_selected = { "Fit selected in view", "Fit selected in view", CT_FUNC, CL_VIEW_MENU | CS_VIS_EN, NULL, NULL, GDK_F, 0,
			{ .func = { VTM_VIEW, 1, RT_META, NULL, &on_view_fit_selected }}, NULL,  &view_fit_all };

/* Windows category */
CMD view_settings = { "Global settings", "Global settings", CT_FUNC, CC_CHECKABLE | CL_ALL_MENUS | CL_MAIN_TBAR | CS_VIS_EN,
			LIB_PATH"settings.png", NULL, 0, 0,
			{ .func = { VTM_ANYTHING, 0, RT_META, &settings_window_visible, &on_show_settings_window }}, NULL,  NULL };
CMD view_units_cmd = { "Unit manager", "Unit manager", CT_FUNC, CL_ALL_MENUS | CL_MAIN_TBAR | CC_CHECKABLE | CS_VIS_EN,
			LIB_PATH"units.png", NULL, 0, 0,
			{ .func = { VTM_ANYTHING, 0, RT_META, &unit_window_visible, &on_show_unit_window }}, NULL,  &view_settings };
CMD view_plugins_cmd = { "Plugin manager", "Plugin manager", CT_FUNC, CL_ALL_MENUS | CL_MAIN_TBAR | CC_CHECKABLE | CS_VIS_EN,
			LIB_PATH"plugin.png", NULL, 0, 0,
			{ .func = { VTM_ANYTHING, 0, RT_META, &plugin_window_visible, &on_show_plugin_window }}, NULL,  &view_units_cmd };
CMD view_clipboard = {"Clipboard", "View clipboard", CT_FUNC, CL_ALL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_ANYTHING, 0, RT_META, NULL, &on_view_clipboard }}, NULL,  &view_plugins_cmd };
CMD view_local_uh = { "Local undo history", "Local undo history", CT_FUNC, CL_LOCAL_MENUS | CS_VIS_EN,
			LIB_PATH"history.png", NULL, 0, 0,
			{ .func = { VTM_TLO | VTM_CT_TLO, 1, RT_META, NULL, &on_undo_history }}, NULL,  &view_clipboard };
CMD view_uh_cmd = { "Global undo history", "Global undo history", CT_FUNC, CL_ALL_MENUS | CL_MAIN_TBAR | CS_VIS_EN,
			LIB_PATH"history.png", NULL, 0, 0,
			{ .func = { VTM_ANYTHING, 0, RT_META, NULL, &on_show_undo_history }}, NULL,  &view_local_uh };
CMD view_ub_cmd = { "Universe browser", "Universe browser", CT_FUNC, CL_ALL_MENUS | CL_MAIN_TBAR | CC_CHECKABLE | CS_VIS_EN,
			LIB_PATH"browser.png", NULL, GDK_U, GDK_CONTROL_MASK,
			{ .func = { VTM_ANYTHING, 0, RT_META, &univ_browser_visible, &on_show_universe_browser }}, NULL,  &view_uh_cmd };
CMD view_style = { "Style palette", "Style palette", CT_FUNC, CC_CHECKABLE | CL_ALL_MENUS | CL_MAIN_TBAR | CS_VIS_EN,
			LIB_PATH"color.png", NULL, 0, 0,
			{ .func = { VTM_ANYTHING, 0, RT_META, &style_window_visible, &on_show_style_window }}, NULL,  &view_ub_cmd };
CMD view_property_cmd = { "Property window", "Property window", CT_FUNC, CL_ALL_MENUS | CC_CHECKABLE | CS_VIS_EN, 0, NULL,
			GDK_P, GDK_SHIFT_MASK | GDK_CONTROL_MASK,
			{ .func = { VTM_ANYTHING, 0, RT_META, &property_window_visible, &on_show_property_window }}, NULL,  &view_style };
CMD view_console_cmd = { "Console", "Scheme console", CT_FUNC, CL_ALL_MENUS | CL_MAIN_TBAR | CC_CHECKABLE | CS_VIS_EN,
			LIB_PATH"console.png", NULL, GDK_C, GDK_SHIFT_MASK,
			{ .func = { VTM_ANYTHING, 0, RT_META, &console_window_visible, &on_show_console_window }}, NULL,  &view_property_cmd };

/* Debug category */

CMD plugin_menu_dump_ = { "Dump menu plugins", "Dump menu plugins", CT_FUNC, CL_MAIN_MENU | CS_VIS_EN, NULL, NULL, GDK_F3, 0,
			{ .func = { VTM_ANYTHING, 0, RT_META, NULL, &plugin_menu_dump }}, NULL,  NULL };
CMD debug_context = { "Dump context", "Dump context", CT_FUNC, CL_MAIN_MENU | CS_VIS_EN, NULL, NULL, GDK_F2, 0,
			{ .func = { VTM_ANYTHING, 0, RT_META, NULL, &on_context_dump }}, NULL,  &plugin_menu_dump_ };
CMD debug_meta_op2 = { "Meta operation on two TLOs", "", CT_FUNC, CL_UB_MENU | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_TLO, 2, RT_META, NULL, NULL }}, NULL,  &debug_context };
CMD debug_meta_op = { "Meta operation on one TLO", "", CT_FUNC, CL_UB_MENU | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_TLO, 1, RT_META, NULL, NULL }}, NULL,  &debug_meta_op2 };
CMD debug_kerntest = { "Kerntest", "Kernel test", CT_FUNC, CL_ALL_MENUS | CS_VIS_EN, 0, NULL, 0, 0,
			{ .func = { VTM_ANYTHING, 0, RT_META, NULL, &on_kerntest }}, NULL,  &debug_meta_op };
CMD debug_dump = { "Kernel dump", "Kernel dump", CT_FUNC, CL_EVERYWHERE | CS_VIS_EN, LIB_PATH"debug.png", NULL, 0, 0,
			{ .func = { VTM_ANYTHING, 0, RT_META, NULL, &on_dump }}, NULL,  &debug_kerntest };

/* Help category */
CMD about_cmd = {"About ...", "About ...", CT_FUNC, CL_ALL_MENUS | CS_VIS_EN, NULL, NULL, 0, 0,
			{ .func = { VTM_ANYTHING, 0, RT_META, NULL, &on_show_about_window }}, NULL,  NULL };
CMD vrr_online = {"VRR online", "VRR online", CT_FUNC, CL_ALL_MENUS | CS_VIS_EN, LIB_PATH"help.png", 0, 0, 0,
			{ .func = { VTM_ANYTHING, 0, RT_META, NULL, &on_open_www }}, NULL,  &about_cmd };
CMD print_help = {"Help", "Show help", CT_FUNC, CL_EVERYWHERE | CS_VIS_EN, LIB_PATH"help.png", 0, GDK_F1, 0,
			{ .func = { VTM_ANYTHING, 0, RT_META, NULL, &on_help }}, NULL,  &vrr_online };

/* Category initialization */
#ifndef GUI_DEBUG
CMD debug_ctg = { "Debug", "", CT_CATEGORY, 0, NULL, NULL, 0, 0, { .ctg = { 0, &debug_dump, NULL }}, NULL, NULL };
#else
CMD debug_ctg = { "Debug", "", CT_CATEGORY, CL_MAIN_MENU | CL_MAIN_TBAR | CL_UB_MENU, NULL, NULL, 0, 0,
		{ .ctg = { 0, &debug_dump, NULL }}, NULL, NULL };
#endif
CMD help_ctg = { "Help", "", CT_CATEGORY, CL_ALL_MENUS | CL_MAIN_TBAR, NULL, NULL, 0, 0,
		{ .ctg = { 0, &print_help, NULL }}, NULL, &debug_ctg };
CMD windows_ctg = { "Windows", "", CT_CATEGORY, CL_ALL_MENUS | CL_MAIN_TBAR, NULL, NULL, 0, 0,
		{ .ctg = { 0, &view_console_cmd, NULL }}, NULL, &help_ctg };
struct cmd plugin_ctg = { "Plugins", "Plugin functions", CT_CATEGORY, CL_VIEW | CS_VIS_EN | CC_CHECKABLE, LIB_PATH"plugin.png", NULL, 0, 0,
			{ .ctg = { CL_VIEW_TBAR, NULL, NULL }}, NULL, &windows_ctg };
CMD export_ctg = { "Export", "", CT_CATEGORY, CL_UB_MENU | CL_VIEW_MENU, NULL, NULL, 0, 0,
		{ .ctg = { 0, &file_export_pdf, NULL }}, NULL, &plugin_ctg };
CMD file_ctg = { "File", "", CT_CATEGORY, CL_ALL_MENUS | CL_MAIN_TBAR, NULL, NULL, 0, 0,
		{ .ctg = { 0, &file_new_cmd, NULL }}, NULL, &export_ctg };
CMD page_ctg = { "Page", "", CT_CATEGORY, CL_UB_MENU | CL_VIEW_MENU, NULL, NULL, 0, 0, { .ctg = { 0, &new_view_cmd, NULL }}, NULL, &file_ctg };
CMD group_ctg = { "Group", "", CT_CATEGORY, CL_LOCAL_MENUS, NULL, NULL, 0, 0, { .ctg = { 0, &group_new_view, NULL }}, NULL, &page_ctg };
CMD edit_g_sep = { .type = CT_SEPARATOR, .flags = CL_LOCAL_MENUS, .next = &group_ctg };
CMD view_ctg = { "View", "", CT_CATEGORY, CL_VIEW_MENU, NULL, NULL, 0, 0, { .ctg = { 0, &view_fit_selected, NULL }}, NULL, &edit_g_sep };
CMD path_ctg = { "Path", "", CT_CATEGORY, CL_VIEW_MENU | CL_VIEW_TBAR, NULL, NULL, 0, 0,
		{ .ctg = { 0, &path_switch, NULL }}, NULL, &view_ctg };
CMD snap_ctg = { "Snap", "", CT_CATEGORY, CL_VIEW_MENU | CL_VIEW_TBAR, NULL, NULL, 0, 0,
		{ .ctg = { 0, &snap_hangers_cmd, NULL }}, NULL, &path_ctg };
CMD new_text_ctg = { "Texts", "Texts", CT_CATEGORY, CL_VIEW | CS_VIS_EN | CC_CHECKABLE, LIB_PATH"group_text.png", NULL, 0, 0,
			{ .ctg = { CL_VIEW_TBAR, &new_tex_text, NULL }}, NULL, &step_back_sep };
CMD new_ellipse_ctg = { "Elliptic arcs", "Elliptic arcs", CT_CATEGORY, CL_VIEW | CS_VIS_EN | CC_CHECKABLE, LIB_PATH"group_ellipse.png",
			NULL, 0, 0, { .ctg = { CL_VIEW_TBAR, &new_ellipse_3p, NULL }}, NULL, &new_text_ctg };
CMD new_circ_ctg = { "Circular arcs", "Circular arcs", CT_CATEGORY, CL_VIEW | CS_VIS_EN | CC_CHECKABLE, LIB_PATH"group_arc.png",
			NULL, 0, 0, { .ctg = { CL_VIEW_TBAR, &new_arc_3p, NULL }}, NULL, &new_ellipse_ctg };
CMD new_seg_ctg = { "Beziers", "Beziers", CT_CATEGORY, CL_VIEW | CS_VIS_EN | CC_CHECKABLE, LIB_PATH"group_bezier.png", NULL, 0, 0,
			{ .ctg = { CL_VIEW_TBAR, &new_segment, NULL }}, NULL, &new_circ_ctg };
CMD new_point_ctg = { "Points, decorations", "Points, decorations", CT_CATEGORY, CL_VIEW | CS_VIS_EN | CC_CHECKABLE,
			LIB_PATH"decoration.png", NULL, 0, 0,
			{ .ctg = { CL_VIEW_TBAR, &newpt_cmd, NULL }}, NULL, &new_seg_ctg };
struct cmd create_ctg = { "Create", "", CT_CATEGORY, CL_VIEW, NULL, NULL, 0, 0, { .ctg = { 0, &mode_select, NULL }}, NULL, &snap_ctg };

CMD edit_ctg = { "Edit", "", CT_CATEGORY, CL_UB_MENU | CL_VIEW_MENU | CL_WUH_TBAR | CL_LWUH_TBAR, NULL, NULL, 0, 0,
			{ .ctg = { 0, &edit_copy_cmd, NULL }}, NULL, &create_ctg };
struct cmd main_ctg = { "All", "", CT_CATEGORY, CL_EVERYWHERE, NULL, NULL, 0, 0, { .ctg = { 0, &edit_ctg, NULL }}, NULL, NULL };
