#ifndef __VRR_VISUALISATION_H__
#define __VRR_VISUALISATION_H__

/** @file
 * GUI - The Vrr Visualisation Layer for displaying graphic objects.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#define  VIS_SELECTOR_ENLARGEMENT 2

enum
{
	VIS_FLAGS_SELECTOR = 1,
	VIS_FLAGS_SELECTOR_BROKEN = 128,
	VIS_FLAGS_SNAP = 2,
	VIS_FLAGS_GRID = 4,
	VIS_FLAGS_TF_MOVE = 8,
	VIS_FLAGS_TF_RESIZE = 16,
	VIS_FLAGS_TF_ROTATE = 32,
	VIS_FLAGS_TF_SECOND = 48,
	VIS_FLAGS_TF_SKEW = 64
};

enum
{
	VIS_SHOW_NONE,
	VIS_SHOW_ALL,
	VIS_SHOW_SELECTED,
	VIS_SHOW_SPECIFIC
};

typedef void (* simple_cb) (void * data);

struct visualisation
{
	struct go_group *grp;		/**<top level group */

	uns flags;
	struct go_hook go_hook;

	u32 default_stroke_color, default_fill_color;
	u32 selected_stroke_color, selected_fill_color;
	real default_stroke_width;
	real selected_stroke_addition;

	struct geom_rectangle selected_bbox;
	struct vcl_path * selector;

	struct geom_point snap_point;
	struct vcl_affinity * snap_img;

	struct geom_point g_center;
	struct geom_point g_xaxis;
	struct geom_point g_yaxis;

	unsigned char anchor_mode;
	unsigned char hanger_mode;
	struct go *anchor_specific;
	struct go *hanger_specific;
	struct vcl_le_area *ha_show;

	struct geom_point transform_points [3];
	struct vcl_affinity * (transform_imgs [8]);
	struct vcl_affinity * (transform2_imgs [23]);
	int gadgets;

	struct geom_transform2 *grid_trans;
	struct vcl_property * grid_prop;
	struct vcl_affinity * grid;

	struct vcl_le_area * area;	/**< VCL area */
	struct vcl_canvas * canvas;	/**< VCL canvas */
	struct vcl_group * root;
	struct vcl_group * internal_root;
	struct vcl_affinity * zoom;

	simple_cb changed_cb;
	void * changed_data;
};


GtkWidget * visualisation_init (struct visualisation * v, struct go_group * grp,
			    simple_cb changed_cb, void * changed_data);

void visualisation_finish (struct visualisation * v);

void visualisation_set_snap (struct visualisation *v,
			real x, real y, UNUSED struct go * g1, UNUSED struct go * g2);

void visualisation_unset_snap (struct visualisation *v);

void visualisation_set_gadget_visible (struct visualisation *v, int trans_id, int visible);

void visualisation_set_center_gadget (struct visualisation *v, real x, real y);

void visualisation_set_xaxis_gadget (struct visualisation *v, real x, real y);

void visualisation_set_yaxis_gadget (struct visualisation *v, real x, real y);

void visualisation_set_tf_move (struct visualisation *v, real x, real y);

void visualisation_set_tf_resize (struct visualisation *v, real x, real y);

void visualisation_set_tf_rotate (struct visualisation *v, real x, real y);

void visualisation_set_tf_skew (struct visualisation *v, real x, real y);

void visualisation_unset_tf_move (struct visualisation *v);

void visualisation_unset_tf_second (struct visualisation *v);

void visualisation_unset_tf_skew (struct visualisation *v);

void visualisation_set_grid (struct visualisation *v, struct geom_transform2 *t, uns color);

void visualisation_unset_grid (struct visualisation *v);

void
visualisation_set_show_anchors_mode (struct visualisation *v, uns mode, struct go * go);

void
visualisation_set_show_hangers_mode (struct visualisation *v, uns mode, struct go * go);

void visualisation_gui_rectangle_update (struct visualisation *v, struct vcl_path ** r,
	struct geom_point a, struct geom_point b, struct geom_point c );

void visualisation_gui_rectangle_destroy (struct visualisation *v, struct vcl_path ** r);

void visualisation_gui_fifi_update (struct visualisation *v, struct vcl_affinity ** fifi, struct geom_point x );

void visualisation_gui_fifi_destroy (struct visualisation *v, struct vcl_affinity ** fifi);
#ifndef __VRR_VISUALISATION_H__
#define __VRR_VISUALISATION_H__

/** \file gui/visualisation.h
 The Vrr Visualisation Layer for displaying graphic objects.
*/

#define  VIS_SELECTOR_ENLARGEMENT 2

enum
{
	VIS_FLAGS_SELECTOR = 1,
	VIS_FLAGS_SELECTOR_BROKEN = 128,
	VIS_FLAGS_SNAP = 2,
	VIS_FLAGS_GRID = 4,
	VIS_FLAGS_TF_MOVE = 8,
	VIS_FLAGS_TF_RESIZE = 16,
	VIS_FLAGS_TF_ROTATE = 32,
	VIS_FLAGS_TF_SECOND = 48,
	VIS_FLAGS_TF_SKEW = 64
};

enum
{
	VIS_SHOW_NONE,
	VIS_SHOW_ALL,
	VIS_SHOW_SELECTED,
	VIS_SHOW_SPECIFIC
};

typedef void (* simple_cb) (void * data);

struct visualisation
{
	struct go_group *grp;		/**<top level group */

	uns flags;
	struct go_hook go_hook;

	u32 default_stroke_color, default_fill_color;
	u32 selected_stroke_color, selected_fill_color;
	real default_stroke_width;
	real selected_stroke_addition;

	struct geom_rectangle selected_bbox;
	struct vcl_path * selector;

	struct geom_point snap_point;
	struct vcl_affinity * snap_img;

	struct geom_point g_center;
	struct geom_point g_xaxis;
	struct geom_point g_yaxis;

	unsigned char anchor_mode;
	unsigned char hanger_mode;
	struct go *anchor_specific;
	struct go *hanger_specific;
	struct vcl_le_area *ha_show;

	struct geom_point transform_points [3];
	struct vcl_affinity * (transform_imgs [8]);
	struct vcl_affinity * (transform2_imgs [23]);
	int gadgets;

	struct geom_transform2 *grid_trans;
	struct vcl_affinity * grid;

	struct vcl_le_area * area;	/**< VCL area */
	struct vcl_canvas * canvas;	/**< VCL canvas */
	struct vcl_group * root;
	struct vcl_group * internal_root;
	struct vcl_affinity * zoom;

	simple_cb changed_cb;
	void * changed_data;
};


GtkWidget * visualisation_init (struct visualisation * v, struct go_group * grp,
			    simple_cb changed_cb, void * changed_data);

void visualisation_finish (struct visualisation * v);

void visualisation_set_snap (struct visualisation *v,
			real x, real y, UNUSED struct go * g1, UNUSED struct go * g2);

void visualisation_unset_snap (struct visualisation *v);

void visualisation_set_gadget_visible (struct visualisation *v, int trans_id, int visible);

void visualisation_set_center_gadget (struct visualisation *v, real x, real y);

void visualisation_set_xaxis_gadget (struct visualisation *v, real x, real y);

void visualisation_set_yaxis_gadget (struct visualisation *v, real x, real y);

void visualisation_set_tf_move (struct visualisation *v, real x, real y);

void visualisation_set_tf_resize (struct visualisation *v, real x, real y);

void visualisation_set_tf_rotate (struct visualisation *v, real x, real y);

void visualisation_set_tf_skew (struct visualisation *v, real x, real y);

void visualisation_unset_tf_move (struct visualisation *v);

void visualisation_unset_tf_second (struct visualisation *v);

void visualisation_unset_tf_skew (struct visualisation *v);

void visualisation_set_grid (struct visualisation *v, struct geom_transform2 *t);

void visualisation_unset_grid (struct visualisation *v);

void
visualisation_set_show_anchors_mode (struct visualisation *v, uns mode, struct go * go);

void
visualisation_set_show_hangers_mode (struct visualisation *v, uns mode, struct go * go);

void visualisation_gui_rectangle_update (struct visualisation *v, struct vcl_path ** r,
	struct geom_point a, struct geom_point b, struct geom_point c );

void visualisation_gui_rectangle_destroy (struct visualisation *v, struct vcl_path ** r);

void visualisation_gui_fifi_update (struct visualisation *v, struct vcl_affinity ** fifi, struct geom_point x );

void visualisation_gui_fifi_destroy (struct visualisation *v, struct vcl_affinity ** fifi);

/** \brief Function converts image coords to pixel coords (in view)

    \param pi input point
    \param po output point
 */
void visualisation_coords_to_pixels (struct visualisation *v, const struct geom_point *pi, struct geom_point *po);

void visualisation_coords_to_vcoords (struct visualisation *v, const struct geom_point *pi, struct geom_point *po);

void visualisation_vcoords_to_coords (struct visualisation *v, const struct geom_point *pi, struct geom_point *po);

/** \brief Function converts pixel coords (in view) to image coords

    \param pi input point
    \param po output point
 */

void visualisation_pixels_to_coords (struct visualisation *v, const struct geom_point *pi, struct geom_point *po);

void visualisation_pixels_to_vcoords (struct visualisation *v, const struct geom_point *pi, struct geom_point *po);

/** \brief Function returns bbox of visible objects in image coords */
void visualisation_get_visible_bbox (struct visualisation *v, struct geom_rectangle *bbox);

/** \brief Function returns bbox of selected objects in view coords */
void visualisation_get_selected_bbox (struct visualisation *v, struct geom_rectangle *bbox);

/** \brief Function returns bbox of selector rectangle in view coords */
void
visualisation_get_selector_bbox (struct visualisation *v, struct geom_rectangle *bbox);

/** \brief Function moves visualisation by dx, dy milimeters */
void visualisation_move (struct visualisation *v, real dx, real dy);

/** \brief Function moves visualisation so that (0, 0) in image coords will be at (x, y) in view coords */
void visualisation_absolut_move (struct visualisation *v, real x, real y);

/** \brief Function moves visualisation so that (x, y) in image coords will be at (0, 0) in view coords */
void visualisation_center (struct visualisation *v, real x, real y);

/** \brief Function moves visualisation by dx, dy pixels */
void visualisation_move_px (struct visualisation *v, real dx, real dy);

/** \brief Function moves visualisation so that absolute center will be at (x, y) in pixels */
void visualisation_absolut_move_px (struct visualisation *v, real x, real y);

/** \brief Function rotates visualisation by relative center and angle radians */
void visualisation_rotate (struct visualisation *v, real angle);

/** \brief Function scales visualisation by relative center and ratio scale_x in x and scale_y in y*/

void visualisation_scale (struct visualisation *v, real scale_x, real scale_y);

/**
    \brief Function sets orientation of the view.

    \param v identifier of view
    \param x_center x coord of the new center of the view (in image coords)
    \param y_center y coord of the new center of the view (in image coords)
    \param x_scale scale factor in x axis (in image coords)
    \param y_scale scale factor in y axis (in image coords)
    \param angle set rotation angle (rotation is centered in the center of the view)
*/
void visualisation_set_orientation (struct visualisation *v, real x_center, real y_center, real x_scale, real y_scale, real angle);

/**
    \brief Function transforms orientation of the view.

    \param v identifier of view
    \param t applied transformation (primary is matrix from image coords to view coords)
*/
void visualisation_transform (struct visualisation *v, const struct geom_transform2 *t);

/** 
 * \brief Scale/move so that the rectangle fills the view in at least one dimension.
 */
int visualisation_fit_box( struct visualisation * vis, struct geom_rectangle bbox );

#endif

/** \brief Function converts image coords to pixel coords (in view)

  	\param v identifier of view
    \param pi input point
    \param po output point
 */
void visualisation_coords_to_pixels (struct visualisation *v, const struct geom_point *pi, struct geom_point *po);

void visualisation_coords_to_vcoords (struct visualisation *v, const struct geom_point *pi, struct geom_point *po);

void visualisation_vcoords_to_coords (struct visualisation *v, const struct geom_point *pi, struct geom_point *po);

/** \brief Function converts pixel coords (in view) to image coords

	\param v identifier of view
    \param pi input point
    \param po output point
 */

void visualisation_pixels_to_coords (struct visualisation *v, const struct geom_point *pi, struct geom_point *po);

void visualisation_pixels_to_vcoords (struct visualisation *v, const struct geom_point *pi, struct geom_point *po);

/** \brief Function returns bbox of visible objects in image coords */
void visualisation_get_visible_bbox (struct visualisation *v, struct geom_rectangle *bbox);

/** \brief Function returns bbox of selected objects in view coords */
void visualisation_get_selected_bbox (struct visualisation *v, struct geom_rectangle *bbox);

/** \brief Function returns bbox of selector rectangle in view coords */
void
visualisation_get_selector_bbox (struct visualisation *v, struct geom_rectangle *bbox);

/** \brief Function moves visualisation by dx, dy milimeters */
void visualisation_move (struct visualisation *v, real dx, real dy);

/** \brief Function moves visualisation so that (0, 0) in image coords will be at (x, y) in view coords */
void visualisation_absolut_move (struct visualisation *v, real x, real y);

/** \brief Function moves visualisation so that (x, y) in image coords will be at (0, 0) in view coords */
void visualisation_center (struct visualisation *v, real x, real y);

/** \brief Function moves visualisation by dx, dy pixels */
void visualisation_move_px (struct visualisation *v, real dx, real dy);

/** \brief Function moves visualisation so that absolute center will be at (x, y) in pixels */
void visualisation_absolut_move_px (struct visualisation *v, real x, real y);

/** \brief Function rotates visualisation by relative center and angle radians */
void visualisation_rotate (struct visualisation *v, real angle);

/** \brief Function scales visualisation by relative center and ratio scale_x in x and scale_y in y*/

void visualisation_scale (struct visualisation *v, real scale_x, real scale_y);

/**
    \brief Function sets orientation of the view.

    \param v identifier of view
    \param x_center x coord of the new center of the view (in image coords)
    \param y_center y coord of the new center of the view (in image coords)
    \param x_scale scale factor in x axis (in image coords)
    \param y_scale scale factor in y axis (in image coords)
    \param angle set rotation angle (rotation is centered in the center of the view)
*/
void visualisation_set_orientation (struct visualisation *v, real x_center, real y_center, real x_scale, real y_scale, real angle);

/**
    \brief Function transforms orientation of the view.

    \param v identifier of view
    \param t applied transformation (primary is matrix from image coords to view coords)
*/
void visualisation_transform (struct visualisation *v, const struct geom_transform2 *t);

/** 
 * \brief Scale/move so that the rectangle fills the view in at least one dimension.
 */
int visualisation_fit_box( struct visualisation * vis, struct geom_rectangle bbox );

#endif
