#ifndef _KERNEL_CONST_H_
#define _KERNEL_CONST_H_

/** @file
 *
 *  Kernel - Enumerations and other constants	
 *
 * Copyright (c) 2004-2007    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */


#define SCONST


/***************************
 *                         *
 *      Object types       *
 *                         *
 ***************************/


enum o_kind
{
	T_OBJ = 17,
	T_GO  = 29,
	T_HANGER = 34,	// Ignore this - not used in o.kind (scheme's constant)
	T_ANCHOR = 76,	// Ignore this - not used in o.kind (scheme's constant)
};

enum obj_type
{
	OT_UNIVERSE,
	OT_DOCUMENT,
	OT_TLO,
	OT_TEMP,
	OT_ZOMBIE,
	OT_MAX,
};

enum go_type
{
	GOT_GROUP,
	GOT_POINT,
	GOT_SEGMENT,
	GOT_BEZIER,
	GOT_ELARC,
	GOT_PARAMETRIC_POINT,
	GOT_INTERSECTION,
	GOT_TEXT,
	GOT_TEX_TEXT,
	GOT_DECORATOR_POINT,
	GOT_ARROW,
	GOT_MAX,
};

enum go_group_subtype
{
	GOST_GROUP,
	GOST_GROUP_PATH,
	GOST_GROUP_TLG,
	GOST_GROUP_TEMP,
	GOST_GROUP_ZOMBIE,
	GOST_GROUP_MAX,
};



/**************************
 *                        *
 *     Object flags       *
 *                        *
 **************************/


enum obj_flag
{
	OF_SELECTED = 0x1,
	OF_TSORT_ACTIVE = 0x2, // used only in tlo
	OF_TSORT_DIRTY = 0x4, // used only in tlo
	OF_TSORT_PRESORT = 0x8, // used only in tlo
};

enum go_flag
{
	GOF_SELECTED = 0x1,
	GOF_TSORT = 0x2,
	GOF_INVISIBLE = 0x4,
	
	// Internal flags
	GOF_PARPOINT_INVALID_X = 0x40, // used only in parametric point
	GOF_PARPOINT_INVALID_Y = 0x80, // used only in parametric point
	GOF_INTERSECTION_INVALID_X = 0x100, // used only in intersection
	GOF_INTERSECTION_INVALID_Y = 0x200, // used only in intersection
	GOF_ELARC_START = 0x400, // used only in elarc
	GOF_ELARC_DIF = 0x800, // used only in elarc
};



/******************************************
 *                                        *
 *  Objects dependence realization types  *
 *                                        *
 ******************************************/


enum quantity_type
{
	QT_POINT,
	QT_CURVE,
	QT_MAX,
};

enum linearity_type
{
	LT_NONE,
	LT_SIMILARITY,
	LT_AFFINITY,
};



/****************************
 *                          *
 *    Property and unit     *
 *                          *
 ****************************/


enum prop_type
{
SCONST	PT_UNS,
SCONST	PT_POINTER,
SCONST	PT_STRING,
SCONST	PT_REAL,
SCONST	PT_MAX,
};

enum prop_subtype_uns
{
SCONST	PTU_UNSPECIFIED,
SCONST	PTU_COLOR,
SCONST	PTU_FONT,
// allowed values are 0 .. GEOM_CONIC_ARC_iP for PTU_CONIC_TYPE_iP
SCONST	PTU_CONIC_TYPE_0P,
SCONST	PTU_CONIC_TYPE_1P,
SCONST	PTU_CONIC_TYPE_2P,
SCONST	PTU_CONIC_TYPE_3P,
SCONST	PTU_BOOLEAN,
SCONST	PTU_CAP_STYLE,
SCONST	PTU_JOIN_STYLE,
SCONST	PTU_STROKE_STYLE,
SCONST	PTU_ALIGNMENT_X,
SCONST	PTU_ALIGNMENT_Y,
SCONST	PTU_ARROW_FRONT,
SCONST	PTU_ARROW_BACK,
SCONST	PTU_ARROW_ALIGN,
SCONST	PTU_MAX,
};

enum prop_subtype_pointer
{
SCONST	PTP_UNSPECIFIED,
SCONST	PTP_TRANSFORM,
	PTP_TEX_PROCESS,	// internal
SCONST	PTP_MAX,
};

enum prop_subtype_string
{
SCONST	PTS_UNSPECIFIED,
SCONST	PTS_FILE_NAME,
SCONST	PTS_LARGE_TEXT,
SCONST	PTS_MAX,
};

enum prop_subtype_real
{
SCONST	PTR_UNSPECIFIED,
SCONST	PTR_POSITIVE,
SCONST	PTR_COORDINATE,
SCONST	PTR_LENGTH,
SCONST	PTR_ANGLE_PI,
SCONST	PTR_ANGLE_2PI,
SCONST	PTR_ANGLE_4PI,
SCONST	PTR_REFERENCE,
SCONST	PTR_MAX,
};

enum prop_flags
{
SCONST	PTF_VIRTUAL = 0x1,
SCONST	PTF_READ_ONLY = 0x2,
SCONST	PTF_SAVE = 0x4,
	PTF_RECYCLABLE = 0x8, // A flag for the property recycler in GUI.
	PTF_UNDO_CREATE = 0x40,	// An internal flag.
	PTF_UNDO_DELETE = 0x80,	// An internal flag.
};

enum prop_quantity
{
SCONST	PQ_NONE,
SCONST	PQ_LENGTH,
SCONST	PQ_ANGLE,
SCONST	PQ_REFERENCE,
SCONST	PQ_MAX,
};

enum prop_cap_style
{
SCONST	PSC_BUTT,
SCONST	PSC_ROUND,
SCONST	PSC_PROJECTING,
SCONST	PSC_MAX,
};

enum prop_join_style
{
SCONST	PSJ_MITER,
SCONST	PSJ_ROUND,
SCONST	PSJ_BEVEL,
SCONST	PSJ_MAX,
};

enum prop_stroke_style
{
SCONST PSS_SOLID,
SCONST PSS_DASHED,
SCONST PSS_DOTTED,
SCONST PSS_DASH_DOTTED,
SCONST PSS_MAX,
};

enum prop_alignment_x
{
SCONST	PTAX_REF_LEFT,
SCONST	PTAX_REF_CENTER,
SCONST	PTAX_REF_RIGHT,
SCONST	PTAX_REF_RELATIVE,
SCONST	PTAX_BBOX_LEFT,
SCONST	PTAX_BBOX_CENTER,
SCONST	PTAX_BBOX_RIGHT,
SCONST	PTAX_BBOX_RELATIVE,
SCONST	PTAX_MAX,
};

enum prop_alignment_y
{
SCONST	PTAY_BASELINE,
SCONST	PTAY_BBOX_BOTTOM,
SCONST	PTAY_BBOX_CENTER,
SCONST	PTAY_BBOX_TOP,
SCONST	PTAY_BBOX_RELATIVE,
SCONST	PTAY_MAX,
};

enum prop_arrow_alignment
{
SCONST	PTAA_FRONT,
SCONST	PTAA_BACK,
SCONST	PTAA_MAX,
};

enum prop_arrow_front
{
SCONST	PTAF_STRAIGHT,
SCONST	PTAF_PARABOLIC,
SCONST	PTAF_MAX,
};

enum prop_arrow_back
{
SCONST	PTAB_NONE,
SCONST	PTAB_STRAIGHT,
SCONST	PTAB_POLY,
SCONST	PTAB_PARABOLIC,
SCONST	PTAB_MAX,
};

#define PROP_SUBTYPE_MAX MAX( MAX(PTU_MAX, PTP_MAX), MAX(PTR_MAX, PTS_MAX) )

/* Size of unit array. This value is also used as error value in property unit functions.
 * Minimum value is PQ_MAX and maximum is 255. */
#define PROP_UNIT_MAX 255

/* Size of buffer which is used for dump property. */
#define PROP_DUMP_SIZE 256

/* Initial value of properties */
#define DEFAULT_STROKE_WIDTH 0.15
#define PROP_INIT_FILL_COLOR 0	// black: red 0, green 0, blue 0, alpha 0
#define PROP_INIT_STROKE_COLOR 255	// black: red 0, green 0, blue 0, alpha 255
#define PROP_INIT_STROKE_CAP PSC_BUTT
#define PROP_INIT_STROKE_JOIN PSJ_MITER
#define PROP_INIT_STROKE_STYLE PSS_SOLID
#define PROP_INIT_STROKE_WIDTH DEFAULT_STROKE_WIDTH
#define PROP_INIT_TEXT_ALIGNMENT_X PTAX_REF_RELATIVE
#define PROP_INIT_TEXT_ALIGNMENT_Y PTAY_BBOX_RELATIVE
#define PROP_INIT_TEXT_FILL_COLOR PROP_INIT_STROKE_COLOR

/*
 * Type of visualization's hook
 */

enum changed_kind
{
	CK_CHANGED,
	CK_ALTERED,
	CK_TRANSFORMED
};



/**************************
 *                        *
 *      Transaction       *
 *                        *
 **************************/


enum trans_log_call
{
	TRANS_LOG_FAIL = 0x1,
	TRANS_LOG_SUCCESS = 0x2,
	TRANS_LOG_ALWAYS = TRANS_LOG_FAIL | TRANS_LOG_SUCCESS,
};


/***********************
 *                     *
 *      Settings       *
 *                     *
 ***********************/


/** the user settings directory */
#define VRR_DIR				"/.vrr"
/** The global settings file */
#define GLOBAL_SETTINGS_FILE		"/settings"
/** Maximum volume of undo history */
#define MAX_UNDO_HISTORY_VOLUME_KB	 100
/** Maximum items of undo history */
#define MAX_UNDO_HISTORY_COUNT	 100
/** Maximum memory for cache in geomlib */
#define MAX_CACHE_MEMORY_LIMIT_KB	1024
/** Maximum items in cache */
#define MAX_CACHE_COUNT	10000
/** Minimum items in cache */
#define MIN_CACHE_COUNT 10


#endif
