/*
 * Kernel - Groups
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "kernel/internal.h"

#define GO_ID GROUP
#define GO_SUBID GROUP

struct hanger_info hanger_go_group_tlg_info[] = {
	{ NULL, QT_POINT, 0, go_group_info, OFFSETOF(struct hanger_smart, q), NULL, 1 },
	{ NULL, QT_CURVE, 0, go_group_info, OFFSETOF(struct hanger_smart, q), NULL, 1 },
};



void go_group_recompute(struct go_group *g, struct geom_rectangle *bbox UNUSED)
{
	g->bbox_all.ur.x = NAN;
}

void go_group_transform(struct go_group *g, struct geom_transform *t UNUSED, struct geom_rectangle *bbox)
{
	go_group_recompute(g, bbox);
}

struct go_group *go_group_new(void)
{
	struct go_group *g = xmalloc_zero(sizeof(struct go_group));
	group_init(g, GOST_GROUP);
	g->bbox_all.ur.x = NAN;
	return g;
}


#undef GO_SUBID
#define GO_SUBID PATH

PROP_STYLES
PROP_VIRTUAL(PROP_STYLES_INIT)


void go_path_recompute(struct go_path *g, struct geom_rectangle *bbox UNUSED)
{
	struct go *a, *b;
	struct go_info *ai, *bi;
	
	geom_rectangle_set_empty(bbox);

	a = group_get_first_go(&g->group);
	if(!a)
		return;
	ai = go_get_info(a);
	if(ai->a_start == -1 || ai->h_end == -1)
		TRANS_FAIL(("You cannot add this into a path."), ("Path %s at %p contains go %s at %p of subtype %s.", go_lookup_name(&g->group.go), g, go_lookup_name(a), a, ai->name->text));
	if(trans_tlo->expand_ref_count)
		geom_rectangle_merge(bbox, go_get_rtree_bbox(a), bbox);

	for(; (b = group_get_next_go(&g->group, a)); a = b, ai = bi)
	{
		bi = go_get_info(b);
		if(bi->a_start == -1 || bi->h_end == -1)
			TRANS_FAIL(("You cannot add this into a path."), ("Path %s at %p contains go %s at %p of subtype %s.", go_lookup_name(&g->group.go), g, go_lookup_name(b), b, bi->name->text));
		if(go_get_anchor(b, bi->a_start)->h != go_get_hanger(a, bi->h_end))
			TRANS_FAIL(("Path must be continuous."), ("Go %s at %p doesn't hang on go %s at %p.", go_lookup_name(b), b, go_lookup_name(a), a));
		
		if(trans_tlo->expand_ref_count)
			geom_rectangle_merge(bbox, go_get_rtree_bbox(b), bbox);
	}

	// FIXME instead of 3 there can be exact value based on arrow angle
	if(!GEOM_ETEST(geom_rectangle_enlarge(bbox, g -> group . stroke_width * 3 + 1, bbox)))
		TRANS_FAIL(("Numeric error."), ("geom_rectangle_enlarge fail."));

	if(!GEOM_ETEST(geom_rectangle_enlarge(bbox, g -> group . stroke_width * 3 + 1, bbox))) 
		TRANS_FAIL(("Numeric error."), ("geom_rectangle_enlarge fail."));

	g->group.bbox_all = *bbox;
	geom_path_after_change(&g->group.geom);
}

void go_path_transform(struct go_path *g, struct geom_transform *t UNUSED, struct geom_rectangle *bbox)
{
	go_path_recompute(g, bbox);
}

struct go_path *go_path_new(void)
{
	struct go_path *g = xmalloc_zero(sizeof(struct go_path));
	group_init(&g->group, GOST_GROUP_PATH);

	PROP_STYLES_SET_DEFAULT_VALUE(&g->group);

	geom_instance_zero_init(go2geom(&g->group.go), GEOM_CLASS(path));
	geom_path_create(&g->group.geom);
	geom_rectangle_set_empty(&g->group.bbox_all);
	return g;
}

struct go_path * group_pathize_selected_go(struct go_group * group)
{
	ASSERT(trans_present && trans_tlo && trans_tlo->obj.type == OT_TLO);
	ASSERT(group && go_get_tlo(&group->go) == trans_tlo);

	struct go * go = group_get_first_selected_go(group);
	if (!go)
		trans_fail("No selected objects.");
	struct go * go_n, * go_nn;
	struct go_path * path;
	struct hanger * h;
	struct anchor * a;

	TSORT_NO_RECOMPUTE_BEGIN
		path = go_path_new();
		go_link_after(&path->group.go, group, go);
		go_relink(go, &path->group, NULL);
		go_n = go;
	
		for (;;)
		{
			if (!(h = go_get_hanger_end(go_n)))
				trans_fail("Incompatible objects selected.");
			WALK_HANGER_ANCHOR(h, a)
				if (go_is_selected(go_nn = anchor_get_go(a)) && a == go_get_anchor_start(go_nn))
					break;
			if (a)
			{
				go_relink(go_nn, &path->group, go_n);
				go_n = go_nn;
			}
			else
				break;
		}
		go_n = go;
		for (;;)
		{
			if (!(a = go_get_anchor_start(go_n)))
				trans_fail("Incompatible objects selected.");
			h = a->h;
			if (hanger_is_mc(h) || h != go_get_hanger_end(go_nn = hanger_get_go(h)) || 
				!go_is_selected(go_nn) )
				break;
			else
			{
				go_relink(go_nn, &path->group, NULL);
				go_n = go_nn;
			}
		}
	TSORT_NO_RECOMPUTE_END
	return path;
}

#undef GO_SUBID
#define GO_SUBID TLG


void go_tlg_recompute(struct go_tlg *g, struct geom_rectangle *bbox)
{
	go_group_recompute(&g->group, bbox);
}

void go_tlg_transform(struct go_tlg *g, struct geom_transform *t, struct geom_rectangle *bbox)
{
	go_group_transform(&g->group, t, bbox);
}


#undef GO_SUBID

struct go_info go_group_info[GOST_GROUP_MAX] = {
	[GOST_GROUP] = { NULL, 0, 0, 0, NULL, -1, -1, -1, -1, 0, 0, NULL, 0, 0, &o_prop_name },
	[GOST_GROUP_PATH] = { NULL, 0, 0, 0, NULL, -1, -1, -1, -1, 0, 0, NULL, 0, 0, &go_path_prop_stroke_style  },
	[GOST_GROUP_TLG] = { NULL, 0, 0, OFFSETOF(struct go_tlg, mc), hanger_go_group_tlg_info, -1, -1, -1, -1, 0, 0, NULL, 0, 0, &o_prop_name },
	[GOST_GROUP_TEMP] = { NULL, 0, 0, OFFSETOF(struct go_tlg, mc), NULL, -1, -1, -1, -1, 0, 0, NULL, 0, 0, &o_prop_name },
	[GOST_GROUP_ZOMBIE] = { NULL, 0, 0, OFFSETOF(struct go_tlg, mc), NULL, -1, -1, -1, -1, 0, 0, NULL, 0, 0, &o_prop_name },
};


void go_group_init(void)
{
//	PINIT(go_group) ();
	go_group_info[GOST_GROUP].name = string_lookup_ref("group");
	go_group_info[GOST_GROUP_PATH].name = string_lookup_ref("path");
	go_group_info[GOST_GROUP_TLG].name = string_lookup_ref("top-level-group");
	go_group_info[GOST_GROUP_TEMP].name = string_lookup_ref("temp-group");
	go_group_info[GOST_GROUP_ZOMBIE].name = string_lookup_ref("zombie-group");
	string_go_type[GOT_GROUP] = string_lookup_ref("group");
	hanger_go_group_tlg_info[QT_POINT].name = string_lookup_ref("mouse-click-position");
	hanger_go_group_tlg_info[QT_CURVE].name = string_lookup_ref("mouse-click-curve");
	go_path_virtual_prop_init();
}

void go_group_finish(void)
{
	go_path_virtual_prop_finish();
	string_unref(go_group_info[GOST_GROUP].name);
	string_unref(go_group_info[GOST_GROUP_PATH].name);
	string_unref(go_group_info[GOST_GROUP_TLG].name);
	string_unref(go_group_info[GOST_GROUP_TEMP].name);
	string_unref(go_group_info[GOST_GROUP_ZOMBIE].name);
	string_unref(hanger_go_group_tlg_info[QT_POINT].name);
	string_unref(hanger_go_group_tlg_info[QT_CURVE].name);
}

