/*
 * Kernel - Segments
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "kernel/internal.h"

#define GO_ID SEGMENT
#define GO_SUBID SEGMENT

PROP_STYLES
PROP_ALL_REHANG(end, PROP_STYLES_LAST, geom.b, 1, "Hanger of end point must be a mouse click to change position.", 0)
PROP_ALL_REHANG(start, PSTRUCT(end_x), geom.a, 0, "Hanger of start point must be a mouse click to change position.", 0)

PROP_VIRTUAL(PROP_STYLES_INIT,
		&PSTRUCT(start_x), "endpoint-1-x", &PSTRUCT(start_y), "endpoint-1-y",
		&PSTRUCT(end_x), "endpoint-2-x", &PSTRUCT(end_y), "endpoint-2-y")

INFO_ANCHOR_2(
		"start", QT_POINT, LT_AFFINITY,
		"end", QT_POINT, LT_AFFINITY)

INFO_HANGER_3(
		"curve", QT_CURVE, h[0], NULL, LT_AFFINITY,
		"start", QT_POINT, geom.a, NULL, LT_AFFINITY,
		"end", QT_POINT, geom.b, NULL, LT_AFFINITY)

INFO_GO_1("segment",
		SEGMENT, "segment", 3, 0, 1, 2, -1, 2, 0, 1, start_x)

void go_segment_recompute(struct go_segment *s, struct geom_rectangle *bbox)
{
	s->geom.a = hanger_get_position(s->a[0].h);
	s->geom.b = hanger_get_position(s->a[1].h);
	geom_rectangle_init_from_points(bbox, &s->geom.a, &s->geom.b);
	if(!GEOM_ETEST(geom_rectangle_enlarge(bbox, s->stroke_width * 0.71 + 1, bbox)))
		TRANS_FAIL(("Numeric error."), ("geom_rectangle_enlarge fail."));
	geom_segment_after_change(&s->geom);
}

void go_segment_transform(struct go_segment *s, struct geom_transform *t UNUSED, struct geom_rectangle *bbox)
{
	go_segment_recompute(s, bbox);
}

struct go_segment *go_segment_new(struct hanger *h1, struct hanger *h2)
{
	struct go_segment *s = xmalloc_zero(sizeof(struct go_segment));
	go_init(&s->go, GOT_SEGMENT, GOST_SEGMENT, h1, h2);
	
	PROP_STYLES_SET_DEFAULT_VALUE(s);
	s->unit_end_x = s->unit_end_y = s->unit_start_x = s->unit_start_y = PROP_UNIT_MAX;
	return s;
}

