/*
 * Kernel - Texts
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG
#include "kernel/internal.h"
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include "font/font.h"
#include "import/import.h"


/* FIXME: nacitat z initsouboru ci menit v options */
static char *tex_tmp_path = "/tmp";
static char *tex_path = "tex";

#define TEXT_BASELINE        0x1
#define TEXT_LEFT_REF_POINT  0x2
#define TEXT_RIGHT_REF_POINT 0x4

static void get_text_bbox(struct geom_rectangle *text_box, struct geom_point *anchor, struct geom_point *rpos,
		struct geom_point *apos, struct geom_transform *trans, real ref_width, uns align_x, uns align_y, struct geom_rectangle *tlo_box)
{
	real x, y, rel_width, rel_x, rel_y;
	uns flags;

	// FIXME: could be faster
	flags = 0;
	rel_x = rpos->x;
	rel_y = rpos->y;
	switch (align_x) {
	case PTAX_REF_LEFT: rel_x = 0; flags |= TEXT_LEFT_REF_POINT | TEXT_RIGHT_REF_POINT; break;
	case PTAX_REF_CENTER: rel_x = 0.5; flags |= TEXT_LEFT_REF_POINT | TEXT_RIGHT_REF_POINT; break;
	case PTAX_REF_RIGHT: rel_x = 1; flags |= TEXT_LEFT_REF_POINT | TEXT_RIGHT_REF_POINT; break;
	case PTAX_REF_RELATIVE: flags |= TEXT_LEFT_REF_POINT | TEXT_RIGHT_REF_POINT; break;
	case PTAX_BBOX_LEFT: rel_x = 0; break;
	case PTAX_BBOX_CENTER: rel_x = 0.5; break;
	case PTAX_BBOX_RIGHT: rel_x = 1; break;
	case PTAX_BBOX_RELATIVE: break;
	default: BUG();
	}
	switch (align_y) {
	case PTAY_BASELINE: flags |= TEXT_BASELINE; break;
	case PTAY_BBOX_BOTTOM: rel_y = 0; break;
	case PTAY_BBOX_CENTER: rel_y = 0.5; break;
	case PTAY_BBOX_TOP: rel_y = 1; break;
	case PTAY_BBOX_RELATIVE: break;
	default: BUG();
	}
	
	/* X coordinate of (text main refpoint - anchor) before transformation  */
	if (flags & TEXT_RIGHT_REF_POINT)
		rel_width = ref_width;
	else
		rel_width = text_box->ur.x;

	if (flags & TEXT_LEFT_REF_POINT)
		x = 0;
	else {
		x = -text_box->ll.x;
		rel_width -= text_box->ll.x;
	}
	x += apos->x - rel_width * rel_x;

	/* Y coordinate of (text main refpoint - anchor) before transformation */
	if (flags & TEXT_BASELINE)
		y = 0;
	else
		y = (text_box->ll.y - text_box->ur.y) * rel_y - text_box->ll.y;
	y += apos->y;

	/* dest = src transformed from text coordinates to tlo coordinates */
	trans->coef[0][2] = x * trans->coef[0][0] + y * trans->coef[0][1] + anchor->x;
	trans->coef[1][2] = x * trans->coef[1][0] + y * trans->coef[1][1] + anchor->y;
	trans->flags &= GEOM_TRANSFORM_SIMILAR;

#ifdef LOCAL_DEBUG
	GEOM_DUMP_ENTER(1, "get_text_bbox: ");
	GEOM_SDUMP("anchor="); geom_point_dump(anchor);
	GEOM_SDUMP("rpos="); geom_point_dump(rpos);
	GEOM_SDUMP("apos="); geom_point_dump(apos);
	GEOM_SDUMP("trans="); geom_transform_dump(trans);
	GEOM_DUMP_LEAVE;
#endif

	if(!(GEOM_RTEST(trans->coef[0][2]) && GEOM_RTEST(trans->coef[1][2])))
		TRANS_FAIL(("Numeric error."), ("Invalid visualization point."));

	if(!GEOM_ETEST(geom_rectangle_transform(text_box, trans, tlo_box)))
		TRANS_FAIL(("Numeric error."), ("geom_rectangle_transform fail"));
}

static inline int is_rpos_valid(real r) { return r >= 0 && r <= 1; }

#define GO_ID TEXT
#define GO_SUBID TEXT

PROP_INVISIBLE PROP_FILL_COLOR
PROP_ALL_WITHOUT_UNIT(transform, PSTRUCT(fill_color), PT_POINTER, PTP_TRANSFORM, transformation, t,
		1, (("Cannot fail."), ("Bug, fix me.")), CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL_WITHOUT_UNIT(alignment_y, PSTRUCT(transform), PT_UNS, PTU_ALIGNMENT_Y, alignment_y, u,
		value.u < PTAY_MAX, (("Invalid value."), ("Alignment Y cannot be %d.", value.u)), CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL_WITHOUT_UNIT(alignment_x, PSTRUCT(alignment_y), PT_UNS, PTU_ALIGNMENT_X, alignment_x, u,
		value.u < PTAX_MAX, (("Invalid value."), ("Alignment X cannot be %d.", value.u)), CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL(apos_y, PSTRUCT(alignment_x), PT_REAL, PTR_COORDINATE, apos.y, r,
		1, (("Cannot fail."), ("Bug, fix me.")), CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL(apos_x, PSTRUCT(apos_y), PT_REAL, PTR_COORDINATE, apos.x, r,
		1, (("Cannot fail."), ("Bug, fix me.")), CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL(rpos_y, PSTRUCT(apos_x), PT_REAL, PTR_REFERENCE, rpos.y, r,
      is_rpos_valid(value.r), (("Invalid value."), ("rpos_y cannot be %f.", value.r)), CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL(rpos_x, PSTRUCT(rpos_y), PT_REAL, PTR_REFERENCE, rpos.x, r,
      is_rpos_valid(value.r), (("Invalid value."), ("rpos_x cannot be %f.", value.r)), CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL_REHANG(coordinate, PSTRUCT(rpos_x), geom.point, 0, "Hanger must be a mouse click to change position.", 0)
PROP_ALL(size, PSTRUCT(coordinate_x), PT_REAL, PTR_LENGTH, font_size, r,
		value.r > 0, (("Size of font must be positive."), ("Size cannot be %f.", value.r)), CHANGED, LOW, PTF_RECYCLABLE, UINT_MAX)
PROP_ALL_WITHOUT_UNIT(font, PSTRUCT(size), PT_UNS, PTU_FONT, font_id, u, font_id_valid(value.u), (("Bad font id."), ("Bad font id %d.", value.u)), CHANGED, LOW, PTF_RECYCLABLE, UINT_MAX)

static struct prop_virtual PSTRUCT(text);

static void PSET(text) (struct o *o, prop_value text)
{
	ASSERT_GO_SUBTYPE_ID(o);
	struct go_text *t = (struct go_text*)o;

	trans_new_undo_prop_virtual(o, &PSTRUCT(text), (prop_value){.s = t->text});
	string_unref(t->text);
	t->text = string_ref(text.s);
	tsort_dirty_recompute(&t->go);
}

static void PLOW(text) (struct o *o, prop_value text)
{
	ASSERT_GO_SUBTYPE_ID(o);
	struct go_text* t = (struct go_text*)o;
	string_unref(t->text);
	t->text = string_ref(text.s);
}

PROP_GET(text, text, s)
PROP_UNIT_GET_DIE(text)
PROP_UNIT_SET_DIE(text)
PROP_STRUCT(text, PSTRUCT(font), PT_STRING, PTS_LARGE_TEXT, 0, UINT_MAX)

PROP_VIRTUAL(&PSTRUCT(text), string_text->text, &PSTRUCT(size), string_font_size->text, &PSTRUCT(font), string_font->text,
		&PSTRUCT(coordinate_x), string_coordinate_x->text, &PSTRUCT(coordinate_y), string_coordinate_y->text,
		&PSTRUCT(rpos_x), string_rpos_x->text, &PSTRUCT(rpos_y), string_rpos_y->text,
		&PSTRUCT(apos_x), string_apos_x->text, &PSTRUCT(apos_y), string_apos_y->text, &PSTRUCT(transform), string_transform->text,
		&PSTRUCT(alignment_x), string_alignment_x->text, &PSTRUCT(alignment_y), string_alignment_y->text,
		PROP_INVISIBLE_INIT, PROP_FILL_COLOR_INIT)


INFO_ANCHOR_1("point", QT_POINT, LT_AFFINITY)
INFO_HANGER_0

static int text_get_string_bbox(struct go_text *t, struct geom_rectangle *bbox)
{
	struct font_ctl ctl;

	font_ctl_init(&ctl);
	ctl.font_id = t->font_id;
	ctl.font_size = t->font_size;
	ctl.charmap = FONT_CHARMAP_STD;
	if(font_get_string_bbox(&ctl, t->text->text, bbox) < 0)
		return 0;
	font_ctl_cleanup(&ctl);

	return 1;
}

void go_text_recompute(struct go_text *t, struct geom_rectangle *bbox)
{
	t->geom.point = hanger_get_position(t->a->h);

	// Zjisteni bboxy stringu v LFT souradnicich.
	if(!text_get_string_bbox(t, &t->string_bbox))
		TRANS_FAIL(("Numeric error."), ("text_get_string_bbox fail."));

	// Doplneni transformacni matice a spocitani bboxu textu v tlo souradnicich.
	get_text_bbox(&t->string_bbox, &t->geom.point, &t->rpos, &t->apos, t->transformation, t->string_bbox.ur.x, t->alignment_x, t->alignment_y, &t->text_bbox);
	
	*bbox = t->text_bbox;
	geom_rectangle_merge_point(bbox, &t->geom.point, bbox);
	geom_item_after_change(&t->geom);
}

void go_text_transform(struct go_text *text, struct geom_transform *tr, struct geom_rectangle *bbox)
{
	struct geom_transform *t = text->transformation;
	text->transformation = xmalloc(sizeof(struct geom_transform));
	trans_new_undo_prop_virtual((struct o*)text, &PSTRUCT(transform), (prop_value){.t = t});
	text->transformation->coef[0][2] = text->transformation->coef[1][2] = 0;
	if(geom_transform_merge(t, tr, text->transformation) != 0)
		TRANS_FAIL(("Numeric error."), ("geom_transform_merge fail"));
	go_text_recompute(text, bbox);
}

struct go_text *go_text_new(struct hanger *refpoint, real rpos_x, real rpos_y, real apos_x, real apos_y, string text, int fnt_id, real fnt_size)
{
	if(!font_id_valid(fnt_id))
		trans_fail("Bad font id.");
	if(fnt_size <= 0)
		trans_fail("The size of font must be great than zero.");
	struct go_text *t = xmalloc_zero(sizeof(struct go_text));

	go_init(&t->go, GOT_TEXT, GOST_TEXT, refpoint);
	t->rpos = (struct geom_point) { rpos_x, rpos_y };
	t->apos = (struct geom_point) { apos_x, apos_y };
	t->text = string_ref(text);
	t->font_id = fnt_id;
	t->font_size = fnt_size;
	t->transformation = xmalloc(sizeof(struct geom_transform));
	*t->transformation = geom_transform_identity;
	t->alignment_x = PROP_INIT_TEXT_ALIGNMENT_X;
	t->alignment_y = PROP_INIT_TEXT_ALIGNMENT_Y;
	t->fill_color = PROP_INIT_TEXT_FILL_COLOR;
	t->unit_rpos_x = t->unit_rpos_y = t->unit_apos_x = t->unit_apos_y = t->unit_coordinate_x = t->unit_coordinate_y = PROP_UNIT_MAX;

	return t;
}

INFO_GO_1("text",
		TEXT, "text", 0, -1, -1, -1, -1, 1, -1, -1, text)


#undef GO_ID
#undef GO_SUBID


#define GO_ID TEX
#define GO_SUBID TEX

static struct tex_process* tex_process(string text)
{
	ASSERT(trans_present);
	char buf[500];
	char dirnm[20] = "vrrtexXXXXXX";
	char olddir[500];
	char *tmpdir;
	FILE *dvifile;
	int fd, r, cnt = 0, ret = 0;
	string header = prop_lookup(void2o(trans_tlo), string_tex_header, PT_STRING, (prop_value){ .s = string_tex_header_default }).s;
	string footer = prop_lookup(void2o(trans_tlo), string_tex_footer, PT_STRING, (prop_value){ .s = string_tex_footer_default }).s;
	struct tex_process *process = xmalloc_zero(sizeof(struct tex_process));
	process->glyphs = garr_new(32, sizeof(struct tex_glyph));

	DBG("Running TeX: %s\n", text->text);

	if (!getcwd(olddir, sizeof(olddir)))
		goto err7;
	chdir(tex_tmp_path);
#ifdef HAVE_MKDTEMP
	tmpdir = mkdtemp(dirnm);
	if (!tmpdir)
		goto err6;
#else
	int try = 0;
	while (1)
	{
		tmpdir = mktemp(dirnm);
		if (!tmpdir)
			goto err6;
		if (!mkdir(tmpdir, 0700))
			break;
		if (errno != EEXIST || try++ == 10)
			goto err6;
	}
#endif
	if (chdir(tmpdir))
		goto err5;

	if ((fd = creat("job.tex", 0600))<0)
		goto err4;

	if (write(fd, header->text, header->length)<0
			|| write(fd, text->text, text->length)<0
			|| write(fd, footer->text, footer->length)<0)
		goto err3;
	close(fd);

	if (snprintf(buf, sizeof(buf), "%s --interaction batchmode job.tex >/dev/null", tex_path)>=(int)sizeof(buf))
		goto err3;

	r = system(buf);
	if (r<0 || !WIFEXITED(r) || WEXITSTATUS(r)!=0)
		goto err2;

	if (!(dvifile = fopen("job.dvi", "r")))
		goto err2;

	if ((cnt = import_dvi(dvifile, process->glyphs))<0)
		goto err1;
	process->glyphs_cnt = cnt;

	ret = 1;
err1:
	fclose(dvifile);
err2:
	unlink("job.dvi");
	unlink("job.log");
err3:
	unlink("job.tex");
err4:
	chdir("..");
err5:
	rmdir(tmpdir);
err6:
	chdir(olddir);
err7:
	if (!ret)
	{
		garr_destroy(process->glyphs);
		xfree(process);
		if (cnt < 0)
			trans_fail(dvi_err_msgbuf);
		else
			trans_fail("Error processing TeX source.");
	}

	struct geom_rectangle bbox;
	bbox.ll.x = bbox.ll.y = INFINITY;
	bbox.ur.x = bbox.ur.y = -INFINITY;
	for (uns i = 0; i<process->glyphs_cnt; i++) {
		struct tex_glyph *gl = (struct tex_glyph*) garr_get(process->glyphs, i);
		bbox.ur.x = MAX(bbox.ur.x, gl->xpos + gl->bbox.ur.x);
		bbox.ur.y = MAX(bbox.ur.y, gl->ypos + gl->bbox.ur.y);
		bbox.ll.x = MIN(bbox.ll.x, gl->xpos + gl->bbox.ll.x);
		bbox.ll.y = MIN(bbox.ll.y, gl->ypos + gl->bbox.ll.y);
	}
	if (!process->glyphs_cnt)
		bzero(&bbox, sizeof(bbox));
	process->bbox = bbox;
	process->text = string_ref(text);
	return tex_process_ref(process);
}

// Property process is internal only. It cannot be found in property list.
PROP_ALL_WITHOUT_UNIT(process, o_prop_name, PT_POINTER, PTP_TEX_PROCESS, process, v,
		(BUG(), 0), (("Die"), ("Bug, fix me.")), CHANGED, LOW, 0, UINT_MAX)

PROP_INVISIBLE PROP_FILL_COLOR
PROP_ALL_WITHOUT_UNIT(transform, PSTRUCT(fill_color), PT_POINTER, PTP_TRANSFORM, transformation, t,
		1, (("Cannot fail."), ("Bug, fix me.")), CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL_WITHOUT_UNIT(alignment_y, PSTRUCT(transform), PT_UNS, PTU_ALIGNMENT_Y, alignment_y, u,
		value.u < PTAY_MAX, (("Invalid value."), ("Alignment Y cannot be %d.", value.u)), CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL_WITHOUT_UNIT(alignment_x, PSTRUCT(alignment_y), PT_UNS, PTU_ALIGNMENT_X, alignment_x, u,
		value.u < PTAX_MAX, (("Invalid value."), ("Alignment X cannot be %d.", value.u)), CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL(apos_y, PSTRUCT(alignment_x), PT_REAL, PTR_COORDINATE, apos.y, r,
		1, (("Cannot fail."), ("Bug, fix me.")), CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL(apos_x, PSTRUCT(apos_y), PT_REAL, PTR_COORDINATE, apos.x, r,
		1, (("Cannot fail."), ("Bug, fix me.")), CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL(rpos_y, PSTRUCT(apos_x), PT_REAL, PTR_REFERENCE, rpos.y, r,
      is_rpos_valid(value.r), (("Invalid value."), ("rpos_y cannot be %f.", value.r)), CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL(rpos_x, PSTRUCT(rpos_y), PT_REAL, PTR_REFERENCE, rpos.x, r,
      is_rpos_valid(value.r), (("Invalid value."), ("rpos_x cannot be %f.", value.r)), CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL_REHANG(coordinate, PSTRUCT(rpos_x), geom.point, 0, "Hanger must be a mouse click to change position.", 0)

static struct prop_virtual PSTRUCT(text);

static void PSET(text) (struct o *o, prop_value text)
{
	ASSERT_GO_SUBTYPE_ID(o);
	struct go_tex_text *t = (struct go_tex_text*)o;
	struct tex_process *process = tex_process(text.s);
	trans_new_undo_prop_virtual(o, &PSTRUCT(process), (prop_value){ .v = t->process });
	t->process = process;
	tsort_dirty_recompute(&t->go);
}

PROP_SET_LOW_DIE(text, "warning: ISO C99 requires rest arguments to be used")
PROP_GET(text, process->text, s)
PROP_UNIT_GET_DIE(text)
PROP_UNIT_SET_DIE(text)
PROP_STRUCT(text, PSTRUCT(coordinate_x), PT_STRING, PTS_LARGE_TEXT, 0, UINT_MAX)

PROP_VIRTUAL(&PSTRUCT(text), string_text->text,
		&PSTRUCT(coordinate_x), string_coordinate_x->text, &PSTRUCT(coordinate_y), string_coordinate_y->text,
		&PSTRUCT(rpos_x), string_rpos_x->text, &PSTRUCT(rpos_y), string_rpos_y->text,
		&PSTRUCT(apos_x), string_apos_x->text, &PSTRUCT(apos_y), string_apos_y->text, &PSTRUCT(transform), string_transform->text,
		&PSTRUCT(alignment_x), string_alignment_x->text, &PSTRUCT(alignment_y), string_alignment_y->text,
		&PSTRUCT(process), "tex-process", PROP_INVISIBLE_INIT, PROP_FILL_COLOR_INIT)


INFO_ANCHOR_1("point", QT_POINT, LT_AFFINITY)
INFO_HANGER_0


void go_tex_text_recompute(struct go_tex_text *t, struct geom_rectangle *bbox)
{
	t->geom.point = hanger_get_position(t->a->h);

	// Doplneni transformacni matice a spocitani bboxu textu v tlo souradnicich.
	get_text_bbox(&t->process->bbox, &t->geom.point, &t->rpos, &t->apos, t->transformation, t->process->bbox.ur.x, t->alignment_x, t->alignment_y, &t->text_bbox);

	*bbox = t->text_bbox;
	geom_rectangle_merge_point(bbox, &t->geom.point, bbox);
	geom_item_after_change(&t->geom);
}

void go_tex_text_transform(struct go_tex_text *tex, struct geom_transform *tr, struct geom_rectangle *bbox)
{
	struct geom_transform *t = tex->transformation;
	tex->transformation = xmalloc(sizeof(struct geom_transform));
	trans_new_undo_prop_virtual((struct o*)tex, &PSTRUCT(transform), (prop_value){.t = t});
	if(geom_transform_merge(t, tr, tex->transformation) != 0)
		TRANS_FAIL(("Numeric error."), ("geom_transform_merge fail"));
	tex->transformation->coef[0][2] = tex->transformation->coef[1][2] = 0;
	go_tex_text_recompute(tex, bbox);
}

struct go_tex_text *go_tex_text_new_process(struct hanger *refpoint, real rpos_x, real rpos_y,
		real apos_x, real apos_y, struct tex_process *process)
{
	struct go_tex_text *t = xmalloc_zero(sizeof(struct go_tex_text));

	go_init(&t->go, GOT_TEX_TEXT, GOST_TEX_TEXT, refpoint);
	t->process = process;
	t->rpos = (struct geom_point) { rpos_x, rpos_y };
	t->apos = (struct geom_point) { apos_x, apos_y };
	t->transformation = xmalloc(sizeof(struct geom_transform));
	*t->transformation = geom_transform_identity;
	t->alignment_x = PROP_INIT_TEXT_ALIGNMENT_X;
	t->alignment_y = PROP_INIT_TEXT_ALIGNMENT_Y;
	t->fill_color = PROP_INIT_TEXT_FILL_COLOR;
	t->unit_rpos_x = t->unit_rpos_y = t->unit_apos_x = t->unit_apos_y = t->unit_coordinate_x = t->unit_coordinate_y = PROP_UNIT_MAX;

	return t;
}

struct go_tex_text *go_tex_text_new(struct hanger *ll, real refp_x, real refp_y, real marg_x, real marg_y, string text)
{
	return go_tex_text_new_process(ll, refp_x, refp_y, marg_x, marg_y, tex_process(text));
}

INFO_GO_1("tex-text",
		TEX, "tex-text", 0, -1, -1, -1, -1, 1, -1, -1, text)

void go_tex_text_rerun_tex_process(struct go_tex_text *t)
{
	ASSERT_GO_SUBTYPE_ID(t);
	ASSERT(go_get_tlo(&t->go) == trans_tlo);
	struct tex_process *process = tex_process(t->process->text);
	trans_new_undo_prop_virtual(void2o(t), &PSTRUCT(process), (prop_value){ .v = t->process });
	t->process = process;
	tsort_dirty_recompute(&t->go);
}

void group_rerun_tex_process(struct go_group *group)
{
	ASSERT_GROUP(group);
	ASSERT(go_get_tlo(&group->go) == trans_tlo);
	struct go *go;

	TSORT_NO_RECOMPUTE_BEGIN
		WALK_GO_RECURSIVE(&group->go, go)
			if(go->type == GOT_TEX_TEXT)
				go_tex_text_rerun_tex_process((struct go_tex_text*)go);
	TSORT_NO_RECOMPUTE_END
}

	/* FIXME:

	  Chybi after_change, jestli na to ma nekdo umet volat exoticke geomlibovke fce, ktere namaji predefinovanou virtualni fci u bodu
	  ale spis by se to ani jako geom_point_item chovat nemuselo... staci geom_item.

	  Bbox se uz nak pocita, ale jeste chybi nekolik veci.


	  Definice pozice textu:

	  Text, co vyleze z Tomova renderitka ma bbox kolem referencniho bodu textu (ten je v pocatku). Referencni bod "Ahoj" je levy dolni roh "A"...
	  levy dolni bod Tomova bboxu muze byt kdekoli (napr. "j" jde pod referencni bod).

	  Pak go_text obsahuje 2 dalsi informace: relativni a absolutni posun.
	  Relativni znamena posunuti o nasobek sirky bboxu a absolutni o konstantu. Analogycky pro Y.

	  Na vysledny posunuty text se dal aplikuje libovolna linearni transformace (geom_transform s vynulovanym posunem).

	  A nakonec se k souradnicim textu pricte souradnice anchoru.

	  Celkove je to afine transformovatelne - staci ztransformovat anchor a linearni transformaci pronasobit s 2x2 podmatici transformace.


	  2 realy by se daly usetrit, kdyby v geom_transform nebyly 2 nuly (napr. by se tam ulozily souradnice anchoru,
	  aby se lin. transformace + posun do anchoru delalo jednim volani geom_xxx_transform).
	  Mozna by to ale znamenalo na nekterych mistech nevolat primo geomlibovske fce na transformaci, protoze by tam ty nenuly prekazely.


	  Co chce vykreslovatko:

	  Jednak linearni transformaci a jednak plne ztransformovany levy dolni bod Tomova bboxu
	  (absolutne a relativne posunuty, linearne ztransformovany a posunuty o anchor).
	  Aspon jestli se to nemenilo...

	  Tenhle bod se ted nikde nepocita, protoze nevim, kam by to melo patrit... a vlastne ani kam to ulozit...


	  Hangery textu:

	  Kdysi jsem psal hangery pro rohy textu, buhvi kam to zmizelo :-(

	  Ty by se spocitaly z Tomova bboxu (pripadne by se dal zvetsit o dalsi property "margin"... aktualni margin funguje jako ten absolutni posun),
	  na ktery by se pouzil absolutni a relativni posun, linearni transformace a posun do anchoru.


	  Pavel

	  */

/*
 * Dopocita posledni sloupec transformacni matice, ktera prevadi body ze souradnicoveho systemu textu do souradnicoveho systemu TLO:
 *
 *   T(0,0) = referencni bod textu v souradnicovem systemu TLO = (trans->coef[0][2], trans->coef[1][2]) = bod, kam vykreslovatko vykresluje
 *   T(text_box->ll) = levy dolni roh bitmapy v -||-
 *   T(text_box->ur) = pravy horni roh bitmapy v -||-
 *   ...
 *
 * Heureka, a uz je vyuziti pro vsechny realy matice transformace.
 *
 * Tahle fce by se mela volat po kazde zmene property textu/TeXtu.
 * Pak uz by melo stacit jen zavolat hook pro vykreslovatko, ktere si muze referencni bod textu zjistit z matice)
 * a updatovat rtree.
 *
 *
 * Vyznam parametru:
 *   trans - staci vyplnena 2x2 podmatice s linearni transformaci textu
 *   text_box - bbox textu z Tomovy struktury
 *   ref_width - sirka textu = vzdalenost mezi dvojici referencnich bodu textu (Tom vi o co de)...
 *               dokud se to nijak nepocita, je vhodne sem zadat text_box->ur.x (pravy referencni bod = pravy okraj bboxu)
 *               FIXME: upravit v go_text_recompute
 *   rpos - relativni pozice textu (x/y_refpnt... bylo by prehlednejsi to dat do jednoho geom_point a prejemenovat rel_pos)
 *   apos - absolutni posunuti textu (x/y_margin... to bych prejmenoval na abs_pos)
 *   anchor - souradnice anchoru
 *   flags - nastaveni zarovnani textu:
 *     TEXT_BASELINE - ignoruje se relativni posun v Y a anchor se umisti na baselinu textu
 *     TEXT_LEFT_REF_PNT - pri X-ovem relativnim posunu se pocita s levym referencnim bodem... jinak s levym okrajem bboxu
 *     TEXT_RIGHT_REF_PNT - pri X-ovem relativnim posunu se pocita s pravym referencnim bodem... jinak s pravym okrajem bboxu
 *
 * Minimalne ten flag BASELINE by mel jit nastavit pri editaci textu (property)... ty ostatni bud editovatelne nebo je nak
 * pevne nastavit... ale to by melo jit kdykoli zmenit.
 *
 * (Ne)funkcnost se da vizualne ozkouset na pocitani bboxu...
 *
 */

