/*
 * Kernel - Referenced strings
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "kernel/internal.h"

#ifdef DEBUG_REF_STRING
#define DBG_STRING(...) msg(__VA_ARGS__)
#else
#define DBG_STRING(...) do { } while(0)
#endif


/*
 * Strings are stored in following hash table.
 */

#define HASH_NODE struct string_entry
#define HASH_PREFIX(x) hts_##x
#define HASH_KEY_ENDSTRING text
#define HASH_WANT_CLEANUP
#define HASH_WANT_FIND
#define HASH_WANT_NEW
#define HASH_WANT_REMOVE
#define HASH_GIVE_INIT_KEY
static inline void hts_init_key(string s, char *str) { strcpy(s->text, str); s->length = strlen(str); }
#include "lib/sht.h"


static struct string_entry dead_list = { &dead_list, 0, 0, {'a'}};


#define REF(_name, _string) string_##_name = string_lookup_ref(_string)

void string_init(void)
{
	hts_init();

	REF(name, "name");
	o_prop_name.p.key = string_name;
	REF(noname, "noname");
	REF(filename, "filename");
	REF(file_unsaved, "unsaved");

	REF(obj_type[OT_UNIVERSE], "universum");
	REF(obj_type[OT_DOCUMENT], "document");
	REF(obj_type[OT_TLO], "page");
	REF(obj_type[OT_ZOMBIE], "zombie");
	REF(obj_type[OT_TEMP], "temp");

	REF(quantity_type[QT_POINT], "point");
	REF(quantity_type[QT_CURVE], "curve");
	
	REF(prop_type[PT_UNS], "uns");
	REF(prop_type[PT_POINTER], "pointer");
	REF(prop_type[PT_STRING], "string");
	REF(prop_type[PT_REAL], "real");

	REF(conic_type[GEOM_CONIC_ARC_START_ENTIRE], "start-entire");
	REF(conic_type[GEOM_CONIC_ARC_START_DIF], "start-dif");
	REF(conic_type[GEOM_CONIC_ARC_POINT_ENTIRE], "point-entire");
	REF(conic_type[GEOM_CONIC_ARC_POINT_DIF], "point-dif");
	REF(conic_type[GEOM_CONIC_ARC_CCW], "ccw");
	REF(conic_type[GEOM_CONIC_ARC_CW], "cw");
	REF(conic_type[GEOM_CONIC_ARC_SMALLER], "smaller");
	REF(conic_type[GEOM_CONIC_ARC_BIGGER], "bigger");
	REF(conic_type[GEOM_CONIC_ARC_MIDDLE], "middle");
	REF(conic_type[GEOM_CONIC_ARC_OPPOSITE], "opposite");
	REF(conic_type[GEOM_CONIC_ARC_ENTIRE], "entire");

	REF(stroke_cap_type[PSC_BUTT], "butt");
	REF(stroke_cap_type[PSC_ROUND], "round");
	REF(stroke_cap_type[PSC_PROJECTING], "projecting");

	REF(stroke_join_type[PSJ_MITER], "miter");
	REF(stroke_join_type[PSJ_ROUND], "round");
	REF(stroke_join_type[PSJ_BEVEL], "bevel");

	REF(stroke_style_type[PSS_SOLID], "solid");
	REF(stroke_style_type[PSS_DASHED], "dashed");
	REF(stroke_style_type[PSS_DOTTED], "dotted");
	REF(stroke_style_type[PSS_DASH_DOTTED], "dash-dotted");

	REF(alignment_x_type[PTAX_REF_LEFT], "refpoints-left");
	REF(alignment_x_type[PTAX_REF_CENTER], "refpoints-center");
	REF(alignment_x_type[PTAX_REF_RIGHT], "refpoints-right");
	REF(alignment_x_type[PTAX_REF_RELATIVE], "refpoints-relative");
	REF(alignment_x_type[PTAX_BBOX_LEFT], "bbox-left");
	REF(alignment_x_type[PTAX_BBOX_CENTER], "bbox-center");
	REF(alignment_x_type[PTAX_BBOX_RIGHT], "bbox-right");
	REF(alignment_x_type[PTAX_BBOX_RELATIVE], "bbox-relative");
	
	REF(alignment_y_type[PTAY_BASELINE], "baseline");
	REF(alignment_y_type[PTAY_BBOX_BOTTOM], "bbox-bottom");
	REF(alignment_y_type[PTAY_BBOX_CENTER], "bbox-center");
	REF(alignment_y_type[PTAY_BBOX_TOP], "bbox-top");
	REF(alignment_y_type[PTAY_BBOX_RELATIVE], "bbox-relative");

	REF(arrow_align_type[PTAA_FRONT], "front");
	REF(arrow_align_type[PTAA_BACK], "back");

	REF(arrow_front_type[PTAF_STRAIGHT], "straight");
	REF(arrow_front_type[PTAF_PARABOLIC], "parabolic");

	REF(arrow_back_type[PTAB_NONE], "none");
	REF(arrow_back_type[PTAB_STRAIGHT], "straight");
	REF(arrow_back_type[PTAB_POLY], "polygonal");
	REF(arrow_back_type[PTAB_PARABOLIC], "parabolic");
	
	REF(coordinate_x, "coordinate-x");
	REF(coordinate_y, "coordinate-y");

	REF(invisible, "invisible");
	REF(fill_color, "fill-color");
	REF(stroke_color, "stroke-color");
	REF(stroke_width, "stroke-width");
	REF(stroke_cap, "stroke-cap");
	REF(stroke_join, "stroke-join");
	REF(stroke_style, "stroke-style");

	REF(text, "text");
	REF(font, "font");
	REF(font_size, "size");
	REF(rpos_x, "relative-position-x");
	REF(rpos_y, "relative-position-y");
	REF(apos_x, "absolute-shift-x");
	REF(apos_y, "absolute-shift-y");
	REF(alignment_x, "alignment-x");
	REF(alignment_y, "alignment-y");
	REF(tex_header, "tex_header");
	REF(tex_footer, "tex_footer");
	REF(tex_header_default, "\\nopagenumbers\\vglue0pt\n");
	REF(tex_footer_default, "\\bye\n");

	REF(transform, "transform");
	
	REF(max_history_volume, "max-undo-history-volume");
	REF(max_history_count, "max-undo-history-count");
	REF(max_cache_memory, "max-cache-memory");
	REF(max_cache_count, "max-cache-count");
	REF(min_cache_count, "min-cache-count");
}

static void string_array_unref(string *array, int size)
{
	for(int i = 0; i < size; i ++)
		string_unref(array[i]);
}

void string_finish(void)
{
	string_array_unref(string_obj_type, OT_MAX);
	string_array_unref(string_go_type, GOT_MAX);
	string_array_unref(string_conic_type, GEOM_CONIC_ARC_MAX);
	string_array_unref(string_stroke_cap_type, PSC_MAX);
	string_array_unref(string_stroke_join_type, PSJ_MAX);
	string_array_unref(string_stroke_style_type, PSS_MAX);
	string_array_unref(string_prop_type, PT_MAX);
	string_array_unref(string_quantity_type, QT_MAX);
	string_array_unref(string_alignment_x_type, PTAX_MAX);
	string_array_unref(string_alignment_y_type, PTAY_MAX);
	string_array_unref(string_arrow_front_type, PTAF_MAX);
	string_array_unref(string_arrow_back_type, PTAB_MAX);
	string_array_unref(string_arrow_align_type, PTAA_MAX);

	string array[] = { NAMES_STRINGS };
	for(int i = sizeof(array) / sizeof(string) - 1; i >= 0; i--)
		string_unref(array[i]);

	string_cleanup();

#ifdef DEBUG
	HASH_FOR_ALL(hts, walk)
	{
		DANGEROUS(0, "There are %d references to the string %s.\n", walk->ref_count, walk->text);
	}
	HASH_END_FOR;
#endif

	hts_cleanup();
}

/**
 * Cleanup the dead list.
 * Function is called in the end of transaction.
 */
void string_cleanup(void)
{
	string s;
	while(s = dead_list.dead_next, s != &dead_list)
	{
		dead_list.dead_next = s->dead_next;
		if(s->ref_count)
			s->dead_next = NULL;
		else
		{
			DBG_STRING("  Remove string %s at %p\n", s->text, s);
			hts_remove(s);
		}
	}
}

#undef string_ref
string string_ref(string s DEBUG_REF_STRING_ARGS)
{
	s->ref_count++;
	DBG_STRING("  Increase reference: string %s at %p, new ref_count %d, file %s, line %d, fce %s\n", s->text, s, s->ref_count, file, line, fce);
	return s;
}

#undef string_unref
string string_unref(string s DEBUG_REF_STRING_ARGS)
{
	ASSERT(s->ref_count > 0);
	s->ref_count--;
	DBG_STRING("  Decrease reference: string %s at %p, new ref_count %d, file %s, line %d, fce %s\n", s->text, s, s->ref_count, file, line, fce);

	if(s->ref_count > 0 || s->dead_next)
		return s;

	s->dead_next = dead_list.dead_next;
	dead_list.dead_next = s;
	return s;
}

#undef string_lookup
/** Lookup for given string and create it if not exists. No new reference is increased. */
string string_lookup(const char* str DEBUG_REF_STRING_ARGS)
{
	string s = hts_find((char*)str);
	if(!s)
	{
		s = hts_new((char*)str);
		s->ref_count = 0;
		s->dead_next = dead_list.dead_next;
		dead_list.dead_next = s;
	}
	DBG_STRING("  String lookup: string %s at %p, new ref_count %d, file %s, line %d, fce %s\n", s->text, s, s->ref_count, file, line, fce);
	return s;
}

#undef string_lookup_ref
/** Lookup for given string and create it if not exists. Increase one new reference. */
string string_lookup_ref(const char* str DEBUG_REF_STRING_ARGS)
{
	string s = hts_find((char*)str);
	if(s)
		s->ref_count++;
	else
	{
		s = hts_new((char*)str);
		s->ref_count = 1;
		s->dead_next = NULL;
	}
	DBG_STRING("  Increase reference: string %s at %p, new ref_count %d, file %s, line %d, fce %s\n", s->text, s, s->ref_count, file, line, fce);
	return s;
}

/** Lookup for given string and return NULL if not exists. */
string string_find(const char* str)
{
	return hts_find((char*)str);
}

