#ifndef _KERNEL_STRINGS_H_
#define _KERNEL_STRINGS_H_

/*
 * Kernel - Referenced strings
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

/* This is used for debugging. */
#define DEBUG_ARGS , const char* file, int line, const char* fce
#ifndef DEBUG_REF_STRING
#define DEBUG_REF_STRING_ARGS
#else
#define DEBUG_REF_STRING_ARGS DEBUG_ARGS
#endif



/**
 * Structure for referenced string
 */

typedef struct string_entry
{
	struct string_entry *dead_next;
	uns ref_count;
	uns length;
	char text[1];
} *string;


/**
 * String handling functins.
 *
 * Macro DEBUG_REF_STRING_ARGS is used for debugging. You can ignore it.
 */

string string_ref(string str DEBUG_REF_STRING_ARGS);
string string_unref(string str DEBUG_REF_STRING_ARGS);
string string_lookup(const char* str DEBUG_REF_STRING_ARGS);
string string_lookup_ref(const char* str DEBUG_REF_STRING_ARGS);
string string_find(const char* str);

	
/**
 * Those strings are used as constants.
 */
#define NAMES_STRINGS \
	string_name, string_noname, \
	string_filename, string_file_unsaved, \
	string_coordinate_x, string_coordinate_y, \
	string_invisible, string_fill_color, string_stroke_color, \
	string_stroke_width, string_stroke_join, string_stroke_cap, string_stroke_style, \
	string_text, string_font, string_font_size, \
	string_rpos_x, string_rpos_y, string_apos_x, string_apos_y, \
	string_alignment_x, string_alignment_y, \
	string_tex_header, string_tex_footer, string_tex_header_default, string_tex_footer_default, \
	string_transform, \
	string_max_history_volume, string_max_history_count, \
	string_max_cache_memory, string_max_cache_count, string_min_cache_count

extern string NAMES_STRINGS,
	string_prop_type[],
	string_go_type[], string_obj_type[],
	string_quantity_type[],
	string_conic_type[], string_stroke_cap_type[], string_stroke_join_type[],
	string_stroke_style_type[],
	string_alignment_x_type[], string_alignment_y_type[],
	string_arrow_front_type[], string_arrow_back_type[], string_arrow_align_type[];


/**
 * Initializating and cleanup functions.
 */

void string_init(void);
void string_finish(void);
void string_cleanup(void);


/* This is used for debugging. */
#ifdef DEBUG_REF_STRING
#define string_ref(s) string_ref(s, __FILE__, __LINE__, __func__)
#define string_unref(s) string_unref(s, __FILE__, __LINE__, __func__)
#define string_lookup(s) string_lookup(s, __FILE__, __LINE__, __func__)
#define string_lookup_ref(s) string_lookup_ref(s, __FILE__, __LINE__, __func__)
#endif

#endif
