/*
 * VRRLIB - Cache.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "lib/lib.h"
#include <string.h>
#include "lib/cache.h"


/* Global cache. */

struct cache global_cache;


/* cache_init */

void cache_init(struct cache *cache)
{
	ASSERT(cache);

	memset(cache, 0, sizeof(struct cache));
	cache->memory_limit = cache->memory_free = CACHE_DEFAULT_MEMORY_LIMIT;
	cache->count_min = CACHE_DEFAULT_COUNT_MIN;
	cache->count_max = CACHE_DEFAULT_COUNT_MAX;
	clist_init(&cache->lru_list);
}


/* cache_destroy */

void cache_destroy(struct cache *cache)
{
	uns i;
	
	ASSERT(cache);
	
	cache_remove_all(cache);
	for (i = 0; i < CACHE_MAX_USERS; i++)
		if (cache->users[i])
			cache_user_destroy(cache->users[i], cache);
#ifdef DEBUG
	if ((uns)cache->memory_free != cache->memory_limit)
		die("! Cache memory usage info is not zero while destroying cache.\n"
			"There are propably invalid calls to cache_used_inc() and cache_used_dec().\n");
#endif
}


/* cache_set_limits */

void cache_set_limits(struct cache *cache, uns memory_limit, uns count_min, uns count_max)
{
	ASSERT(cache != NULL);
	if (memory_limit != CACHE_LEAVE_MEMORY_LIMIT) {
		cache->memory_free += memory_limit - cache->memory_limit;
		cache->memory_limit = memory_limit;
	}
	if (count_min != CACHE_LEAVE_COUNT_MIN)
		cache->count_min = count_min;
	if (count_max != CACHE_LEAVE_COUNT_MAX)
		cache->count_max = count_max;
	ASSERT(count_min <= count_max);
	while (cache->count > cache->count_max)
		cache_remove_lru(cache);
	cache_used_update(cache);
}


/* cache_user_init */

void cache_user_init(struct cache_user *user, struct cache *cache, void (*callback)(struct cache_node *node))
{
	uns i;

	ASSERT(cache != NULL);
	ASSERT(user != NULL);

	/* Zero id is unused for better debugging. */
	for (i = 1; i < CACHE_MAX_USERS; i++)
		if (!cache->users[i]) {
			user->user_id = i;
			user->callback = callback;
			cache->users[i] = user;
			return;
		}

	die("No unused id for new cache user.\n");
}


/* cache_user_destroy */

void cache_user_destroy(struct cache_user *user, struct cache *cache)
{
	ASSERT(user != NULL);
	ASSERT(cache != NULL);
	ASSERT(user->user_id < CACHE_MAX_USERS && cache->users[user->user_id] == user);

	cache->users[user->user_id] = NULL;
}


