/*
 *	Sherlock Library -- Circular Linked Lists
 *
 *	(c) 2003 Martin Mares <mj@ucw.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU Lesser General Public License.
 *
 *
 * Modification -- Counted Circular Linked Lists
 *
 * (c) 2004 Jirka Fink <fink@atrey.karlin.mff.cuni.cz>
 */

#ifndef _SHERLOCK_CCLISTS_H
#define _SHERLOCK_CCLISTS_H

typedef struct ccnode {
  struct ccnode *next, *prev;
} ccnode;

typedef struct cclist {
  struct ccnode head;
  uns count;
} cclist;

static inline void *cclist_head(cclist *l)
{
  return (l->head.next != &l->head) ? l->head.next : NULL;
}

static inline void *cclist_tail(cclist *l)
{
  return (l->head.prev != &l->head) ? l->head.prev : NULL;
}

static inline void *cclist_next(cclist *l, ccnode *n)
{
  return (n->next != &l->head) ? (void *) n->next : NULL;
}

static inline void *cclist_prev(cclist *l, ccnode *n)
{
  return (n->prev != &l->head) ? (void *) n->prev : NULL;
}

static inline int cclist_empty(cclist *l)
{
  return (l->head.next == &l->head);
}

#define CCLIST_WALK(n,list) for(n=(void*)(list).head.next; (ccnode*)(n) != &(list).head; n=(void*)((ccnode*)(n))->next)
#define CCLIST_WALK_DELSAFE(n,list,tmp) for(n=(void*)(list).head.next; tmp=(void*)((ccnode*)(n))->next, (ccnode*)(n) != &(list).head; n=(void*)tmp)

static inline void cclist_insert_after(cclist *l, ccnode *what, ccnode *after)
{
  ccnode *before = after->next;
  what->next = before;
  what->prev = after;
  before->prev = what;
  after->next = what;
  l->count++;
}

static inline void cclist_insert_before(cclist *l, ccnode *what, ccnode *before)
{
  ccnode *after = before->prev;
  what->next = before;
  what->prev = after;
  before->prev = what;
  after->next = what;
  l->count++;
}

static inline void cclist_add_tail(cclist *l, ccnode *n)
{
  cclist_insert_before(l, n, &l->head);
}

static inline void cclist_add_head(cclist *l, ccnode *n)
{
  cclist_insert_after(l, n, &l->head);
}

static inline void cclist_remove(cclist *l, ccnode *n)
{
  ccnode *before = n->prev;
  ccnode *after = n->next;
  before->next = after;
  after->prev = before;
  l->count--;
}

static inline void cclist_init(cclist *l)
{
  ccnode *head = &l->head;
  head->next = head->prev = head;
  l->count = 0;
}

#endif
