
#ifndef _LIB_FLISTS_H_
#define _LIB_FLISTS_H_

/*
 * VRRLIB - Fifo unidirectional linked lists
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "slists.h"

typedef struct flist
{
  struct snode head;
  struct snode *tail;
} flist;

static inline int flist_empty(flist *l)
{
	return l->head.next == &l->head;
}

static inline void* flist_head(flist *l)
{
	return flist_empty(l) ? NULL : l->head.next;
}

static inline void* flist_remove_head(flist *l)
{
	if(flist_empty(l))
		return NULL;
	struct snode *s = l->head.next;
	l->head.next = l->head.next->next;
	s->next = NULL;
	if(flist_empty(l))
		l->tail = &l->head;
	return s;
}

static inline void flist_remove_after(flist *l, snode *n)
{
	ASSERT(n->next != &l->head);
	struct snode *s = n->next;
	n->next = n->next->next;
	s->next = NULL;
	if(n->next == &l->head)
		l->tail = n;
}

static inline snode* flist_prev(flist *l, snode *n)
{
	for(snode *p = &l->head; p->next != &l->head; p = p->next)
		if(p->next == n)
			return p;
	BUG();
}

static inline void flist_remove(flist *l, snode *n)
{
	flist_remove_after(l, flist_prev(l, n));
}

static inline void flist_add_tail(flist *l, snode *n)
{
	l->tail->next = n;
	l->tail = n;
	n->next = &l->head;
}

static inline void flist_init(flist *l)
{
	l->head.next = &l->head;
	l->tail = &l->head;
}

static inline int flist_is_initial(flist *l)
{
	return l->head.next == &l->head && l->tail == &l->head;
}

#endif
