/*
 * VRRLIB - Global cache header
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "lib/cache.h"

#ifndef CACHE_PREFIX
#define CACHE_PREFIX(x) gcache_##x
#endif

#ifndef CACHE_STRUCT
#define CACHE_STRUCT global_cache
#endif

#ifndef CACHE_USER
#define CACHE_USER CACHE_PREFIX(user)
#endif

#ifndef CACHE_TYPE
#define CACHE_TYPE struct CACHE_PREFIX(node)
#endif

#ifndef CACHE_CALLBACK
#define CACHE_CALLBACK CACHE_PREFIX(callback)
#endif

static inline void CACHE_PREFIX(init) (void)
{ cache_user_init(&(CACHE_USER), &(CACHE_STRUCT), (void (*)(struct cache_node *))&(CACHE_CALLBACK)); }

static inline void CACHE_PREFIX(destroy) (void)
{ cache_user_destroy(&(CACHE_USER), &(CACHE_STRUCT)); }

static inline void CACHE_PREFIX(add) (CACHE_TYPE *node)
{
	cache_add(&(CACHE_STRUCT), &(CACHE_USER), (struct cache_node *)node);
#ifdef CACHE_NODE_SIZE
	cache_used_inc(&(CACHE_STRUCT), CACHE_NODE_SIZE(node));
#endif	
}

static inline void CACHE_PREFIX(remove) (CACHE_TYPE *node)
{
	cache_remove(&(CACHE_STRUCT), (struct cache_node *)node);
#ifdef CACHE_NODE_SIZE
	cache_used_dec(&(CACHE_STRUCT), CACHE_NODE_SIZE(node));
#endif	
}

static inline void CACHE_PREFIX(touch) (CACHE_TYPE *node)
{ cache_touch(&(CACHE_STRUCT), (struct cache_node *)node); }

static inline void CACHE_PREFIX(used_inc) (uns size)
{ cache_used_inc(&(CACHE_STRUCT), size); }

static inline void CACHE_PREFIX(used_dec) (uns size)
{ cache_used_dec(&(CACHE_STRUCT), size); }

static inline void CACHE_PREFIX(used_mod) (int size)
{ cache_used_mod(&(CACHE_STRUCT), size); }

#ifdef CACHE_NODE_SIZE
static inline void CACHE_PREFIX(node_inc) (CACHE_TYPE UNUSED *node)
{ cache_used_inc(&(CACHE_STRUCT), CACHE_NODE_SIZE(node)); }

static inline void CACHE_PREFIX(node_dec) (CACHE_TYPE UNUSED *node)
{ cache_used_dec(&(CACHE_STRUCT), CACHE_NODE_SIZE(node)); }
#endif

#undef CACHE_PREFIX
#undef CACHE_STRUCT
#undef CACHE_USER
#undef CACHE_TYPE
#undef CACHE_NODE_SIZE
#undef CACHE_CALLBACK

