#ifndef _VRR_LIB_LIB_H
#define _VRR_LIB_LIB_H

/*
 * VRRLIB - Main project header
 *
 * Copyright (c) 2004-2007    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#define _BSD_SOURCE
#define _POSIX_C_SOURCE 199506L
#define _XOPEN_SOURCE 600

#include "build/config.h"

#include <stdarg.h>
#include <stdint.h>

#include "build/types.h"

/* used floating-point type */
#ifdef FLOAT_REAL
#define SCN_REAL "%f"
typedef float real;
#else
#define SCN_REAL "%lf"
typedef double real;
#endif

#ifndef NULL
#define NULL (void *)0
#endif


/* Useful macros */

#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))
#define CLAMP(x,min,max) ({ int _t=x; (_t < min) ? min : (_t > max) ? max : _t; })
#define ABS(x) ((x) < 0 ? -(x) : (x))
#define ARRAY_SIZE(a) (sizeof(a)/sizeof(*(a)))
#define GLUE(a,b) a##b
#define GLUE_(a,b) a##_##b
#define GLUEE(a,b) GLUE(a,b)
#define GLUE3(a,b,c) a##b##c
#define GLUE_3(a,b,c) a##_##b##_##c
#define GLUEE3(a,b,c) GLUE3(a,b,c)
#define GLUE4(a,b,c,d) a##b##c##d
#define GLUE_4(a,b,c,d) a##_##b##_##c##_##d
#define GLUEE4(a,b,c,d) GLUE4(a,b,c,d)
#define XMKSTR(a) VMKSTR(a)
#define VMKSTR(...) #__VA_ARGS__
#define MKSTR(a) # a

#define BITCOUNT(x)	(((BX_(x)+(BX_(x)>>4)) & 0x0F0F0F0F) % 255)
#define  BX_(x)		((x) - (((x)>>1)&0x77777777)	\
			- (((x)>>2)&0x33333333)		\
			- (((x)>>3)&0x11111111))


/* Structures handling macros */

#define OFFSETOF(s, i) ((int)&((s *)0)->i)
#define SKIP_BACK(s, i, p) ((s *)((char *)p - OFFSETOF(s, i)))
#define SKIP_OFFSET(s, p, o) ((s *)((char *)p + (o)))
#define ALIGN(s, a) (((s)+a-1)&~(a-1))
#define UNALIGNED_PART(ptr, type) (((long) (ptr)) % sizeof(type))


/* GCC specific attributes */
#define WARN_UNUSED_RESULT
#ifdef __GNUC__
#define NONRET __attribute__((noreturn))	/* Function doesn't return */
#define UNUSED __attribute__((unused))		/* Unused function argument */
#define PACKED __attribute__((packed))		/* Smallest possible structure alignment */
#define CONST __attribute__((const))		/* Function doesn't look outside its body */
#define PURE __attribute__((const))

#if __GNUC__ >= 4 || __GNUC__ == 3 && __GNUC_MINOR__ >= 4
#undef WARN_UNUSED_RESULT
#define WARN_UNUSED_RESULT __attribute__((warn_unused_result))	/* Warn if caller ignores the function's result */
#endif

#ifdef C99_PEDANTIC
#define FORMAT(_index, _first)
#else
#define FORMAT(_index, _first) __attribute__((format (printf, _index, _first)))
#endif

#define likely(x) __builtin_expect((x),1)	/* Condition is in most cases true */
#define unlikely(x) __builtin_expect((x),0)	/* dtto false */
#else
#define NONRET
#define UNUSED
#define PACKED
#define CONST
#define PURE
#define FORMAT(_index, _first)
#define likely(x) (x)
#define unlikely(x) (x)
#endif

/* Our own memory allocation */

#ifdef DEBUG_MALLOC

#define xmalloc malloc
#define xmalloc_zero(size) calloc(size, 1)
#define xfree free
#define xrealloc realloc
#define xstrdup strdup

#else

void *xmalloc(uns size);
void *xmalloc_nonfail(uns size); /* does not fail if the allocation has been unsuccesful */
void *xmalloc_zero(uns size);	/* also zeroes the buffer */
void xfree(void *);
void *xrealloc(void *, uns size);
char *xstrdup(const char* string);
char *xstrndup(const char *string, uns n);

#endif

/* Logging and debugging */

void msg(const char *, ...) FORMAT(1,2);
void err(const char *, ...) FORMAT(1,2);
void die(const char *, ...) FORMAT(1,2) NONRET;

extern int lib_stdout_fd;
extern int lib_stderr_fd;

#ifdef LOCAL_DEBUG
#define DBG(...) msg(__VA_ARGS__)
#define DBGLN(...) do{ msg("%s:%d ", __FILE__, __LINE__); msg(__VA_ARGS__); }while(0)
#else
#define DBG(...) do { } while(0)
#define DBGLN(...) do{}while(0)
#endif

#ifdef DEBUG
void assert_failed(const char *assertion, const char *file, int line) NONRET;
#define ASSERT(x) do { if (unlikely(!(x))) assert_failed(#x, __FILE__, __LINE__); } while(0)
#else
#define ASSERT(x) do{}while(0)
#endif

void internal_error(const char *file, int line) NONRET;
#define BUG() internal_error(__FILE__, __LINE__)

#endif
