/*
 * VRRLIB - Logging and debugging
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "lib/lib.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>

int lib_stdout_fd = 1;
int lib_stderr_fd = 2;

/* prints message to stderr */
static void do_log(int fd, const char *s, va_list args)
{
	char buf[512];
	int p;

	p = vsnprintf(buf, sizeof(buf), s, args);
	if (p >= (int)sizeof(buf))
		strcpy(buf, "???\n");
	write(fd, buf, MIN((int)sizeof(buf), p+1)-1);
}

/* normal log, no other action */
void msg(const char *s, ...)
{
	va_list args;

	va_start(args, s);
	do_log(lib_stdout_fd, s, args);
	va_end(args);
}

/* error log, no other action */
void err(const char *s, ...)
{
	va_list args;

	va_start(args, s);
	do_log(lib_stderr_fd, s, args);
	va_end(args);
}

/* logs s and then exits */
void NONRET die(const char *s, ...)
{
	va_list args;

	va_start(args, s);
	do_log(2, s, args);
	va_end(args);
#ifdef DEBUG
	abort();
#else
	exit(1);
#endif
}

#ifdef DEBUG
void NONRET assert_failed(const char *assertion, const char *file, int line)
{
	die("Assertion `%s' failed at %s:%d\n", assertion, file, line);
}
#endif

void NONRET internal_error(const char *file UNUSED, int line UNUSED)
{
#ifdef DEBUG
	die("Bug found at %s:%d\n", file, line);
#else
	die("Internal error\n");
#endif
}

