#ifndef _LIB_SLISTS_H_
#define _LIB_SLISTS_H_

/*
 * VRR library - simple linked lists
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

typedef struct snode {
  struct snode *next;
} snode;

typedef struct slist {
  struct snode head;
} slist;

static inline void *slist_head(slist *l)
{
  return l->head.next;
}

static inline snode *slist_remove_head(slist *l)
{
	snode *n = l->head.next;
	if(n)
		l->head.next = n->next;
	return n;
}


static inline void *slist_next(snode *n)
{
  return n->next;
}

static inline void *slist_prev(slist *l, snode *n)
{
  for(snode* p = &l->head; p; p = p->next)
	  if(p->next == n)
		  return p;
  return NULL;
}

static inline int slist_empty(slist *l)
{
  return (l->head.next == NULL);
}

#define SLIST_WALK(n,slist) for(n=(void*)(slist).head.next; n; n=(void*)((snode*)(n))->next)
#define SLIST_WALK_DELSAFE(n,slist,tmp) for(n=(void*)(slist).head.next; n && (tmp=(void*)((snode*)(n))->next, 1); n=(void*)tmp)

static inline void slist_insert_after(snode *what, snode *after)
{
  what->next = after->next;
  after->next = what;
}

static inline void slist_add(slist *l, snode *n)
{
  slist_insert_after(n, &l->head);
}

static inline void slist_remove_after(snode *what)
{
  what->next = what->next->next;
}

static inline void slist_remove(slist *l, snode *n)
{
  n = slist_prev(l, n);
  n->next = n->next->next;
}

static inline int slist_remove_safe(slist *l, snode *n)
{
  n = slist_prev(l, n);
  if(n)
  {
	  n->next = n->next->next;
	  return 1;
  }
  return 0;
}

static inline void slist_init(slist *l)
{
  l->head.next = NULL;
}

#endif
