;;; Vrr REPL and console support
;;;
;;; Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
;;; This program is distributed under GNU General Public License, see COPYING.

(define-module (vrr console))

(use-modules (vrr misc))
(use-modules (vrr low))
(use-modules (vrr support))

(use-modules (vrr gui))

;; actions: none


;;; Vrr REPL and console support

(define implicit-tlo 'active)
(define int-mod #f)

(define (set-interaction-module! mod)
  (set! int-mod mod))

(define (set-implicit-tlo! tlo)
  (if (or (not tlo)
	  (eq? tlo 'active)
	  (proxy? tlo 'obj 'page)
	  (proxy? tlo 'obj 'universum))
      (set! implicit-tlo tlo)
      (typearg-error "set-implicit-tlo!" 1 tlo)))

(define (get-implicit-tlo)
  (if (eq? implicit-tlo 'active)
      (active-page)
      implicit-tlo))

(define (make-vrr-console-port console)
  (make-soft-port
   (vector (lambda (c) (vrr-console-write-char console c))
	   (lambda (s) (vrr-console-write-string console s))
	   #f
	   (lambda () (vrr-console-read-char console))
	   #f)
   "rw"))

(define vrr-repl-prompt-1 "> ")
(define vrr-repl-prompt-2 "... ")

(define (vrr-repl-init console)
  (display vrr-repl-prompt-1 (vrr-console-get-output-port console)))

(define (vrr-repl-iter console)
  (call-with-io-ports
   (vrr-console-get-input-port console)
   (vrr-console-get-output-port console)
   (lambda () (vrr-repl-iter-int console))))

(define (display-exception . e)
  (if (= (length e) 5)
      (apply display-error (cons* #f (current-output-port) (cdr e)))
      (format #t "Error - exception: ~s\n" e))
  *unspecified*)


(define (vrr-repl-eval call)
  (if (get-implicit-tlo)
      (trans-main (get-implicit-tlo)
		  (lambda ()
		    (eval call int-mod))
		  display-exception
		  (format #f "~s" call))
      (catch #t
	     (lambda () (eval call int-mod))
	     display-exception)))

(define (vrr-repl-read console)
  (vrr-console-begin-blockread console)
  (let ((data (read)))
    (vrr-console-end-blockread console)
    data))

(define (vrr-repl-iter-int console)
  (catch 'misc-error
	 (lambda ()
	   (let loop ((data (vrr-repl-read console)))
	     (if (not (eof-object? data))
		 (let ((val (vrr-repl-eval data)))
		   (if (not (unspecified? val))
		       (begin
			 (write val)
			 (newline)))
		   (loop (vrr-repl-read console)))
		 (display vrr-repl-prompt-1))))
	 (lambda err
	   (if (and (> (length err) 2) (string=? (caddr err) "end of file"))
	       (display vrr-repl-prompt-2)
	       (begin
	         (vrr-console-end-blockread console)
		 (display "ERROR ")
		 (write err)
		 (newline)
		 (vrr-repl-iter-int console))))))

(define (call-with-io-ports input output thunk)
  (let ((ci (current-input-port))
	(co (current-output-port)))
    (set-current-input-port input)
    (set-current-output-port output)
    (thunk)
    (set-current-input-port ci)
    (set-current-output-port co)))

