;;; public (high-level) interface functions
;;;
;;; Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
;;; This program is distributed under GNU General Public License, see COPYING.

(define-module (vrr high))

(use-modules (vrr misc))
(use-modules (vrr low))
(use-modules (vrr support))
(use-modules (vrr property))
(use-modules (vrr save))
(use-modules (vrr load))

;; actions: fills tables (cons-table)


;; *************************
;; part 1 - type information
;; *************************

(re-export proxy? proxy-kind proxy-type proxy-subtype)

(define-public (obj? x) (proxy? x 'obj))
(define-public (go? x) (proxy? x 'go))
(define-public (anchor? x) (proxy? x 'anchor))
(define-public (hanger? x) (proxy? x 'hanger))



;; *****************
;; reexport
;; *****************

(re-export lookup load-doc save-doc get-property set-property! get-conic set-conic!
	   trans-main trans trans-fail undo redo universe 
	   save-as-ps save-as-eps save-as-pdf save-as-svg plugin-call load-plugin)

;; ****************************
;; part 2 - creation of objects
;; ****************************

(define (vrr-link go) (go-link-head go (tlo-get-tlg (get-trans-tlo))))


(define-public make-document doc-new)
(define-public make-page tlo-new)

(define-public (make-common-document name)
  (vrr-type-assert "make-common-doc" name 'string)
  (let ((d (make-document)))
    (set-property! d 'name name)
    (n-set-parent! d universe)
    d))

(define-public (make-common-page name doc)
  (vrr-type-assert "make-common-page" name 'string doc 'obj)
  (let ((p (make-page)))
    (set-property! p 'name name)
    (n-set-parent! p doc)
    p))


(define-public (coords x y)
  (vrr-type-assert "coords" x 'real y 'real)
  (hanger-new-mc-xy x y))

(define-public (coords-c p)
  (vrr-type-assert "coords-c" p (complex? p) 'complex-number)
  (hanger-new-mc-xy (real-part p) (imag-part p)))


(define-public (make-point h)
  (vrr-type-assert "make-point" h 'hanger)
  (vrr-link (go-point-new h)))

(define-public (make-segment h1 h2)
  (vrr-type-assert "make-segment" h1 'hanger h2 'hanger)
  (vrr-link (go-segment-new h1 h2)))


(define-public (make-ellipse-by-center-2-radii-rotation c r1 r2 rot)
  (vrr-type-assert "make-ellipse-by-center-2-radii-rotation" c 'hanger r1 'real r2 'real rot 'real)
  (vrr-link (go-elarc-new-xyr c rot r1 r2 0 7 (code-conic 'start-entire))))

(define-public (make-ellipse-by-2-foci-point f1 f2 p)
  (vrr-type-assert "make-ellipse-by-2-foci-point" f1 'hanger f2 'hanger p 'hanger)
  (vrr-link (go-elarc-new-foci f1 f2 p (code-conic 'point-entire))))

(define-public (make-ellipse-by-3-points-smallest p1 p2 p3)
  (vrr-type-assert "make-ellipse-by-3-points-smallest" p1 'hanger p2 'hanger p3 'hanger)
  (vrr-link (go-elarc-new-3small p1 p2 p3 (code-conic 'point-entire))))

(define-public (make-ellipse-by-3-points-rotation-eccentricity p1 p2 p3 rot ecc)
  (vrr-type-assert "make-ellipse-by-3-points-rotation-eccentricity"
		   p1 'hanger p2 'hanger p3 'hanger rot 'real ecc 'real)
  (vrr-link (go-elarc-new-3ecc p1 p2 p3 rot ecc (code-conic 'point-entire))))

(define-public (make-ellipse-by-center-point-rotation-eccentricity c p rot ecc)
  (vrr-type-assert "make-ellipse-by-center-point-rotation-eccentricity"
		   c 'hanger p 'hanger rot 'real ecc 'real)
 (vrr-link (go-elarc-new-xy1ecc c p rot ecc (code-conic 'point-entire))))


(define-public make-ellipse/c2rr make-ellipse-by-center-2-radii-rotation)
(define-public make-ellipse/2fp make-ellipse-by-2-foci-point)
(define-public make-ellipse/3ps make-ellipse-by-3-points-smallest)
(define-public make-ellipse/3pre make-ellipse-by-3-points-rotation-eccentricity)
(define-public make-ellipse/cpre make-ellipse-by-center-point-rotation-eccentricity)
(define-public make-ellipse make-ellipse-by-center-2-radii-rotation)


(define-public (make-elarc-by-center-2-radii-rotation-2-angles c r1 r2 rot a1 a2 style)
  (vrr-type-assert "make-elarc-by-center-2-radii-rotation-2-angles" 
		   c 'hanger r1 'real r2 'real rot 'real a1 'real a2 'real
		   style (number? (code-conic style)) 'conic-style-symbol)
  (vrr-link (go-elarc-new-xyr c rot r1 r2 a1 a2 (code-conic style))))

(define-public (make-elarc-by-3-points-smallest p1 p2 p3 style)
  (vrr-type-assert "make-elarc-by-3-points-smallest" p1 'hanger p2 'hanger p3 'hanger
		   style (number? (code-conic style)) 'conic-style-symbol)
  (vrr-link (go-elarc-new-foci p1 p2 p3 (code-conic style))))


(define-public make-elarc/c2rr2a make-elarc-by-center-2-radii-rotation-2-angles)
(define-public make-elarc/3ps make-elarc-by-3-points-smallest)


(define-public (make-quadratic-bezier p1 p2 p3)
  (vrr-type-assert "make-quadratic-bezier" p1 'hanger p2 'hanger p3 'hanger)
  (vrr-link (go-bezier-new-quadratic p1 1 p2 1 p3 1)))

(define-public (make-quadratic-rational-bezier p1 w1 p2 w2 p3 w3)
  (vrr-type-assert "make-quadratic-rational-bezier"
		   p1 'hanger w1 'real p2 'hanger w2 'real p3 'hanger w3 'real)
  (vrr-link (go-bezier-new-quadratic p1 w1 p2 w2 p3 w3)))

(define-public (make-cubic-bezier p1 p2 p3 p4)
  (vrr-type-assert "make-cubic-bezier" p1 'hanger p2 'hanger p3 'hanger p4 'hanger)
  (vrr-link (go-bezier-new-cubic p1 1 p2 1 p3 1 p4 1)))

(define-public (make-cubic-rational-bezier p1 w1 p2 w2 p3 w3 p4 w4)
  (vrr-type-assert "make-cubic-rational-bezier"
		   p1 'hanger w1 'real p2 'hanger w2 'real p3 'hanger w3 'real p4 'hanger w4 'real)
  (vrr-link (go-bezier-new-cubic p1 w1 p2 w2 p3 w3 p4 w4)))

(define-public make-bezier/2 make-quadratic-bezier)
(define-public make-bezier/2r make-quadratic-rational-bezier)
(define-public make-bezier/3 make-cubic-bezier)
(define-public make-bezier/3r make-cubic-rational-bezier)


(define-public (make-tex-text h str)
  (vrr-type-assert "make-tex-text" h 'hanger str 'string)
  (vrr-link (go-tex-text-new h 0 0 0 0 str)))

(define-public (make-text h str font-family font-style font-size)
  (vrr-type-assert "make-text" h 'hanger str 'string font-family 'string font-style 'string font-size 'real)
  (vrr-link (go-text-new h 0 0 0 0 str (font-load-by-family-style font-family font-style) font-size)))


(define-public (make-parametric-point go p)
  (vrr-type-assert "make-parametric-point" go 'go p 'real)
  (vrr-link (go-parametric-point-new go p)))

(define-public (make-parametric-point-on-hanger h p)
  (vrr-type-assert "make-parametric-point-on-hanger" h 'hanger p 'real)
  (vrr-link (go-parametric-point-new-on-hanger h p #f 0 0)))


(define-public (make-intersection g1 g2 p)
  (vrr-type-assert "make-intersection" g1 'go g2 'go p 'real)
  (vrr-link (go-intersection-new g1 g2 p)))

(define-public (make-intersection-on-hanger h1 h2 p)
  (vrr-type-assert "make-intersection-on-hanger" h1 'hanger h2 'hanger p 'real)
  (vrr-link (go-intersection-new-on-hanger h1 h2 p #f 0 0)))


(define-public (make-decorator-point h n radius rot)
  (vrr-type-assert "make-decorator-point" h 'hanger n 'natural radius 'real rot 'real)
  (vrr-link (go-decorator-point-new h n radius rot)))


(define-public (make-decorator-arrow g p arrow-length rot)
  (vrr-type-assert "make-decorator-arrow" g 'go p 'real arrow-length 'real rot 'real)
  (vrr-link (go-arrow-new g p arrow-length rot)))

(define-public (make-decorator-arrow-on-hanger h p arrow-length rot)
  (vrr-type-assert "make-decorator-arrow-on-hanger" h 'hanger p 'real arrow-length 'real rot 'real)
  (vrr-link (go-arrow-new-on-hanger h p arrow-length rot)))


(define-public (make-group)
  (vrr-link (go-group-new)))

(define-public (make-path)
  (vrr-link (go-path-new)))



;; ************************
;; part 3 - n (as namespace) functions and hierarchy
;; ************************

(define (obj-prev-son g s)  (obj-get-prev-son s))

(define (vrr-type-assert-o name x1)
  (vrr-type-assert name x1 (proxy? x1 'obj) 'obj))

(define (vrr-type-assert-oo name x1 x2)
  (vrr-type-assert name
		   x1 (proxy? x1 'obj) 'obj
		   x2 (proxy? x2 'obj) 'obj))


;; public

(define-public (n-leaf? x) (proxy? x 'obj 'page))

(define-public (n<=? o1 x2)
  (vrr-type-assert-oo "n<=?" o1 x2)
  (obj-is-ancestor x2 o1))

(define-public (n>=? o1 x2)
  (vrr-type-assert-oo "n>=?" o1 x2)
  (obj-is-ancestor o1 x2))

(define-public (n<? o1 x2)
  (vrr-type-assert-oo "n<?" o1 x2)
  (and (not (eq? o1 x2)) (n<=? o1 x2)))

(define-public (n>? o1 x2)
  (vrr-type-assert-oo "n>?" o1 x2)
  (and (not (eq? o1 x2)) (n>=? o1 x2)))

;(define-public (n-root) ;; FIXME silly, undocumented
;  universe)

(define-public (n-parent o1)
  (vrr-type-assert-o "n-parent" o1)
  (let ((p (obj-get-parent o1)))
    (if (or (proxy? p 'obj 'temp)
	    (proxy? p 'obj 'zombie))
	#f
	p)))


(define-public (n-ancestors o1)
  (vrr-type-assert-o "n-ancestors" o1)
  (collect-list-1r (n-parent o1) n-parent))

(define-public (n-first-child o1)
  (vrr-type-assert-o "n-first-child" o1)
  (obj-get-first-son o1))

(define-public (n-last-child o1)
  (vrr-type-assert-o "n-last-child" o1)
  (obj-get-last-son o1))

(define-public (n-first-sib o1)
  (vrr-type-assert-o "n-first-sib" o1)
  (let ((p (n-parent o1)))
    (if p (n-first-child p) #f)))

(define-public (n-last-sib o1)
  (vrr-type-assert-o "n-last-sib" o1)
  (let ((p (n-parent o1)))
    (if p (n-last-child p) #f)))

(define-public (n-next-sib o1)
  (vrr-type-assert-o "n-next-sib" o1)
  (obj-get-next-son o1))

(define-public (n-prev-sib o1)
  (vrr-type-assert-o "n-prev-sib" o1)
  (obj-get-prev-son o1))

(define-public (n-children o1)
  (vrr-type-assert-o "n-children" o1)
  (collect-list-1r (obj-get-last-son o1) obj-get-prev-son))

(define-public (n-descendants o1)
  (vrr-type-assert-o "n-descendants" o1)
  (collect-tree-r o1 obj-get-last-son obj-prev-son))

(define-public (n-leaves o1)
  (vrr-type-assert-o "n-leaves" o1)
  (collect-tree-r-and-filter o1 obj-get-last-son obj-prev-son n-leaf?))


;; FIXME check circularity ?
(define-public (n-set-parent! o1 o2)
  (vrr-type-assert "n-set-parent!" o1 'obj o2 (or (proxy? o2 'obj) (not o2)) 'obj-or-false)
  (if (n-parent o1) (obj-unlink o1))
  (if o2 (obj-link-tail o1 o2)))

;; support



;; ************************
;; part 4 - g (as group) functions and hierarchy
;; ************************

(define (vrr-type-assert-g name x1)
  (vrr-type-assert name x1 (proxy? x1 'go) 'go))

(define (vrr-type-assert-gg name x1 x2)
  (vrr-type-assert name
		   x1 (proxy? x1 'go) 'go
		   x2 (proxy? x2 'go) 'go))

(define-public (g-leaf? x) (and (proxy? x 'go) (not (proxy? x 'go 'group))))

(define-public (g<=? x1 x2)
  (vrr-type-assert-gg "g<=?" x1 x2)
  (go-is-ancestor x2 x1))

(define-public (g>=? x1 x2)
  (vrr-type-assert-gg "g>=?" x1 x2)
  (go-is-ancestor x1 x2))

(define-public (g<? x1 x2)
  (vrr-type-assert-gg "g<?" x1 x2)
  (and (not (eq? x1 x2)) (g<=? x1 x2)))

(define-public (g>? x1 x2)
  (vrr-type-assert-gg "g>?" x1 x2)
  (and (not (eq? x1 x2)) (g>=? x1 x2)))

(define-public (g-root x1)
  (vrr-type-assert-g "g-root" x1)
  (go-get-root x1))

(define-public (g-parent x1)
  (vrr-type-assert-g "g-parent" x1)
  (go-get-group x1))

(define-public (g-ancestors x1)
  (vrr-type-assert-g "g-ancestors" x1)
  (collect-list-1 (g-parent x1) g-parent))

(define-public (g-first-child x1)
  (vrr-type-assert-g "g-first-child" x1)
  (groupize2 group-get-first-go))

(define-public (g-last-child x1)
  (vrr-type-assert-g "g-last-child" x1)
  (groupize2 group-get-last-go))

(define-public (g-first-sib go)
  (let ((p (g-parent go)))
    (if p (g-first-child p) #f)))
(define-public (g-last-sib go)
  (let ((p (g-parent go)))
    (if p (g-last-child p) #f)))
(define-public (g-next-sib go)
  (let ((p (g-parent go)))
    (if p (group-get-next-go p go) #f)))
(define-public (g-prev-sib go)
  (let ((p (g-parent go)))
    (if p (group-get-prev-go p go) #f)))

(define-public (g-children x1)
  (vrr-type-assert-g "g-children" x1)
  (if (proxy? x1 'go 'group)
      (collect-list-2r x1 group-get-last-go group-get-prev-go)
      '()))

(define-public (g-descendants x1)
  (vrr-type-assert-g "g-descendants" x1)
  (if (proxy? x1 'go 'group)
      (collect-tree-r x1 group-get-last-go group-get-prev-go)
      '()))

(define-public (g-leaves x1)
  (vrr-type-assert-g "g-leaves" x1)
  (if (proxy? x1 'go 'group)
      (collect-tree-r-and-filter x1 group-get-last-go group-get-prev-go g-leaf?)
      '()))

;; FIXME check common parent for go an g2
(define-public (g-set-parent/before go grp g2)
  (vrr-type-assert "g-set-parent/before" go 'go grp (proxy? grp 'go 'group) 'go-group g2 'go)
  (let ((bg (g-prev-sib g2)))
    (if bg
	(go-relink go grp bg)
	(go-relink-head go grp)))
  *unspecified*)

(define-public (g-set-parent/after go grp g2)
  (vrr-type-assert "g-set-parent/after" go 'go grp (proxy? grp 'go 'group) 'go-group g2 'go)
  (go-relink go grp g2)
  *unspecified*)

(define-public (g-set-parent/top go grp)
  (vrr-type-assert "g-set-parent/top" go 'go grp (proxy? grp 'go 'group) 'go-group)
  (go-relink-head go grp)
  *unspecified*)

(define-public (g-set-parent/bottom go grp)
  (vrr-type-assert "g-set-parent/bottom" go 'go grp (proxy? grp 'go 'group) 'go-group)
  (go-relink go grp (g-last-child grp))
  *unspecified*)

;; must be (not (eq? go stop))
;; return list of gos (path from go to root in g tree) in case there is no go 'stop' in path
;; return go 'x' in case there is a go 'stop' in path. 'x' is in path and parent of 'x' is go 'stop'.
 
(define (path-to-root go stop)
  (let loop ((act go)
	     (acc '()))
    (cond
     ((not act) acc)
     ((eq? act stop) (car acc))
     (else (loop (g-parent go) (cons act acc))))))
  
(define-public (g-set-parent! go grp)
  (vrr-type-assert "g-set-parent!" go 'go grp (proxy? grp 'go 'group) 'go-group)

  (if (eq? go grp) (error "circular dependence"))
  (let ((p1 (path-to-root go grp))
	(p2 (path-to-root grp go)))
    (cond 
     ((not (list? p2))
      (error "circular dependence"))
     ((not (list? p1))
      (g-set-parent/before go grp p1))
     (else
      (let loop ((t1 p1)
		 (t2 p2))
	(cond
	 ((eq? (car t1) (car t2))
	  (loop (cdr t1) (cdt t2)))
	 ;; (car t1) and (car t2) is defined, because last elements of p1 and p2 are different
	 ((z< (car t1) (car t2)) 
	  (g-set-parent/top go grp))
	 (else
	  (g-set-parent/bottom go grp)))))))
  *unspecified*)



;; ************************
;; part 5 - d (as dependency) functions and hierarchy
;; ************************

;(define-public (d<=? x1 x2)
;  (vrr-type-assert-gg "d<=?" x1 x2)
;  (go-is-dependent x1 x2)) ; FIXME missing

;(define-public (d>=? x1 x2)
;  (vrr-type-assert-gg "d>=?" x1 x2)
;  (go-is-dependent x2 x1))

;(define-public (d<? x1 x2)
;  (vrr-type-assert-gg "d<?" x1 x2)
;  (and (not (eq? x1 x2)) (d<=? x1 x2)))

;(define-public (d>? x1 x2)
;  (vrr-type-assert-gg "d>?" x1 x2)
;  (and (not (eq? x1 x2)) (d>=? x1 x2)))


(define-public (d-parent go id)
  (vrr-type-assert "d-parent"
		   go (proxy? go 'go) 'go
		   id (symbol? id) 'symbol)
  (let ((a (go-anchor go id)))
    (if a
	(linear-call anchor-hanger hanger-go)
	(error "invalid anchor id ~S" id))))

(define-public (d-children go id)
  (vrr-type-assert "d-children"
		   go (proxy? go 'go) 'go
		   id (symbol? id) 'symbol)
  (let ((h (go-hanger go id)))
    (if h
	(map anchor-go (hanger-anchors h))
	(error "invalid hanger id ~S" id))))

(define-public (d-all-parents go)
  (vrr-type-assert-g "d-all-parents" go)
  (unify-go-list (map (lambda (a) (hanger-go (anchor-hanger a))) (go-anchors go))))

(define-public (d-all-children go)
  (vrr-type-assert-g "d-all-children" go)
  (unify-go-list (map anchor-go (apply append (map hanger-anchors (go-hangers go))))))

; (define-public (d-descendants go) ERROR) ; FIXME missing, undocumented

(define-public (d-set-parent! g1 k1 g2 k2)
  (vrr-type-assert "d-set-parent!"
		   g1 (proxy? g1 'go) 'go
		   k1 (symbol? k1) 'symbol
		   g2 (proxy? g2 'go) 'go
		   k2 (symbol? k2) 'symbol)
  (let ((a (go-anchor g1 k1))
	(h (go-hanger g2 k2)))
    (cond
     ((not a) 
      (error "invalid anchor id ~S" k1))
     ((not h) 
      (error "invalid hanger id ~S" k2))
     ;; FIXME check collision of collision
     (else
      (anchor-rehang a h)))))



;; ************************
;; part 6 - anchors/hangers
;; ************************

(define-public (go-hanger g id)
  (vrr-type-assert "go-hanger" g 'go id 'symbol)
  (go-get-hanger-by-name g id))

(define-public (go-hangers g)
  (vrr-type-assert-g "go-hangers" g)
  (go-get-hangers g))

(define-public (go-anchor g id)
  (vrr-type-assert "go-anchor" g 'go id 'symbol)
  (go-get-anchor-by-name g id))

(define-public (go-anchors g)
  (vrr-type-assert-g "go-anchors" g)
  (go-get-anchors g))

(define-public (anchor-first-sib a)
  (vrr-type-assert "anchor-first-sib" a 'anchor)
  (hanger-first-anchor (anchor-hanger a)))

(define-public (anchor-last-sib a)
  (vrr-type-assert "anchor-last-sib" a 'anchor)
  (let loop ((a a))
    (let ((next (anchor-next-sib a)))
      (if next (loop next) a))))

(define-public (anchor-next-sib a)
  (vrr-type-assert "anchor-next-sib" a 'anchor)
  (hanger-get-next-anchor a))

(define-public (anchor-prev-sib a)
  (vrr-type-assert "anchor-prev-sib" a 'anchor)
  (let loop ((act (anchor-first-sib a))
	     (last #f))
    (if (eq? act a) last (loop (anchor-next-sib act) act))))

(define-public (anchor-hanger a)
  (vrr-type-assert "anchor-hanger" a 'anchor)
  (anchor-get-hanger a))

(define-public (anchor-go a)
  (vrr-type-assert "anchor-go" a 'anchor)
  (anchor-get-go a))

(define-public (hanger-first-anchor h)
  (vrr-type-assert "hanger-first-anchor" h 'hanger)
  (hanger-get-first-anchor h))

(define-public (hanger-last-anchor h)
  (vrr-type-assert "hanger-last-anchor" h 'hanger)
  (let loop ((act (hanger-first-anchor h))
	     (last #f))
    (if act (loop (anchor-next-sib act) act) last)))

(define-public (hanger-anchors h)
  (vrr-type-assert "hanger-anchors" h 'hanger)
  (collect-list-1 (hanger-get-first-anchor h) hanger-get-next-anchor))

(define-public (hanger-go h)
  (vrr-type-assert "hanger-go" h 'hanger)
  (hanger-get-go h))

(define-public (anchor-set-hanger! a h)
  (vrr-type-assert "anchor-set-hanger!" a 'anchor h 'hanger)
  (anchor-rehang a h))


(define-public (anchor-term a)
  (vrr-type-assert "anchor-term" a 'anchor)
  (anchor-get-class-name a))

(define-public (hanger-term h)
  (vrr-type-assert-h "hanger-term" h)
  (hanger-get-class-name a))

;; **********************************
;; part 7 - inter-hieararchy movement
;; **********************************

(define-public (go-superior-page g)
  (vrr-type-assert-g "go-superior-page" g)
  (go-get-tlo g))

(define-public (page-g-root p)
  (vrr-type-assert "page-g-root" p (proxy? p 'obj 'page) 'page)
  (tlo-get-tlg p))



;; *************************
;; part 8 - z-order movement
;; *************************


(define (z<?-cp x1 x2)
  (let loop ((tmp (g-first-sib x1)))
    (cond 
     ((eq? tmp x1) #t)
     ((eq? tmp x2) #f)
     (tmp (loop (zgrpx-next-sib tmp)))
     (else (error "bug in use of z<-cp")))))

(define (z<?-int x1 x2)
  (if (eq? (g-parent x1) (g-parent x2))
      (z<?-cp x1 x2)
      (let loop ((l1 (collect-list-1 x1 g-parent))
		 (l2 (collect-list-1 x2 g-parent)))
	(cond
	 ((or (null? l1) (null? l2))
	  #f)
	 ((eq? (car l1) (car l2))
	  (loop (cdr l1) (cdr l2)))
	 (else
	  (z<?-cp (car l1) (car l2)))))))
	    
(define-public (z<=? x1 x2)
  (vrr-type-assert-gg "z<=?" x1 x2)
  (or (eq? x1 x2) (z<?-int x1 x2)))

(define-public (z>=? x1 x2)
  (vrr-type-assert-gg "z>=?" x1 x2)
  (or (eq? x1 x2) (z<?-int x2 x1)))

(define-public (z<? x1 x2)
  (vrr-type-assert-gg "z<?" x1 x2)
  (and (not (eq? x1 x2)) (z<?-int x1 x2)))

(define-public (z>? x1 x2)
  (vrr-type-assert-gg "z>?" x1 x2)
  (and (not (eq? x1 x2)) (z<?-int x2 x1)))


(define (tmp-relink go after)
  (if after
      (go-relink go (g-parent go) after)
      (go-relink-head go (g-parent go)))
  #t)

(define-public (z-move-top! go)
  (vrr-type-assert-g "z-move-top!" go)
  (let ((temp (g-first-sib go)))
    (and temp (not (eq? temp go)) (tmp-relink go temp))))

(define-public (z-move-up! go)
  (vrr-type-assert-g "z-move-up!" go)
  (let* ((temp (g-prev-sib go))
	(temp2  (and temp (g-prev-sib temp))))
    (and temp (tmp-relink go temp2))))

(define-public (z-move-down! go)
  (vrr-type-assert-g "z-move-down!" go)
  (let ((temp (g-next-sib go)))
    (and temp (tmp-relink go temp))))

(define-public (z-move-bottom! go)
  (vrr-type-assert-g "z-move-bottom!" go)
  (let ((temp (g-last-sib go)))
    (and temp (not (eq? temp go)) (tmp-relink go temp))))



;; ******************
;; part 9 - selection
;; ******************

(define-public (select . os) (for-each go-select os))
(define-public (unselect . os) (for-each go-unselect os))

(define-public (unselect-all)  (tlo-unselect-all-go))

(define-public (select-g-children g)
  (vrr-type-assert-g "unselect-g-children" g)
  (if (proxy? g 'go 'group)
      (group-select-all-go g)))

(define-public (unselect-g-children g)
  (vrr-type-assert-g "unselect-g-children" g)
  (if (proxy? g 'go 'group)
      (group-unselect-all-go g)))

  
(define-public selected? o-is-selected)

(define-public (selected-objs)
  (collect-list-1 (obj-get-first-selected) obj-get-next-selected))

(define-public (selected-g-children x1)
  (vrr-type-assert-g "selected-g-children" x1)
  (if (proxy? x1 'go 'group)
      (collect-list-2 x1 group-get-first-selected-go group-get-next-selected-go)
      '()))

(define-public (selected-g-descendants x1)
  (vrr-type-assert-g "selected-g-descendants" x1)
  (if (proxy? x1 'go 'group)
      (collect-tree-r-and-filter x1 group-get-last-go group-get-prev-go selected?)
      '()))


;; (define-public anchor-name anchor-get-class-name)
;; (define-public hanger-name hanger-get-class-name)



;; ************************
;; part 10 - transformations
;; ************************

(re-export move-selected rotate-selected scale-selected skew-selected transform-selected)

(define-public (remove-selected grp)
  (vrr-type-assert "remove-selected" grp (proxy? grp 'go 'group) 'go-group)
  (group-remove-rehang-selected-go grp))

(define-public (copy-selected src-grp dst-grp)
  (vrr-type-assert "copy-selected" 
		   src-grp (proxy? src-grp 'go 'group) 'go-group
		   dst-grp (proxy? dst-grp 'go 'group) 'go-group)
  (clipboard-copy-selected-go-on-front src-grp dst-grp #f))

;(define-public (copy-selected-after src-grp dst-grp) ; FIXME
;  (vrr-type-assert "copy-selected" 
;		   src-grp (proxy? src-grp 'go 'group) 'go-group
;		   dst-grp (proxy? dst-grp 'go 'group) 'go-group)
;  (clipboard-copy-selected-go src-grp dst-grp XXX #f))


;; MISC

(define-public (trans-page)
  (false-if-exception (get-trans-tlo)))

(define-public plugin-load load-plugin)
(define-public universum universe)

;; TODO: selekce (obj-select, unselect, unselect-all, selected?), go-info,  go-selekce, z-move-after, z-move-before


(define-constructor 'point make-point)
(define-constructor 'segment make-segment)

(define-constructor 'ellipse-by-center-2-radii-rotation-2-angles 
  make-elarc-by-center-2-radii-rotation-2-angles) ;; OBSOLETE

(define-constructor 'ellipse-by-center-2-radii-rotation make-ellipse-by-center-2-radii-rotation)
(define-constructor 'ellipse-by-2-foci-point make-ellipse-by-2-foci-point)
(define-constructor 'ellipse-by-3-points-smallest make-ellipse-by-3-points-smallest)
(define-constructor 'ellipse-by-3-points-rotation-eccentricity make-ellipse-by-3-points-rotation-eccentricity)
(define-constructor 'ellipse-by-center-point-rotation-eccentricity make-ellipse-by-center-point-rotation-eccentricity)

(define-constructor 'quadratic-bezier make-quadratic-rational-bezier)
(define-constructor 'cubic-bezier make-cubic-rational-bezier)

(define-constructor 'text make-text)
(define-constructor 'tex-text make-tex-text)

(define-constructor 'parametric-point make-parametric-point-on-hanger)
(define-constructor 'intersection make-intersection-on-hanger)
(define-constructor 'decorator-point make-decorator-point)
(define-constructor 'decorator-arrow make-decorator-arrow-on-hanger)

(define-constructor 'group make-group)
(define-constructor 'path make-path)
