;;; miscellaneous functions
;;;
;;; Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
;;; This program is distributed under GNU General Public License, see COPYING.

(define-module (vrr misc))

;; actions: none

(define-public (make-thunk fn . args)
  (lambda () (apply fn args)))

(define-public (curry fn arg1)
  (lambda args (apply fn (cons arg1 args))))

(define-public (safe-force p)
  (if (promise? p) (force p) p))

(define-public (natural? n)
  (and (integer? n) (exact? n) (>= n 0)))

(define-public (every predicate list)
  (or (null? list)
      (and (predicate (car list))
	   (every predicate (cdr list)))))

(define-public (flatten data)
  (let rec ((node data)
	    (tail '()))
    (cond
     ((pair? node)
      (rec (car node) (rec (cdr node) tail)))
     ((null? node)
      tail)
     (else
      (cons node tail)))))

(define-public (true . x) #t)
(define-public (id x) x)

(define-public (flag? number flag) (positive? (logand number flag)))

(define-public (linear-call init . fns)
  (let loop ((val init)
	     (todo fns))
    (if (or (not val) (null? todo))
	val
	(loop ((car todo) val) (cdr todo)))))

(define-public (collect-list-1 o f)
  (let rec ((arg o))
    (if arg
	(cons arg (rec (ff arg)))
	'())))

(define-public (collect-list-2 o f1 f2)
  (let rec ((arg (f1 o)))
    (if arg
	(cons arg (rec (f2 o arg)))
	'())))

(define-public (collect-list-1r o f)
  (let rec ((arg o)
	    (acc '()))
    (if arg
	(rec (f arg) (cons arg acc))
	acc)))

(define-public (collect-list-2r o f1 f2)
  (let rec ((arg (f1 o))
	    (acc '()))
    (if arg
	(rec (f2 o arg) (cons arg acc))
	acc)))

(define-public (collect-tree-r o f1 f2)
  (let subtree ((o o)
		(acc1 '()))
    (let rec ((n (f1 o))
	      (acc2 acc1))
      (if n
	  (rec (f2 o n)
	       (cons n (subtree n acc2)))
	  acc2))))

(define-public (collect-tree-r-and-filter o f1 f2 pred)
  (lambda (o)
    (let subtree ((o o)
		  (acc1 '()))
      (let rec ((n (f1 o))
		(acc2 acc1))
	(if n
	    (rec (f2 o n)
		 (if (pred n)
		     (cons n (subtree n acc2))
		     (subtree n acc2)))
	    acc2)))))
