;;; Save mechanism
;;;
;;; Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
;;; This program is distributed under GNU General Public License, see COPYING.

(define-module (vrr save))

(use-modules (vrr misc))
(use-modules (vrr low))
(use-modules (vrr support))
(use-modules (vrr property))

(export save-doc)

;; actions: none

;; FIXME - check whether id property is different for every go
;; in one tlo and be string or undefined

(define (group-children x1)
  (if (proxy? x1 'go 'group)
      (collect-list-2r x1 group-get-last-go group-get-prev-go)
      '()))

(define (direct-write? go)
  (not (or (proxy? go 'go 'group 'top-level-group)
	   (go-has-dependent go))))

(define (save-internal doc port)
  (define (write-space x) (write-char #\space port) (write x port))
  (define (write-open-list) (write-char #\space port) (write-char #\( port))
  (define (write-close-list) (write-char #\) port))

  (define (write-properties obj next)
    (write-open-list)
    (for-each (lambda (x) (write-space x)) next)
    (property-for-each obj (lambda (name value type subtype flags)
			     (if (flag? flags CC_PTF_SAVE)
				 (write-space (list name (prop-decode type subtype value))))))
    (write-close-list))

  (define (save-hanger h)
    (cond
     ((and (hanger-is-mc h) (hanger-is-position h))
      (write-open-list)
      (write 'coords port)
      (write-space (hanger-get-pos-x h))
      (write-space (hanger-get-pos-y h))
      (write-close-list))
     ((and (hanger-is-mc h) (hanger-is-curve h))
      (write-space #f))
     ((hanger-is-curve h)
      (link-go (hanger-get-go h)))
     (else
      (write-open-list)
      (let ((tp (hanger-get-class-name h)))
	(write tp port)
	(link-go (hanger-get-go h)))
      (write-close-list))))

  (define (link-go go)
    (if (direct-write? go)
	(write-go go #f)
	(let ((id (get-property go 'id)))
	  (if id
	      (write-space id)
	      (write-space (proxy2ptr go))))))

  (define (save-go go)
    (if (not (direct-write? go))
	(write-go go #t)))

  (define (write-go go want-id)
    (newline port)
    (write-open-list)
    (write (proxy-subtype go) port)

    (if (not (proxy? go 'go 'group))
	(for-each (lambda (src)
		    (cond
		     ((proxy? src 'anchor)
		      (save-hanger (anchor-get-hanger src)))
					;		 ((proxy? src 'go)
					;		  (link-go src))
		     (else
		      (write-space src))))
		  (go-get-sources go)))

    (write-properties go (append
			  (if (and want-id
				   (not (proxy? go 'go 'group 'top-level-group))
				   (not (get-property go 'id)))
			      (list (list 'id (proxy2ptr go)))
			      '())
			  (if (proxy? go 'go 'elliptic-arc)
			      (list (list 'conic (get-conic go)))
			      '())))

    (if (proxy? go 'go 'group)
	(begin
	  (write-open-list)
	  (for-each link-go (group-children go))
	  (write-close-list)))

    (write-close-list))


  (define (save-tlo-content tlo)
    (trans-main tlo (lambda ()
		      (tsort-tlo tlo save-go))))

  ;; body of save-internal
  (let rec ((obj doc))
    (if obj
	(begin
	  (newline port)
	  (write-open-list)
	  (write (proxy-type obj) port)
	  (if (eq? obj doc)
	      (write-space 0.6))
	  (write-properties obj '())
	  (if (not (eq? obj universe))
	      (begin
		(write-open-list)
		(if (proxy? obj 'obj 'page)
		    (save-tlo-content obj)
		    (rec (obj-get-first-son obj)))
		(write-close-list)))
	  (write-close-list)
	  (if (not (eq? obj doc))
	      (rec (obj-get-next-son obj)))))))

(define (save-doc doc filename)
  (let ((do-backup (file-exists? filename)))
    (if do-backup
        (rename-file filename (string-append filename "~")))
    (trans-main universe
	      (lambda ()
		(call-with-output-file filename
		  (lambda (port) (save-internal doc port))))
	      (lambda (exc . args) 
	        (if do-backup (rename-file (string-append filename "~") filename))
		(throw exc args)))))
