/*
 * VCL module
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG

#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "kernel/kernel.h"
#include "font/font.h"

vcl_vtable *vcl_vt_char;


// NODE interface

static void
vcl_char_set_data (UNUSED struct vcl_char * obj, UNUSED int data)
{ BUG (); }

static int
vcl_char_get_data (UNUSED struct vcl_char * obj)
{ BUG (); }

static void
vcl_char_set_parent (UNUSED struct vcl_char * obj, UNUSED  void * parent)
{ BUG (); }

static void *
vcl_char_get_parent (UNUSED struct vcl_char * obj)
{ BUG (); }

static void
vcl_char_get_bbox (struct vcl_char * obj, struct geom_rectangle * bbox)
{ *bbox = obj -> bbox; }


// MASK interface

static void
vcl_char_render (struct vcl_char * obj, const struct geom_transform *t,
		 UNUSED const struct vcl_rectangle *bbox, const struct geom_rectangle *bbox_local,
		 u8** buff, uns * buff_len, struct vcl_rectangle *retbox)
{
  if (geom_rectangle_have_intersection (&(obj -> bbox), bbox_local))
    {
      struct font_ctl ctl;
      font_ctl_init(&ctl);

      ctl . charmap = obj -> charmap;
      ctl . font_id = obj -> font_id;
      ctl . font_size = obj -> font_size;
      ctl . transform = *t;

      ctl . bitmap = *buff;
      ctl . bsize = *buff_len;

      if (font_render_glyph(&ctl, obj -> code) >= 0)
	{
	  retbox -> lx = ctl . xoff;
	  retbox -> ly = ctl . yoff;
	  retbox -> hx = ctl . xoff + ctl . width;
	  retbox -> hy = ctl . yoff + ctl . height;
	}
      else vcl_rectangle_set_empty (retbox);

      *buff = ctl . bitmap;
      *buff_len = ctl . bsize;

      ctl . bitmap = NULL;
      ctl . bsize = 0;
      font_ctl_cleanup(&ctl);
    }
  else vcl_rectangle_set_empty (retbox);
}


// specific functions

extern void
vcl_char_init (struct vcl_char *obj,  uns code, uns charmap,  int font_id, real font_size, struct geom_rectangle bbox)
{
  vcl_object_init (obj, vcl_vt_char);
  obj -> code = code;
  obj -> charmap = charmap;
  obj -> font_id = font_id;
  obj -> font_size = font_size;
  obj -> bbox = bbox;
}

void
vcl_char_main (void)
{
  vcl_vt_char = vcl_class_new ("char");

  vcl_class_add_interface (vcl_vt_char, vcl_if_node);
  vcl_class_add_method_li (vcl_mid_node_set_data, vcl_char_set_data);
  vcl_class_add_method_li (vcl_mid_node_get_data, vcl_char_get_data);
  vcl_class_add_method_li (vcl_mid_node_set_parent, vcl_char_set_parent);
  vcl_class_add_method_li (vcl_mid_node_get_parent, vcl_char_get_parent);
  vcl_class_add_method_li (vcl_mid_node_get_bbox, vcl_char_get_bbox);

  vcl_class_add_interface (vcl_vt_char, vcl_if_mask);
  vcl_class_add_method_li (vcl_mid_mask_render, vcl_char_render);
}
