/*
 * VCL module
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>

vcl_vtable *vcl_vt_grid;


// OBJECT interface

static void
vcl_grid_destroy (struct vcl_grid * obj)
{
  ASSERT (obj -> parent == NULL);
  xfree (obj);
}


// NODE interface

static void
vcl_grid_set_data (struct vcl_grid * obj, int data)
{
  obj -> data = data;
}

static int
vcl_grid_get_data (struct vcl_grid * obj)
{
  return (obj -> data);
}

static void
vcl_grid_set_parent (struct vcl_grid * obj, void * parent)
{
  obj -> parent = parent;
}

static void *
vcl_grid_get_parent (struct vcl_grid * obj)
{
  return (obj -> parent);
}

static void
vcl_grid_get_bbox (struct vcl_grid * obj UNUSED, struct geom_rectangle * bbox)
{
  // FIXME +/- infinity ?
  bbox -> ll . x = -1000000;
  bbox -> ll . y = -1000000;
  bbox -> ur . x = +1000000;
  bbox -> ur . y = +1000000;
}


// SHAPE interface

struct tmp_data
{
  struct geom_rectangle int_bbox;
  real i;
  uns second;
};

static void
return_line (real x1, real y1, real x2, real y2,
	    const struct geom_transform *t,
	    struct geom_point ** buff,
	    uns * buff_size, uns *closed, uns *points)
{
  if (*buff_size < 2)
    {
      *buff_size = 2;
      *buff = xrealloc (*buff, *buff_size * sizeof (struct geom_point));
    }

  struct geom_point pi;

  pi . x = x1;
  pi . y = y1;
  geom_point_transform (&pi, t, *buff + 0);

  pi . x = x2;
  pi . y = y2;
  geom_point_transform (&pi, t, *buff + 1);

  *closed = 0;
  *points = 2;
}

static void *
vcl_grid_polygonize (UNUSED struct vcl_grid * obj, void * it, UNUSED double delta,
		     const struct geom_transform *t,
		     const struct geom_rectangle *bbox,
		     u8 ** buff, uns * buff_len,
		     uns *closed, uns *points, uns hints)
{
  struct tmp_data *tmp_data;

  if (hints == VCL_SHAPE_CLOSED_ONLY)
  {
    *closed = 1;
    *points = 0;
    return (NULL);
  }

  if (it == NULL)
  {
    tmp_data = xmalloc (sizeof (struct tmp_data));
    tmp_data -> int_bbox . ll . x = floor (bbox -> ll . x);
    tmp_data -> int_bbox . ll . y = floor (bbox -> ll . y);
    tmp_data -> int_bbox . ur . x = ceil (bbox -> ur . x);
    tmp_data -> int_bbox . ur . y = ceil (bbox -> ur . y);
    tmp_data -> i = tmp_data -> int_bbox . ll . x;
    tmp_data -> second = 0;
  }
  else tmp_data = it;

  if (tmp_data -> second == 0)
  {
    return_line (tmp_data -> i, tmp_data -> int_bbox . ll . y,
	         tmp_data -> i, tmp_data -> int_bbox . ur . y,
		 t, (struct geom_point **) buff, buff_len, closed, points);

    tmp_data -> i ++;
    if (tmp_data -> i > tmp_data -> int_bbox . ur . x)
    {
        tmp_data -> i = tmp_data -> int_bbox . ll . y;
	tmp_data -> second = 1;
    }
  }
  else
  {
    return_line (tmp_data -> int_bbox . ll . x, tmp_data -> i,
	         tmp_data -> int_bbox . ur . x, tmp_data -> i,
		 t, (struct geom_point **) buff, buff_len, closed, points);

    tmp_data -> i ++;
    if (tmp_data -> i > tmp_data -> int_bbox . ur . y)
    {
	xfree (tmp_data);
	tmp_data = NULL;
    }
  }
  return (tmp_data);
}


// specific functions

struct vcl_grid *
vcl_grid_new (void)
{
  struct vcl_grid *obj = xmalloc (sizeof (struct vcl_grid));
  vcl_object_init (obj, vcl_vt_grid);
  obj -> parent = NULL;
  obj -> data = 0;

  return (obj);
}

void
vcl_grid_main (void)
{
  vcl_vt_grid = vcl_class_new ("grid");

  vcl_class_add_interface (vcl_vt_grid, vcl_if_object);
  vcl_class_add_method_li (vcl_mid_object_destroy, vcl_grid_destroy);

  vcl_class_add_interface (vcl_vt_grid, vcl_if_node);
  vcl_class_add_method_li (vcl_mid_node_set_data, vcl_grid_set_data);
  vcl_class_add_method_li (vcl_mid_node_get_data, vcl_grid_get_data);
  vcl_class_add_method_li (vcl_mid_node_set_parent, vcl_grid_set_parent);
  vcl_class_add_method_li (vcl_mid_node_get_parent, vcl_grid_get_parent);
  vcl_class_add_method_li (vcl_mid_node_get_bbox, vcl_grid_get_bbox);

  vcl_class_add_interface (vcl_vt_grid, vcl_if_shape);
  vcl_class_add_method_li (vcl_mid_shape_polygonize, vcl_grid_polygonize);
}

