/*
 * VCL module
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>

int vcl_if_object;
int vcl_mid_object_destroy;

static int class_counter = 1;
static int method_counter = 1;
static struct vcl_growing_array vcl_ifaces;

static vcl_vtable *li_class;
static int li_iface;
static int li_defined;

int
vcl_interface_new (const char *name)
{
  int i = vcl_ga_new_item (&vcl_ifaces);
  struct vcl_iface_dsc *idsc = &(VCL_GA_REF (&vcl_ifaces, i, struct vcl_iface_dsc));

  idsc -> first_fn = 0;
  idsc -> fn_count = 0;
  idsc -> name = name;
  return (i);
}

int
vcl_interface_add_method (int iface)
{
  int mid = method_counter ++;
  struct vcl_iface_dsc *idsc = &(VCL_GA_REF (&vcl_ifaces, iface, struct vcl_iface_dsc));

  idsc -> fn_count ++;
  if (idsc -> fn_count == 1) idsc -> first_fn = mid;
  return (mid);
}

vcl_vtable *
vcl_class_new (const char *name)
{
  struct vcl_class_dsc *cld = xmalloc (sizeof (struct vcl_class_dsc));
  vcl_vtable *vt = xmalloc_zero  (method_counter * sizeof (vcl_vtable));

  cld -> name = name;
  cld -> class_id = class_counter ++;
  vt [0] = cld;
  return (vt);
}

static const char *
vcl_class_get_name (vcl_vtable *vt)
{
  struct vcl_class_dsc *dsc = (struct vcl_class_dsc *) vt [0];
  return (dsc -> name);
}

uns
vcl_class_supports_interface_p (vcl_vtable *klass, int iface)
{
  struct vcl_iface_dsc *idsc = &(VCL_GA_REF (&vcl_ifaces, iface, struct vcl_iface_dsc));
  return (klass [idsc -> first_fn] != NULL);
}

void
vcl_class_add_interface (vcl_vtable *klass, int iface)
{
  struct vcl_iface_dsc *idsc = &(VCL_GA_REF (&vcl_ifaces, iface, struct vcl_iface_dsc));

  if (idsc -> fn_count == 0) die ("ERROR: Interface without methods used\n");

  klass [idsc -> first_fn] = (void *) 1;

  if ((li_class != NULL) &&
      (li_defined != VCL_GA_REF (&vcl_ifaces, li_iface, struct vcl_iface_dsc) . fn_count))
    msg ("WARNING: Maybe not defined all methods in interface %s in class %s\n",
	 VCL_GA_REF (&vcl_ifaces, li_iface, struct vcl_iface_dsc) . name,
	 vcl_class_get_name (li_class));

  li_class = klass;
  li_iface = iface;
  li_defined = 0;
}

void
vcl_class_add_method (vcl_vtable *klass, int iface UNUSED, int mid, void *call)
{
  if ((klass [mid] != NULL) && (klass [mid] != (void *) 1))
    {
      struct vcl_class_dsc * dsc = (struct vcl_class_dsc *) klass [0];
      die ("ERROR: method defined twice in %s\n", dsc -> name);
    }

  klass [mid] = call;
  li_defined ++;
}

void
vcl_class_add_method_li (int mid, void *call)
{
  vcl_class_add_method (li_class, li_iface, mid, call);
}

void
vcl_object_init (void * ptr, vcl_vtable *tbl)
{
  struct vcl_object * obj = ptr;
  obj -> vt = tbl;
}

void
vcl_object_main (void)
{
  vcl_ga_init (&vcl_ifaces, 0, sizeof (struct vcl_iface_dsc));

  vcl_if_object = vcl_interface_new ("object");
  vcl_mid_object_destroy = vcl_interface_add_method (vcl_if_object);
}
